/*----------------------------------------------------------------------------*
 * Link.c
 *
 * Michael Geikowski, 6.5.93
 *
 * Implementierung der Member Funktionen der Klassen Link und LinkTable
 *----------------------------------------------------------------------------*/

#include <assert.h>
#include <stdio.h>
#include <string.h>

#include "link.h"
#include "dbinterface.h"

#define BUCKETRANGE 100 	// Anzahl Positionen, die in einen
				// Bucket gehasht werden.

/*----------------------------------------------------------------------------*
 * Klasse Link
 *----------------------------------------------------------------------------*/

Link::~Link()
 { 
 if (LinkText)   delete LinkText;
 if (DefMod)     delete DefMod;
 if (PrevUseMod) delete PrevUseMod;
 if (NextUseMod) delete NextUseMod;
 if (Mod)        delete Mod;
 }

/*----------------------------------------------------------------------------*
 * Klasse LinkTable
 *
 * Die Verweise werden in einer Hashtabelle gespeichert. Es wird offenes
 * Hashing verwendet. Jeder Bucket ist eine Liste von Objekten der Klasse Link.
 * Die Hashfunktion lautet:
 *                          bucket = (Startposition des Links) / BUCKETRANGE 
 *----------------------------------------------------------------------------*/

LinkTable::LinkTable()
 {
 HashTable       = 0;
 HashTableLength = 0;
 }

LinkTable::~LinkTable()
 {
 if (HashTable) delete [] HashTable;
 }

int LinkTable::ReadLinkFile(char* filename)
 {
 Link *l;
 List *LinkList;
 long bucket, maxpos, pos, defpos, prevpos, nextpos, modpos;
 int  ok, leng;
 char *text, *defmodname, *prevmodname, *nextmodname, *destmodname;

 if (! OpenLinkFile(filename,"r",0) )
   {
   return 0;
   } 
 maxpos = GetMaxLinkPos();
 if (maxpos < 0)
   {
   CloseLinkFile();
   return 0;
   }
 HashTableLength = maxpos / BUCKETRANGE + 1;	// Groesse fuer HashTable
 HashTable = new List[HashTableLength];		// Table anlegen
 while(! EndOfLinkFile())			// Eintraege lesen
   {
   text = defmodname = prevmodname = nextmodname = destmodname = 0;
   ok = ReadLink(&pos, &leng, &text,
                 &defpos , &defmodname,
                 &prevpos, &prevmodname,
                 &nextpos, &nextmodname,
                 &modpos , &destmodname );
   bucket = pos / BUCKETRANGE;
   if ( (!ok) || (bucket<0) || (bucket >= HashTableLength) )
     {
     CloseLinkFile();
     return 0;
     }
   l = new Link(pos, leng, text,
                 defpos , defmodname,
                 prevpos, prevmodname,
                 nextpos, nextmodname,
                 modpos , destmodname );
   HashTable[bucket].Append(l);
   }
 CloseLinkFile();
 return 1;
 }

Link* LinkTable::GetLink(long pos)
 {
 assert(HashTable);
 long bucket = pos / BUCKETRANGE;	// suche bucket zu Position pos
 if (bucket >= HashTableLength)		// pos nicht in HashTable
   {
   return 0;
   }
					// durchsuche bucket, in dem pos
					// liegen muesste
 Link *l = (Link*) HashTable[bucket].GetFirstElem();
 while(l)
   {
   if ( (l->SourcePos <= pos) && (pos < l->SourcePos+l->SourceLength) )
     {
     return l;				// Link bzgl. pos gefunden
     }
   l = (Link*) HashTable[bucket].GetNextElem();
   }
					// pos koennte zu erstem Link vor
					// obigem bucket gehoeren.
 for (bucket--; bucket>=0; bucket--)
   {
   if (HashTable[bucket].GetLength()>0)	// Wenn bucket nicht leer
     {
     l = (Link*) HashTable[bucket].GetLastElem();
     if ( (l->SourcePos <= pos) && (pos < l->SourcePos+l->SourceLength) )
       {
       return l;
       }
     else
       {
       return 0;
       }
     }
   }
 return 0;
 }
