/*---------------------------------------------------------------------------*
 * ModWin.c
 *
 * Michael Geikowski, 30.6.93
 *
 * Generierung und Verwaltung der Modul Windows
 *---------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------*
 * Ergaenzung:
 * Olaf Meier,  18.08.94
 * Routinen zur Verwaltung des Strukturbaumfensters sowie zur
 * Typfehler-Anzeige in der Listen-Anzeige des Modul Windows.
 *--------------------------------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>


#include "list.h"
#include "adjlistcxx.h"     // Adjazenzlisten-Darstellung des Strukturbaums
#include "ptlayout.h"       // Layoutberechnung des aktuellen Baumausschnitts
#include "link.h"
#include "navwin.h"
#include "modwin.h"
#include "topwin.h"
#include "dbinterface.h"


#include <Xm/Xm.h>
#include <Xm/CascadeB.h>
#include <Xm/Form.h>
#include <Xm/Frame.h>
#include <Xm/DrawingA.h>     // fuer graphische Strukturbaum-Anzeige
#include <Xm/List.h>
#include <Xm/MessageB.h>
#include <Xm/MwmUtil.h>
#include <Xm/PanedW.h>
#include <Xm/PushB.h>
#include <Xm/RowColumn.h>
#include <Xm/ScrollBar.h>
#include <Xm/ScrolledW.h>
#include <Xm/SelectioB.h>
#include <Xm/Separator.h>
#include <Xm/Text.h>
#include <X11/Shell.h>

// Initiale Windowgroesse
//-----------------------


// Alle Routinen und Variablen fuer das Strukturbaum Window beginnen mit PT...


#define LISTVISIBLEITEMS 5

#define PTCANVASHEIGHT 600        // Zeichenflaeche Strukturbaum-Anzeige
#define PTCANVASWIDTH  1000
#define PTDIALOGHEIGHT 300       // Initioale Groesse der Strukturbaum-Dialogshell     
#define PTDIALOGWIDTH  500

// ScrollBar Identifier  (Strukturbaum-Anzeige)
//---------------------

#define PTSBHOR 1
#define PTSBVER 2

// PromptDialog client_data
//-------------------------

#define PDOK          1
#define PDCANCEL      2

// PromptDialog client_data (Parsetree Window- Eingabe von Max.Level)
//---------------------------------------------------------------------

#define PTPDOK        1
#define PTPDCANCEL    2



// MenuItem client_data
//---------------------

#define WMMAXMODWINS  1
#define WMCLOSE       2
#define GMGOTOTOP    10
#define GMBOOKMARKS  11
#define LMCONTENTS   20
#define LMERROR      21
#define LMTYPING     22   // Neuer Menueeintrag : Display Typing
#define PMDISPLAY    30   // Neuer Menueeintrag : Display Parsetree


// MenuItem client_data (Parsetree Window- Menueeintraege)
//----------------------------------------

#define PTNODELEVELS  10    // Eingabe von Max.Level
#define PTCLOSE       20    // Schliessen des Fensters
#define PTROOT        30    // Anzeige des obersten Baumausschnitts 


// ALIAS Namen fuer einige AdjList Routinen  
//-----------------------------------------

#define GETXPOS()                 GetPos()
#define SETXPOS(x)                SetPos(x)
#define GETYPOS()                 GetTempPos()
#define SETYPOS(y)                SetTempPos(y) 
#define GETWIDGETNR()             GetLevel()
#define SETWIDGETNR(nr)           SetLevel(nr)
#define GETNODEOFFSET()           GetCounter()
#define SETNODEOFFSET(offset)     SetCounter(offset)

// Macros
//--------

#define max(a,b) ((a) > (b)) ? (a) : (b)
#define min(a,b) ((a) < (b)) ? (a) : (b)

class ModWinListElem;

class MenuData : public ListElem	// Client_data fuer MenuCB
 {
 public:

 int ItemID;
 ModWinListElem* mw;
 };

// Bereits geoeffnete Modul Windows werden in einer Liste verwaltet.
// Ein Element dieser Liste hat die Form:
//---------------------------------------

class ModWinListElem : public ListElem
 {
 public:

   ModWinListElem();
   ~ModWinListElem();

   char* name;		// Modulname
   Widget theshell;	// shell-, list- und textwidget des Modul Windows
   Widget thelist;
   Widget thetext;
   long*  ListDestPos;	// Arrays von Position und Laenge der Sprungziele
   int*   ListDestLen;	// der einzelnen Listenelemente
   long   SourcePos;	// Position und Laenge des aktuellen Symbols
   int    SourceLength;
   int    touched;	// Zaehler zur Berechnung des 'Alters' des Fensters
   LinkTable *lt;	// Array aller von dem Fenster ausgehenden Links
   List* mdList;	// Liste aller MenuData Objekte des Fensters

// Zusaetzliche Informationen fuer die graphische Parsetree-Anzeige:

   Widget thePTdialogshell;
   Widget thePTcanvas;
   Widget *PTWidgetArray;
   Widget PTScrollBarHor;
   Widget PTScrollBarVer;


   AdjList *thePT;      // fuer den gesamtem Parsetree des Spezifikationsmoduls
   AdjList *theactPTpart;  // fuer den aktuellen Parsetree-Ausschnitt

   List *PTbuttonList;    
   List *PTwininfo;

   int PTWidgetArrayLength;
   int PTx_offset;
   int PTy_offset;
   int PTGraphWidth;
   int PTGraphHeight;
   int maxLevels;    // aktuell eingegebene maximale Levelanzahl (initial: 3)

 };

ModWinListElem::ModWinListElem()
 {
 name=0;
 lt = 0;
 ListDestPos=0;
 ListDestLen=0;
 mdList = new List;
 PTwininfo = new List;
 PTWidgetArray = 0;
 PTWidgetArrayLength = 0;
 PTbuttonList = 0;
 PTGraphWidth = 0;
 PTGraphHeight = 0;
 
 }

ModWinListElem::~ModWinListElem()
 {
 if(PTWidgetArray)
  {
   for(int i = 0; i < PTWidgetArrayLength; i++)
    {
      XtDestroyWidget(PTWidgetArray[i]);
    }
   delete [] PTWidgetArray;
  } 
 XtDestroyWidget(theshell);
 if (name)        delete name;
 if (lt)          delete lt;
 if (ListDestPos) delete ListDestPos;
 if (ListDestLen) delete ListDestLen;
 delete mdList;
 if(PTbuttonList) delete PTbuttonList;    // Parsetree Infos loeschen
 if(PTwininfo) delete PTwininfo;
 if(thePT) delete thePT;
 if(theactPTpart) delete theactPTpart; 
 }

// Modulglobale Variablen
//-----------------------

static Widget thetoplevel;		// Merker
static char* thedir=0;
static char* thesuffix=0;
static List *ModWinList = new List;	// Liste aller offenen Fenster
static int   touch = 0;	   // Zaehler zur Berechnung des 'Alters' der Fenster
static int   maxopenwindows = 2;	// max Anzahl offener Fenster
                                      
static GC PTGCcopy, PTGCdash;           // Linen fuer Parsetree Darstellung
static Dimension PTFontWidth,           // Schriftgroesse fue die Buttonbeschriftung
                 PTFontHeight,
                 PTHalfFontHeight;
                

// Modul-Inhalt oder Fehler oder Typ-Warnungen einlesen
//---------------------------------------------------------

static Boolean ReadContents(ModWinListElem* mw, char* name)
	// Inhaltsverzeichnis zu einem Modul lesen und darstellen
 {
 int  i,listlength,ok;
 char *line;
 XmString line_str;

 XmListSetPos(mw->thelist,1);		// Alten Listeninhalt loeschen
 XmListDeleteAllItems(mw->thelist);
 if (mw->ListDestPos) delete mw->ListDestPos;
 if (mw->ListDestLen) delete mw->ListDestLen;
 mw->ListDestPos=0;
 mw->ListDestLen=0;
 if (! OpenContentsFile(name,"r"))	// Inhaltsverzeichnis lesen
   {
   return False;
   }
 listlength = GetNrOfContentsLines();
 if (listlength<=0)
   {
   CloseContentsFile();
   return False;
   }
 mw->ListDestPos = new long[listlength];	// Sprungziele 
 mw->ListDestLen = new int[listlength];
 for (i=0; i<listlength; i++)
   {
   ok = ReadContentsLine(&line,&(mw->ListDestPos[i]),&(mw->ListDestLen[i]));
   if (!ok)
     {
     CloseContentsFile();
     return False;
     }
   line_str = XmStringCreateSimple(line);	// Zeilenweise anzeigen
   delete line;
   XmListAddItem(mw->thelist,line_str,0);
   XmStringFree(line_str);
   }
 CloseContentsFile();
 return True;
 }

static Boolean ReadErrors(ModWinListElem* mw, char* name)
	// Fehler-Liste zu einem Modul lesen und darstellen
 {
 int  i,err,ok;
 char *line;
 XmString line_str;

 XmListSetPos(mw->thelist,1);		// Alten Listeninhalt loeschen
 XmListDeleteAllItems(mw->thelist);
 if (mw->ListDestPos) delete mw->ListDestPos;
 if (mw->ListDestLen) delete mw->ListDestLen;
 mw->ListDestPos=0;
 mw->ListDestLen=0;
 if (! OpenErrorFile(name,"r"))		// Fehler-Liste lesen
   {
   return False;
   }
 err = GetNrOfErrors();
 if (err<=0)
   {
   CloseErrorFile();
   return False;
   }
 mw->ListDestPos = new long[err];	// Sprungziele
 mw->ListDestLen = new int[err];
 for (i=0; i<err; i++)
   {
   ok = ReadErrorLine(&line,&(mw->ListDestPos[i]),&(mw->ListDestLen[i]));
   if (!ok)
     {
     CloseErrorFile();
     return False;
     }
   line_str = XmStringCreateSimple(line); // Fehler zeilenweise anzeigen
   delete line;
   XmListAddItem(mw->thelist,line_str,0);
   XmStringFree(line_str);
   }
 CloseErrorFile();
 return True;
 }


static Boolean ReadWarnings(ModWinListElem* mw, char* name)
        // Typ-Warnungs-Liste zu einem Modul lesen und darstellen
 {
 int  i,warn,ok;
 char *line;
 XmString line_str;

 XmListSetPos(mw->thelist,1);           // Alten Listeninhalt loeschen
 XmListDeleteAllItems(mw->thelist);
 if (mw->ListDestPos) delete mw->ListDestPos;
 if (mw->ListDestLen) delete mw->ListDestLen;
 mw->ListDestPos=0;
 mw->ListDestLen=0;
 if (! OpenWarningFile(name,"r"))         // Warnungs-Liste lesen
   {
   return False;
   }
 warn = GetNrOfWarnings();
if (warn<=0)
   {
   CloseWarningFile();
   return False;
   }
 mw->ListDestPos = new long[warn];       // Sprungziele
 mw->ListDestLen = new int[warn];
 for (i=0; i<warn; i++)
   {
   ok = ReadWarningLine(&line,&(mw->ListDestPos[i]),&(mw->ListDestLen[i]));
   if (!ok)
     {
     CloseWarningFile();
     return False;
     }
   line_str = XmStringCreateSimple(line); // Warnungen  zeilenweise anzeigen
   delete line;
   XmListAddItem(mw->thelist,line_str,0);
   XmStringFree(line_str);
   }
 CloseWarningFile();
 return True;
 }


// Initiales Einlesen des Parsetrees und Errechnung des Layouts des obersten Baumausschnitts
static void InitPTGraph(ModWinListElem* mw, char* modname)
 {

  char *name;
  long min, max;
  int topx, counter, value, ok, endofnodes, endofedges;
  char t;
  Node *n, *m;
  Arg al[10];
  int ac;
  XmString name_str;
  Dimension width, height, maxwidth;
  MenuData *binfo; 

  void PTNodeCB(Widget w, XtPointer client_data, XtPointer call_data);

  mw->maxLevels = 3;    // Parsetree-Darstellung eines anderen Moduls loeschen
  if(mw->PTbuttonList) delete mw->PTbuttonList; 
  if(mw->thePT) {delete mw->thePT; mw->thePT = 0;}
  if(mw->theactPTpart) {delete mw->theactPTpart; mw->theactPTpart = 0;}
  if(mw->PTWidgetArray)
   {
     for(int i = 0; i < mw->PTWidgetArrayLength; i++)
      {
        XtDestroyWidget(mw->PTWidgetArray[i]);
      }
     delete [] mw->PTWidgetArray;
     mw->PTWidgetArray = 0;
     mw->PTWidgetArrayLength = 0;
   }
  mw->PTGraphHeight = 0;
  mw->PTGraphWidth = 0;

  // Einlesen des gesammten Parsetrees

  if(!OpenParseTreeFile(modname,"r")) return;
  mw->thePT = new AdjList;
  ok = ReadPTNode(&t, &name ,&min, &max, &endofnodes);
  while(!endofnodes && ok)
   { 
     n = mw->thePT->PTAddNode(name, t, min, max);
     ok = ok && ReadPTEdge(&t , &name , &min , &max, &endofedges);
     while(!endofedges && ok)
      {
        m = mw->thePT->PTAddNode(name, t, min ,max);
        n->PTAddEdge(m, t);
        ok = ok && ReadPTEdge(&t, &name, &min, &max, &endofedges);
      }
     ok = ok && ReadPTNode(&t, &name , &min, &max, &endofnodes);
   }

  CloseParseTreeFile();
  
  if(!ok)
   {
     delete mw->thePT;
     mw->thePT = 0;
     return;
   }

// Darstellung des obersten Baumausschnitts
  mw->theactPTpart = new AdjList;
  mw->PTbuttonList = new List;
  n = mw->thePT->GetFirstNode();
  if(n && n->HasPTsuccessor())
   {


       m = CopyAdjList(n,mw->theactPTpart,mw->maxLevels) ;
       m->SetNodeType('U');
       



     ok = PTLayout(mw->theactPTpart, &(mw->PTGraphWidth), &(mw->PTGraphHeight));


     mw->PTGraphWidth = mw->PTGraphWidth * PTFontWidth;
     mw->PTGraphHeight = mw->PTGraphHeight * PTFontHeight;

                                                                      
     counter = 0;
     topx = -1;
     n = mw->theactPTpart->GetFirstNode();
     while(n)
      { 
        n->SETXPOS(n->GETXPOS() * PTFontWidth);
        n->SETYPOS(n->GETYPOS() * PTFontHeight);

        if( topx == -1)
         {
           topx = n->GETXPOS();
         }
        counter++;
        n = mw->theactPTpart->GetNextNode();
      }
     mw->PTWidgetArray = new Widget[counter];
     mw->PTWidgetArrayLength = counter;
     counter = 0;
     n = mw->theactPTpart->GetFirstNode();
     while(n)
      {

        name = n->GetName();
        name_str =  XmStringCreateSimple(name);
        ac = 0;
        XtSetArg(al[ac], XmNlabelString, name_str); ac++;
        XtSetArg(al[ac], XmNx, 10000); ac++;
        XtSetArg(al[ac], XmNy, 10000); ac++;
        mw->PTWidgetArray[counter] = XmCreatePushButton(mw->thePTcanvas,"name",al,ac);
        XtManageChild(mw->PTWidgetArray[counter]);
        binfo = new MenuData;
        binfo->mw = mw;
        binfo->ItemID = counter;
        mw->PTbuttonList->Append(binfo);
        XtAddCallback(mw->PTWidgetArray[counter] , XmNactivateCallback,
                      PTNodeCB, XtPointer(binfo));
        n->SETWIDGETNR(counter);
        counter++;
        XmStringFree(name_str);
        n->SETNODEOFFSET(PTHalfFontHeight);
        n = mw->theactPTpart->GetNextNode();
      }

     ac = 0;
     XtSetArg(al[ac],XmNheight, &height);ac++;
     XtSetArg(al[ac],XmNwidth, &width);ac++;
     XtGetValues(mw->thePTcanvas,  al, ac);
     mw->PTy_offset = 0;
     if(width> mw->PTGraphWidth)
      mw->PTx_offset = (mw->PTGraphWidth - width)/2;
     else
      mw->PTx_offset = topx-width/2;
     maxwidth = max(mw->PTGraphWidth, width);
     value = min(mw->PTx_offset, maxwidth - width);
     value = max(0,value);
     ac = 0;
     XtSetArg(al[ac],XmNmaximum, maxwidth);ac++;
     XtSetArg(al[ac], XmNsliderSize,width); ac++;
     XtSetArg(al[ac], XmNpageIncrement, max(1, width/2));ac++;
     XtSetArg(al[ac], XmNvalue, value); ac++;
     XtSetValues(mw->PTScrollBarHor, al, ac);
     ac = 0;
     XtSetArg(al[ac],XmNmaximum, max(mw->PTGraphHeight, height));ac++;
     XtSetArg(al[ac], XmNsliderSize,height);ac++;
     XtSetArg(al[ac], XmNpageIncrement, max(1,height/2));ac++;
     XtSetArg(al[ac], XmNvalue,mw->PTy_offset);ac++;
     XtSetValues(mw->PTScrollBarVer,al,ac);
       
   }
 
  }

// Loeschen des aktuellen Parsetree-Ausschnitts
// und Darstellung des obersten Strukturbaum-Auschnitts vom gesamten Baum des Spezifikationsmoduls 
static void PTCreateRootTree(ModWinListElem* mw)
 {

  char* name;
  Node *n, *m;
  Arg al[10];
  int ac ;
  int topx, counter, value, ok;
  char t;
  XmString name_str;
  Dimension width, height, maxwidth;
  MenuData *binfo;


  void PTNodeCB(Widget w, XtPointer client_data, XtPointer call_data);
// Loeschen des aktuellen Ausschnitts
  if(mw->thePT)
   {
    if(mw->PTbuttonList) delete mw->PTbuttonList;
    if(mw->theactPTpart) {delete mw->theactPTpart; mw->theactPTpart = 0;}
    if(mw->PTWidgetArray)
     {
      for(int i = 0; i < mw->PTWidgetArrayLength; i++)
       {
         XtDestroyWidget(mw->PTWidgetArray[i]);
       }
      delete [] mw->PTWidgetArray;
      mw->PTWidgetArray = 0;
      mw->PTWidgetArrayLength = 0;
      }
    mw->PTGraphHeight = 0;
    mw->PTGraphWidth = 0;

    mw->theactPTpart = new AdjList;
    mw->PTbuttonList = new List;

    n = mw->thePT->GetFirstNode();
// Darstellung des obersten Ausschnitts
    if(n && n->HasPTsuccessor())
     {
       m = CopyAdjList(n,mw->theactPTpart,mw->maxLevels) ;
       m->SetNodeType('U');   // Oberster Knoten erhaelt Kennung 'U'-pwards


// Layoutberechnung
       ok = PTLayout(mw->theactPTpart, &(mw->PTGraphWidth), &(mw->PTGraphHeight));
       mw->PTGraphWidth = mw->PTGraphWidth * PTFontWidth;
       mw->PTGraphHeight = mw->PTGraphHeight * PTFontHeight;


      counter = 0;
      topx = -1;
      n = mw->theactPTpart->GetFirstNode();
      while(n)
      {
        n->SETXPOS(n->GETXPOS() * PTFontWidth);
        n->SETYPOS(n->GETYPOS() * PTFontHeight);

        if( topx == -1)
         {
           topx = n->GETXPOS();
         }
        counter++;
        n = mw->theactPTpart->GetNextNode();
      }
      mw->PTWidgetArray = new Widget[counter];
      mw->PTWidgetArrayLength = counter;
      counter = 0;
      n = mw->theactPTpart->GetFirstNode();
      while(n)
      {
        name = n->GetName();
        name_str =  XmStringCreateSimple(name);
        ac = 0;
        XtSetArg(al[ac], XmNlabelString, name_str); ac++;
        XtSetArg(al[ac], XmNx, 10000); ac++;
        XtSetArg(al[ac], XmNy, 10000); ac++;
        mw->PTWidgetArray[counter] = XmCreatePushButton(mw->thePTcanvas,"name",al,ac);
        XtManageChild(mw->PTWidgetArray[counter]);
        binfo = new MenuData;
        binfo->mw = mw;
        binfo->ItemID = counter;
        mw->PTbuttonList->Append(binfo);
        XtAddCallback(mw->PTWidgetArray[counter] , XmNactivateCallback,
                      PTNodeCB, XtPointer(binfo));
        n->SETWIDGETNR(counter);
        counter++;
        XmStringFree(name_str);
        n->SETNODEOFFSET(PTHalfFontHeight);
        n = mw->theactPTpart->GetNextNode();
      }

      ac = 0;
      XtSetArg(al[ac],XmNheight, &height);ac++;
      XtSetArg(al[ac],XmNwidth, &width);ac++;
      XtGetValues(mw->thePTcanvas,  al, ac);
      mw->PTy_offset = 0;
      if(width> mw->PTGraphWidth)
       mw->PTx_offset = (mw->PTGraphWidth - width)/2;
      else
       mw->PTx_offset = topx-width/2;
      maxwidth = max(mw->PTGraphWidth, width);
      value = min(mw->PTx_offset, maxwidth - width);
      value = max(0,value);
      ac = 0;
      XtSetArg(al[ac],XmNmaximum, maxwidth);ac++;
      XtSetArg(al[ac], XmNsliderSize,width); ac++;
      XtSetArg(al[ac], XmNpageIncrement, max(1, width/2));ac++;
      XtSetArg(al[ac], XmNvalue, value); ac++;
      XtSetValues(mw->PTScrollBarHor, al, ac);
      ac = 0;
      XtSetArg(al[ac],XmNmaximum, max(mw->PTGraphHeight, height));ac++;
      XtSetArg(al[ac], XmNsliderSize,height);ac++;
      XtSetArg(al[ac], XmNpageIncrement, max(1,height/2));ac++;
      XtSetArg(al[ac], XmNvalue,mw->PTy_offset);ac++;
      XtSetValues(mw->PTScrollBarVer,al,ac);
// Zeichenflaeche loeschen und Expose-Event erzeugen
      XClearArea(XtDisplay(mw->thePTcanvas),XtWindow(mw->thePTcanvas),
                0,0,0,0,1);
   }
 }
}

// Erstellung des aktuellen Baumausschnitts in Abhaengigkeit des aktivierten Knotens
// in der Strukturbaumanzeige
static void CreatePTnewGraph(ModWinListElem* mw)
 {  

    MenuData *binfo;
    Node *n;
    char* name;
    Arg al[10];
    int ac ;
    int topx, counter, value, ok;
    XmString name_str;
    Dimension width, height, maxwidth;

    void PTNodeCB(Widget w, XtPointer client_data, XtPointer call_data);
    // Alte Graphik-Daten loeschen
    if(mw->PTbuttonList) delete mw->PTbuttonList;
    if(mw->PTWidgetArray)
       {
         for(int i = 0; i < mw->PTWidgetArrayLength; i++)
           {
             XtDestroyWidget(mw->PTWidgetArray[i]);
            }
          delete [] mw->PTWidgetArray;
          mw->PTWidgetArray = 0;
          mw->PTWidgetArrayLength = 0;
       }

// neue Graphik-Daten erstellen
    mw->PTbuttonList = new List;


    counter = 0;
    topx = -1;
    n = mw->theactPTpart->GetFirstNode();
    while(n)
     {
       n->SETXPOS(n->GETXPOS() * PTFontWidth);
       n->SETYPOS(n->GETYPOS() * PTFontHeight);

       if( topx == -1)
        {
        topx = n->GETXPOS();
        }
       counter++;
       n = mw->theactPTpart->GetNextNode();
      }

    mw->PTWidgetArray = new Widget[counter];
    mw->PTWidgetArrayLength = counter;
    counter = 0;
    n = mw->theactPTpart->GetFirstNode();
    while(n)
     {
      name = n->GetName();
      name_str =  XmStringCreateSimple(name);
      ac = 0;
      XtSetArg(al[ac], XmNlabelString, name_str); ac++;
      XtSetArg(al[ac], XmNx, 10000); ac++;
      XtSetArg(al[ac], XmNy, 10000); ac++;
      mw->PTWidgetArray[counter] = XmCreatePushButton(mw->thePTcanvas,"name",al,ac);
      XtManageChild(mw->PTWidgetArray[counter]);
      binfo = new MenuData;
      binfo->mw = mw;
      binfo->ItemID = counter;
      mw->PTbuttonList->Append(binfo);
      XtAddCallback(mw->PTWidgetArray[counter] , XmNactivateCallback,
                                 PTNodeCB, XtPointer(binfo));
      n->SETWIDGETNR(counter);
      counter++;
      XmStringFree(name_str);
      n->SETNODEOFFSET(PTHalfFontHeight);
      n = mw->theactPTpart->GetNextNode();
     }
   ac = 0;
   XtSetArg(al[ac],XmNheight, &height);ac++;
   XtSetArg(al[ac],XmNwidth, &width);ac++;
   XtGetValues(mw->thePTcanvas,  al, ac);
   mw->PTy_offset = 0;
   if(width> mw->PTGraphWidth)
     mw->PTx_offset = (mw->PTGraphWidth - width)/2;
   else
     mw->PTx_offset = topx-width/2;
   maxwidth = max(mw->PTGraphWidth, width);
   value = min(mw->PTx_offset, maxwidth - width);
   value = max(0,value);
   ac = 0;
   XtSetArg(al[ac],XmNmaximum, maxwidth);ac++;
   XtSetArg(al[ac], XmNsliderSize,width); ac++;
   XtSetArg(al[ac], XmNpageIncrement, max(1, width/2));ac++;
   XtSetArg(al[ac], XmNvalue, value); ac++;
   XtSetValues(mw->PTScrollBarHor, al, ac);
   ac = 0;
   XtSetArg(al[ac],XmNmaximum, max(mw->PTGraphHeight, height));ac++;
   XtSetArg(al[ac], XmNsliderSize,height);ac++;
   XtSetArg(al[ac], XmNpageIncrement, max(1,height/2));ac++;
   XtSetArg(al[ac], XmNvalue,mw->PTy_offset);ac++;
   XtSetValues(mw->PTScrollBarVer,al,ac);
// Zeichenlaeche loeschen und Expose-Event erzeugen
   XClearArea(XtDisplay(mw->thePTcanvas),XtWindow(mw->thePTcanvas),
                0,0,0,0,1);
  }

// Callback Funktionen
//--------------------

static void DeleteHighlights()
	// Hilfsfunktion : loescht Highlights in allen Fenstern
 {
 ModWinListElem *mw;

 mw = (ModWinListElem*) ModWinList->GetFirstElem();
 while (mw)
   {
   if (mw->SourceLength>0)
     {
     XmTextSetHighlight(mw->thetext,mw->SourcePos,
                                    mw->SourcePos+mw->SourceLength,
                                    XmHIGHLIGHT_NORMAL);
     mw->SourceLength=0;
     }
   mw = (ModWinListElem*) ModWinList->GetNextElem();
   }
 }


// Callbackfunktion (Groesse des Strukturbaumfensters wird veraendert)
static void PTResizeCB(Widget w, XtPointer client_data, XtPointer call_data)
  {

  Arg al[10];
  int ac;
  int value;
  Dimension height, width, maxwidth, maxheight;
  MenuData *info = (MenuData*) client_data;


  ac = 0;
  XtSetArg(al[ac], XmNheight, &height);ac++;
  XtSetArg(al[ac], XmNwidth, &width);ac++;
  XtGetValues( info->mw->thePTcanvas, al, ac);
  maxwidth = max(info->mw->PTGraphWidth,width);
  if(width > info->mw->PTGraphWidth)
   {
    info->mw->PTx_offset = (info->mw->PTGraphWidth - width) / 2;
  }

  value = min(info->mw->PTx_offset, maxwidth - width);
  value = max(0, value);
  if(width <= info->mw->PTGraphWidth)
   {
    info->mw->PTx_offset = value;
   }

  ac = 0;   // Scrollbars neu ausrichten
  XtSetArg(al[ac], XmNsliderSize, width);ac++;
  XtSetArg(al[ac], XmNpageIncrement, max(1, width/2));ac++;
  XtSetArg(al[ac], XmNmaximum, maxwidth);ac++;
  XtSetArg(al[ac], XmNvalue,value);ac++;
  XtSetValues(info->mw->PTScrollBarHor,al,ac);
  maxheight = max(info->mw->PTGraphHeight, height);
  value = min(info->mw->PTy_offset, maxheight-height);
  value = max(0,value);
  ac = 0;
  XtSetArg(al[ac], XmNsliderSize, height); ac++;
  XtSetArg(al[ac], XmNpageIncrement, max(1,height/2));ac++;
  XtSetArg(al[ac], XmNmaximum , max(info->mw->PTGraphHeight, height)); ac++;
  XtSetArg(al[ac], XmNvalue, value); ac++;
  XtSetValues(info->mw->PTScrollBarVer, al, ac);



if(XtWindow(info->mw->thePTcanvas) != NULL)
 XClearArea(XtDisplay( info->mw->thePTcanvas),XtWindow( info->mw->thePTcanvas),
             0,0,0,0,1);

  }           
  
// Callbackfuntion (Schiebebalken des Strukturbaumfensters wurden veraendert)
static void PTScrollCB(Widget w, XtPointer client_data, XtPointer call_data)
  {

  Arg al[10];
  int ac;
  int value;
  Dimension width;
  MenuData *md = (MenuData*) client_data;


  ac =  0;
  XtSetArg(al[ac], XmNvalue, &value); ac++;
  XtGetValues(w,al,ac);
  if(md->ItemID == PTSBHOR)
   {
   ac = 0;
   XtSetArg(al[ac], XmNwidth, &width); ac++;
   XtGetValues(md->mw->thePTcanvas, al, ac);
   if (width <= md->mw->PTGraphWidth)
    {
      md->mw->PTx_offset = value;
    }
   }
  else
   {
   md->mw->PTy_offset = value;
   }

  XClearArea(XtDisplay(md->mw->thePTcanvas), XtWindow(md->mw->thePTcanvas),
             0,0,0,0,1);

            
  }


// Callbackfunktion (Neuzeichnen des Strukturbaumfensters)
static void PTExposeCB(Widget w, XtPointer client_data, XtPointer call_data)
  {


  XmDrawingAreaCallbackStruct* cbp = (XmDrawingAreaCallbackStruct*) call_data;
  XExposeEvent* event = (XExposeEvent*) cbp->event;
  MenuData* info = (MenuData*) client_data;
  Node *n ,*m;
  char t;
  Arg al[10];
  int ac;
  int counter;
  Dimension NodeWidth, NodeHeight;

  if(event->count != 0) return;
  if(! (info->mw->theactPTpart)) return;

  counter = 0;
  n = info->mw->theactPTpart->GetFirstNode();
  while(n)
   {
    ac = 0;
    XtSetArg(al[ac], XmNwidth, &NodeWidth); ac++;
    XtSetArg(al[ac], XmNheight,&NodeHeight); ac++;
    XtGetValues(info->mw->PTWidgetArray[counter], al, ac);
    XMoveWindow(XtDisplay(info->mw->PTWidgetArray[counter]),
                XtWindow(info->mw->PTWidgetArray[counter]),
                n->GETXPOS() - info->mw->PTx_offset - NodeWidth/2,
                n->GETYPOS() - info->mw->PTy_offset - NodeHeight/2);
    counter++;
    m = n->GetFirstEdge();
    while(m)
     { 
       t = n->GetTypeOfCurrentEdge();
       if(t == 'T')
        XDrawLine(XtDisplay(info->mw->thePTcanvas),
                  XtWindow( info->mw->thePTcanvas),
                  PTGCdash,
                  n->GETXPOS() - info->mw->PTx_offset,
                  n->GETYPOS() - info->mw->PTy_offset + n->GETNODEOFFSET(),
                  m->GETXPOS() - info->mw->PTx_offset,
                  m->GETYPOS() - info->mw->PTy_offset - m->GETNODEOFFSET());
       else
        XDrawLine(XtDisplay(info->mw->thePTcanvas),
                  XtWindow(info->mw->thePTcanvas),
                  PTGCcopy,
                  n->GETXPOS() - info->mw->PTx_offset,
                  n->GETYPOS() - info->mw->PTy_offset + n->GETNODEOFFSET(),
                  m->GETXPOS() - info->mw->PTx_offset,
                  m->GETYPOS() - info->mw->PTy_offset - m->GETNODEOFFSET());  
       m = n->GetNextEdge();
       }
    n = info->mw->theactPTpart->GetNextNode();
    }          


  }


// Callbackfunktion (Ein Knoten innerhalb der Strukturbaum-Anzeige wurde angewaehlt) 

static void PTNodeCB(Widget w, XtPointer client_data, XtPointer call_data)
 {

   MenuData* md = (MenuData*) client_data;
   Node *n,*m;
   int  ok,newGraph;
   char t;
   XmTextPosition pos, maxpos;


   n = md->mw->theactPTpart->PTSearchforNodeLevel(md->ItemID);
   if(n)
    { 
      maxpos = XmTextGetLastPosition(md->mw->thetext);
      pos = n->GetPTminpos();
      touch++;
      md->mw->touched = touch;
      DeleteHighlights();
      ClearNavWin();
      md->mw->SourcePos = pos;
      md->mw->SourceLength = n->GetPTmaxpos() - pos + 1;
      XmTextSetHighlight(md->mw->thetext,
                         md->mw->SourcePos,
                         md->mw->SourcePos + md->mw->SourceLength,
                         XmHIGHLIGHT_SELECTED);
      XmTextShowPosition(md->mw->thetext, pos);
      XmTextSetInsertionPosition(md->mw->thetext, maxpos);  
     
     if(n->GetNodeType() == 'U')    // oberster Knoten innerhalb des Baumausschnitts wurde
                                    // angewaehlt.
       { 
         


         m = md->mw->thePT->PTSearchforNode(n);
         if(m->HasPTpredecessor())  // Knoten hat Vorgaenger -> neuer Baumausschnitt
           { 
             newGraph = 1;
             if(md->mw->theactPTpart) delete md->mw->theactPTpart; 
             md->mw->theactPTpart = new AdjList;
             n =  m->PTGoUpwards(md->mw->maxLevels - 1);
             m = CopyAdjList(n, md->mw->theactPTpart, md->mw->maxLevels);
             m->SetNodeType('U'); 
             ok = PTLayout(md->mw->theactPTpart, &(md->mw->PTGraphWidth), &(md->mw->PTGraphHeight));
             md->mw->PTGraphWidth = md->mw->PTGraphWidth * PTFontWidth;
             md->mw->PTGraphHeight = md->mw->PTGraphHeight * PTFontHeight;
           } 
          else
            newGraph = 0;
        }
      else if(n->GetNodeType() == 'D')  // Blattknoten innerhalb des Baumausschnitts wurde
                                        // angewaehlt.
        {
           m = md->mw->thePT->PTSearchforNode(n);
           if(m->HasPTsuccessor())  // Knoten hat Nachfolger -> neuer Baumausschnitt
            {
              newGraph = 1;
              if(md->mw->theactPTpart) delete md->mw->theactPTpart;
              md->mw->theactPTpart = new AdjList;
              n = CopyAdjList(m, md->mw->theactPTpart, md->mw->maxLevels);
              n->SetNodeType('U');
              ok = PTLayout(md->mw->theactPTpart, &(md->mw->PTGraphWidth), &(md->mw->PTGraphHeight));
              md->mw->PTGraphWidth = md->mw->PTGraphWidth * PTFontWidth;
              md->mw->PTGraphHeight = md->mw->PTGraphHeight * PTFontHeight;
            }
           else
             newGraph = 0;
         }
      
        }


   if(newGraph)  //  neuer Baumausschnitt 
     {  
        CreatePTnewGraph(md->mw);  // neuen Baumausschnitt darstellen
     }

  }
  

static void PromptCB(Widget w, XtPointer client_data, XtPointer call_data)
	// Maximale Anzahl offener Fenster wurde mit Prompt Dialog geaendert
 {
 int button = (int) client_data;
 XmSelectionBoxCallbackStruct* cbs = (XmSelectionBoxCallbackStruct*) call_data;
 ModWinListElem* mw;
 char *s;
 int max=0;
 int i,ok;

 switch (button)
   {
   case PDOK:		// Neues Maximum bestimmen
     XmStringGetLtoR(cbs->value,XmSTRING_DEFAULT_CHARSET,&s);
     ok=1;
     for (i=0; s[i] != '\0'; i++)
       ok = ok && isdigit(s[i]);
     ok = ok && (sscanf(s,"%d",&max) == 1);
       XtFree(s);
     if ( (!ok) || (max<=0) ) return;   // Prompt Dialog bleibt offen bei
				        // Fehleingabe
     maxopenwindows = max;		// Schliesse Unerwuenschte Fenster
     while( ModWinList->GetLength() > maxopenwindows )
       {
       ModWinList->DelFirstElem();
       }
     DeleteHighlights();
     ClearNavWin();
     break;
   case PDCANCEL:	// Prompt Dialog schliessen ohne Aenderung
     break;
   default: assert(0);
   };
 XtDestroyWidget(w);
 }



// Callbackfunktion (Knopf in der  Parsetree Dialog-Anzeige wurde angewaehlt) 
static void PTPromptCB(Widget w, XtPointer client_data, XtPointer call_data)
 {
 MenuData *md = (MenuData*) client_data;
 XmSelectionBoxCallbackStruct* cbs = (XmSelectionBoxCallbackStruct*) call_data;
 char *s;
 int max = 0;
 int i,ok;

 switch(md->ItemID)
   {
   case PTPDOK : // Neue, maximale Baumebenenzahl bestimmen 
                 XmStringGetLtoR(cbs->value,XmSTRING_DEFAULT_CHARSET,&s);
                 ok = 1;
                 for (i = 0; s[i] != '\0'; i++)
                   ok = ok && isdigit(s[i]);
                 ok = ok && (sscanf(s,"%d",&max) == 1);
                  XtFree(s);
                 if( (!ok) || (max < 3) || (max > 8)) return; // Fehleingabe !
                 md->mw->maxLevels = max;
                 break;
   case PTPDCANCEL :  // Prompt Dialogfenster ohne Aenderung schliessen
                 break;
   default : assert(0);
   };
 XtDestroyWidget(w);
 }


static void PTMenuCB(Widget w, XtPointer client_data, XtPointer call_data)
   // Ein MenuItem aus dem Parsetree Window wurde ausgewaehlt
 {
 MenuData* md = (MenuData*) client_data;
 char value[10];
 XmString label_str, value_str, title_str;
 Arg al[10];
 int ac;
 MenuData *PTinfo;
 Widget PTPromptDialog;
  

 switch (md->ItemID)    // MenuItem-Kennung
   {
   case PTNODELEVELS : // oeffne Prompt Dialog fuer max. anzuzeigende Baumebenen
                       sprintf(value,"%d",md->mw->maxLevels);
                       value_str = XmStringCreateSimple(value);
                       label_str = XmStringCreateSimple("Max. number of node levels [3..8].");
                       title_str = XmStringCreateSimple("Max. Levels");
                       ac = 0;
                       XtSetArg(al[ac], XmNselectionLabelString,label_str);ac++;
                       XtSetArg(al[ac], XmNtextString,value_str);ac++;
                       XtSetArg(al[ac], XmNautoUnmanage,False);ac++;
                       XtSetArg(al[ac], XmNdialogStyle,XmDIALOG_FULL_APPLICATION_MODAL);ac++;
                       XtSetArg(al[ac], XmNdialogTitle,title_str);ac++;
                       PTPromptDialog = XmCreatePromptDialog(md->mw->thePTdialogshell,"PTprompt",
                                                             al,ac);
                       PTinfo = new MenuData;
                       PTinfo->mw = md->mw;
                       PTinfo->ItemID = PTPDOK;
                       md->mw->PTwininfo->Append(PTinfo); 
                       XtAddCallback(PTPromptDialog,XmNokCallback,PTPromptCB,XtPointer(PTinfo));
                       PTinfo = new MenuData;
                       PTinfo->mw = md->mw;
                       PTinfo->ItemID = PTPDCANCEL;
                       md->mw->PTwininfo->Append(PTinfo);
                       XtAddCallback(PTPromptDialog,XmNcancelCallback,PTPromptCB,
                                                                         XtPointer(PTinfo));
                       XtUnmanageChild(XmSelectionBoxGetChild(PTPromptDialog, XmDIALOG_HELP_BUTTON));
                       XtManageChild(PTPromptDialog);
                       XmStringFree(label_str);
                       XmStringFree(value_str);
                       XmStringFree(title_str);
                       break;
   case PTCLOSE      : XtUnmanageChild(md->mw->thePTdialogshell);  // Schliesen des Strukturbaumfensters
                       break;
   case PTROOT       : PTCreateRootTree(md->mw);   // Root des Strukturbaums anzeigen
                       break;
   default : assert(0);
   };
 }
 
static void MenuCB(Widget w, XtPointer client_data, XtPointer call_data)
	// Ein MenuItem wurde ausgewaehlt
 {
 MenuData* md = (MenuData*) client_data;
 ModWinListElem* mw;
 int ok;
 Arg al[10];
 int ac;
 Widget PromptDialog;
 XmString label_str,value_str,title_str;
 char value[10];

 switch (md->ItemID)	// MenuItem-Kennung
   {
   case WMMAXMODWINS:	// oeffne Prompt Dialog fuer max. offene Fenster
     sprintf(value,"%d",maxopenwindows);
     value_str = XmStringCreateSimple(value);
     label_str = XmStringCreateSimple("Max. number of module windows.");
     title_str = XmStringCreateSimple("Max. Windows");
     ac=0;
     XtSetArg(al[ac],XmNselectionLabelString,label_str); ac++;
     XtSetArg(al[ac],XmNtextString,value_str); ac++;
     XtSetArg(al[ac],XmNautoUnmanage,False); ac++;
     XtSetArg(al[ac],XmNdialogTitle,title_str); ac++;
     XtSetArg(al[ac],XmNdialogStyle,XmDIALOG_FULL_APPLICATION_MODAL);ac++;
     PromptDialog = XmCreatePromptDialog(md->mw->theshell,"prompt",al,ac);
     XtAddCallback(PromptDialog,XmNokCallback,PromptCB,XtPointer(PDOK));
     XtAddCallback(PromptDialog,XmNcancelCallback,PromptCB,XtPointer(PDCANCEL));
     XtUnmanageChild(XmSelectionBoxGetChild(PromptDialog,XmDIALOG_HELP_BUTTON));
     XtManageChild(PromptDialog);
     XmStringFree(label_str);
     XmStringFree(value_str);
     XmStringFree(title_str);
     break;
   case WMCLOSE:	// schliesse Fenster
     mw = (ModWinListElem*) ModWinList->GetFirstElem();
     while (mw && (mw!=md->mw))
       {
       mw = (ModWinListElem*) ModWinList->GetNextElem();
       }
     assert(mw);
     ModWinList->DelCurrentElem();
     DeleteHighlights();
     ClearNavWin();	
     break;
   case GMGOTOTOP:	// Zeige TopWindow an
     XRaiseWindow(XtDisplay(thetoplevel),XtWindow(thetoplevel));
     XtPopup(thetoplevel,XtGrabNone);
     break;
   case GMBOOKMARKS:	// Zeige Lesezeichenfenster an
     DisplayBookmarkWin();
     break;
   case LMCONTENTS:	// Zeige Inhaltsverzeichnis an
     ok = ReadContents(md->mw,md->mw->name);
     break;
   case LMERROR:	// Zeige Fehler-Liste an
     ok = ReadErrors(md->mw,md->mw->name);
     break;
   case LMTYPING:      // Zeige Typ-Warnungs-Liste an
     ok = ReadWarnings(md->mw,md->mw->name);
     break;
   case PMDISPLAY:    // Zeige Parsetree-Graph Window
     if (! XtIsManaged(md->mw->thePTdialogshell)) 
        {
          XtManageChild(md->mw->thePTdialogshell);
          XClearArea(XtDisplay(md->mw->thePTcanvas),XtWindow(md->mw->thePTcanvas),0,0,0,0,1);
                        // expose event wird erzeugt
        } 
     break;
   default: assert(0);
   };
 }

static void TextCB(Widget w, XtPointer client_data, XtPointer call_data)
	// Eine Position im Text-Widget wurd angeklickt
 {
 XmTextVerifyCallbackStruct *cbp = (XmTextVerifyCallbackStruct*) call_data;
 ModWinListElem *mw = (ModWinListElem*) client_data;

 XmTextPosition pos,maxpos;
 XmString LinkName_str,LinkText_str;
 Link *l;
 char *DestName,*LinkText;
 Arg al[10];
 int ac;

 if (!mw->lt)
   {
   return;      // Wenn LinkFile noch nicht gelesen
   }
 maxpos = XmTextGetLastPosition(mw->thetext); // Cursor-park-Position 
 pos    = cbp->newInsert;		      // klick Position
 if (pos == maxpos)
   {
   return;				// Cursor wurde geparkt
   }
 touch++;				// Fenster beruehrt
 mw->touched=touch;
 DeleteHighlights();
 l = mw->lt->GetLink(pos);	// Link suchen
 if (!l)
   {
   ClearNavWin();
   return;			// kein Link
   }
 else				// Highlight und Link in NavWin anzeigen
   {
   l->GetSource(&(mw->SourcePos),&(mw->SourceLength));
   XmTextSetHighlight(mw->thetext,mw->SourcePos,
                                  mw->SourcePos+mw->SourceLength,
                                  XmHIGHLIGHT_SELECTED);
   DisplayLink(l,mw->name);
   }
 }

static void ListCB(Widget w, XtPointer client_data, XtPointer call_data)
	// Zeile in List-Widget wurde gewaehlt
 {
 XmListCallbackStruct *cbp = (XmListCallbackStruct*) call_data;
 ModWinListElem *mw        = (ModWinListElem*) client_data;

 Arg al[10];
 int ac;
 int item,top;
 Link *l;
 XmTextPosition pos,maxpos;

 item = cbp->item_position-1;
 XmListDeselectPos(mw->thelist,item+1);		  // unhighlight
 while ((item>=0) && (mw->ListDestPos[item]==-1)) // Kopf von evtl. 
   {						  // mehrzeiligem Element
   item--;					  // suchen
   }
 assert(item>=0);
 XmListSelectPos(mw->thelist,item+1,False);
 ac=0;						  // evtl. Kopf in sichtbaren
 XtSetArg(al[ac],XmNtopItemPosition,&top); ac++;  // Bereich scrollen
 XtGetValues(mw->thelist,al,ac);
 if (item+1 < top)
   {
   XmListSetPos(mw->thelist,item+1);
   }
 maxpos = XmTextGetLastPosition(mw->thetext); 
 pos    = mw->ListDestPos[item];
 l = mw->lt->GetLink(pos);			// Link suchen
 if (l)						// Wenn Link existiert
   {
   XmTextShowPosition(mw->thetext,pos);	          // evtl. scrollen
   XmTextSetInsertionPosition(mw->thetext,pos);	  // TextCB triggern
   XmTextSetInsertionPosition(mw->thetext,maxpos);// Cursor wegraeumen
   }
 else
   {
   touch++;		// Fenster wurde beruehrt
   mw->touched=touch;
   DeleteHighlights();	// unhighlight und NavWin leeren
   ClearNavWin();
   mw->SourcePos    = pos;		// neues Highlight
   mw->SourceLength = mw->ListDestLen[item];
   XmTextSetHighlight(mw->thetext,mw->SourcePos,
                                  mw->SourcePos+mw->SourceLength,
                                  XmHIGHLIGHT_SELECTED);
   XmTextShowPosition(mw->thetext,pos);	          // evtl. scrollen
   XmTextSetInsertionPosition(mw->thetext,maxpos);// Cursor wegraeumen
   }
 }

// Hilfsfunktionen zur Window Generierung
//---------------------------------------

static Widget CreateMenu(Widget menubar, char* name, char mnemonic)
	// Menu mit Name und Abkuerzung generieren
 {
 Arg al[10];
 int ac;
 Widget   menu,button;
 XmString name_str;

 ac=0;					// menupane erzeugen
 menu = XmCreatePulldownMenu(menubar,name,al,ac);
 name_str = XmStringCreateSimple(name);
 ac=0;					// Button fuer Menu
 XtSetArg(al[ac],XmNsubMenuId,menu); ac++;
 XtSetArg(al[ac],XmNlabelString,name_str); ac++;
 XtSetArg(al[ac],XmNmnemonic,mnemonic); ac++;
 button = XmCreateCascadeButton(menubar,name,al,ac);
 XtManageChild(button);
 XmStringFree(name_str);
 return(menu);
 }

static void CreateMenuItem(Widget menupane, char* name, char mnemonic,
                           int ItemID, ModWinListElem* mw)
	// MenuItem mit Name, Abkuerzung und client_data fuer MenuCB erzeugen
 {
 Arg al[10];
 int ac;
 Widget   item;
 XmString name_str;

 MenuData *md = new MenuData;	// client_data fuer MenuCB()
 md->ItemID=ItemID;		// MenuItem-Kennung
 md->mw    =mw;			// Fenster-Daten
 mw->mdList->Append(md);	// md merken, um es spaeter loeschen zu koennen
 
 name_str = XmStringCreateSimple(name);	// MenuItem erzeugen
 ac=0;
 XtSetArg(al[ac],XmNlabelString,name_str); ac++;
 XtSetArg(al[ac],XmNmnemonic,mnemonic); ac++;
 item = XmCreatePushButton(menupane,name,al,ac);
 XtManageChild(item);
 XtAddCallback(item,XmNactivateCallback,MenuCB,XtPointer(md));
 XmStringFree(name_str);
 }

static void PTCreateGC(ModWinListElem* mw)
  // Graphic Context fuer die Parsetree Anzeigefenster erzeugen
  // (wird nur nach Generierung  des ersten Modulfensters aufgerufen)
 {

 Arg al[10];
 int ac;
 Pixel fg, bg;
 XGCValues vals;

 ac = 0;
 XtSetArg(al[ac],XmNforeground,&fg); ac++;
 XtSetArg(al[ac],XmNbackground,&bg); ac++;
 XtGetValues(mw->thePTcanvas, al,ac);
 vals.foreground = fg;
 vals.background = bg;
 // Durchgezogene Linie fuer Kanten zu Nichtterminal-Symbolen
 PTGCcopy = XtGetGC(mw->thePTcanvas, GCForeground | GCBackground, &vals);

 // Gestichelte Linie fuer Kanten zu Terminal-Symbolen
 vals.foreground = fg;
 vals.background = bg;
 vals.line_style = LineOnOffDash; 
 PTGCdash = XtGetGC(mw->thePTcanvas, GCForeground | GCBackground | GCLineStyle, &vals);
 }

static void PTInitFontConstants(ModWinListElem* mw)
 // Fonthoehe und -breite merken
 // (wird nur nach erstmaliger Modulfenstergenerierung aufgerufen)
 {

  Arg al[10];
  int ac;
  XmString test_str;
  XmFontList fontlist;
  // Fontdaten aus Listenfenster des Modulewindows ermitteln
  ac = 0;
  XtSetArg(al[ac], XmNbuttonFontList, &fontlist);ac++;
  XtGetValues(mw->thePTdialogshell, al,ac);
  test_str =  XmStringCreateSimple("M");
  PTFontHeight = XmStringHeight(fontlist, test_str);
  PTFontWidth =  XmStringWidth(fontlist, test_str);
  XmStringFree(test_str);
  PTHalfFontHeight = PTFontHeight / 2;
 }
      

// Ein Parsetree-Anzeige Window erzeugen
//---------------------------------------

static void CreatePTWindow(ModWinListElem* mw, char* name)

 {

 Arg al[10]; 
 int ac;
 Widget commands, frame, scrollwin, closebutton, maxlevelbutton,
        rootbutton;
 XmString name_str;
 MenuData* PTinfo;

 char* wintitle = new char[strlen(name) + strlen("Parsetree : ") + 1];

 sprintf(wintitle,"Parsetree : %s",name);
 
 ac = 0;
 XtSetArg(al[ac], XmNheight, PTDIALOGHEIGHT);ac++;
 XtSetArg(al[ac], XmNwidth , PTDIALOGWIDTH);ac++;
 XtSetArg(al[ac], XmNtitle, wintitle);ac++;
 XtSetArg(al[ac], XmNresizePolicy, XmRESIZE_NONE);ac++;

// FormDialog
 mw->thePTdialogshell = XmCreateFormDialog(mw->theshell,"PTdialog",al,ac);

 delete wintitle;
// FormDialog/RowColumn
 ac = 0;
 XtSetArg(al[ac], XmNtopAttachment, XmATTACH_FORM); ac++;
 XtSetArg(al[ac], XmNleftAttachment, XmATTACH_FORM); ac++;
 commands = XmCreateRowColumn(mw->thePTdialogshell,"commands",al,ac);
 XtManageChild(commands);


// RowColumn/PushButton

 name_str = XmStringCreateSimple("Root");
 ac = 0;
 XtSetArg(al[ac] , XmNlabelString , name_str); ac++;
 XtSetArg(al[ac] , XmNmnemonic,'R'); ac++;
 rootbutton = XmCreatePushButton(commands, "rootbutton" , al, ac);
 PTinfo = new MenuData;
 PTinfo->mw = mw;
 PTinfo->ItemID = PTROOT;
 mw->PTwininfo->Append(PTinfo);
 XtAddCallback(rootbutton, XmNactivateCallback, PTMenuCB, XtPointer(PTinfo));
 XtManageChild(rootbutton);
 XmStringFree(name_str);

// RowColumn/PushButton
 name_str = XmStringCreateSimple("Max.Levels");
 ac = 0;
 XtSetArg(al[ac] , XmNlabelString , name_str); ac++;
 XtSetArg(al[ac] , XmNmnemonic,'L'); ac++;
 maxlevelbutton = XmCreatePushButton(commands, "maxlevelbutton",al,ac); 
 PTinfo = new MenuData;
 PTinfo->mw = mw;
 PTinfo->ItemID = PTNODELEVELS;
 mw->PTwininfo->Append(PTinfo);
 XtAddCallback(maxlevelbutton, XmNactivateCallback, PTMenuCB, XtPointer(PTinfo));
 XtManageChild(maxlevelbutton);
 XmStringFree(name_str);

// RowColumn/PushButton
 name_str = XmStringCreateSimple("Close");
 ac = 0;
 XtSetArg(al[ac] , XmNlabelString , name_str); ac++;
 XtSetArg(al[ac] , XmNmnemonic,'C');ac++;
 closebutton = XmCreatePushButton(commands, "closebutton",al,ac);
 PTinfo = new MenuData;
 PTinfo->mw = mw;
 PTinfo->ItemID = PTCLOSE;
 mw->PTwininfo->Append(PTinfo);
 XtAddCallback(closebutton,  XmNactivateCallback, PTMenuCB, XtPointer(PTinfo));
 XtManageChild(closebutton);
 XmStringFree(name_str);

// FormDialog/ScrolledWindow
 ac = 0;
 XtSetArg(al[ac], XmNtopAttachment, XmATTACH_FORM);ac++;
 XtSetArg(al[ac], XmNbottomAttachment,XmATTACH_FORM);ac++;
 XtSetArg(al[ac], XmNleftAttachment, XmATTACH_WIDGET);ac++;
 XtSetArg(al[ac], XmNleftWidget, commands);ac++;
 XtSetArg(al[ac], XmNrightAttachment, XmATTACH_FORM); ac++;
 XtSetArg(al[ac], XmNscrollBarDisplayPolicy,XmSTATIC); ac++;
 scrollwin = XmCreateScrolledWindow(mw->thePTdialogshell,"scrollwin",al,ac);
 XtManageChild(scrollwin);

// ScrolledWindow/ScrollBar
 ac = 0;
 XtSetArg(al[ac], XmNorientation, XmHORIZONTAL);ac++;
 XtSetArg(al[ac], XmNmaximum, PTCANVASWIDTH);ac++;
 XtSetArg(al[ac], XmNsliderSize, PTCANVASWIDTH); ac++;
 XtSetArg(al[ac], XmNpageIncrement , PTCANVASWIDTH / 2); ac++;
 mw->PTScrollBarHor = XmCreateScrollBar(scrollwin,"PTScrollBarHor",al, ac);
 XtManageChild(mw->PTScrollBarHor);
 PTinfo = new MenuData;
 PTinfo->mw = mw;
 PTinfo->ItemID = PTSBHOR;
 mw->PTwininfo->Append(PTinfo);
 XtAddCallback(mw->PTScrollBarHor,XmNvalueChangedCallback, PTScrollCB,
                                                        XtPointer(PTinfo));
// ScrolledWindow/ScrollBar
 ac = 0;
 XtSetArg(al[ac], XmNorientation, XmVERTICAL); ac++;
 XtSetArg(al[ac], XmNmaximum, PTCANVASHEIGHT); ac++;
 XtSetArg(al[ac], XmNsliderSize, PTCANVASHEIGHT);ac++;
 XtSetArg(al[ac], XmNpageIncrement , PTCANVASHEIGHT / 2); ac++;
 mw->PTScrollBarVer = XmCreateScrollBar(scrollwin,"PTScrollBarVer", al ,ac);
 XtManageChild(mw->PTScrollBarVer);
 PTinfo = new MenuData;
 PTinfo->mw = mw;
 PTinfo->ItemID = PTSBVER;
 mw->PTwininfo->Append(PTinfo);
 XtAddCallback(mw->PTScrollBarVer,XmNvalueChangedCallback, PTScrollCB,
                                                        XtPointer(PTinfo));
// ScrolledWindow/Frame
 ac = 0;
 XtSetArg(al[ac], XmNshadowType , XmSHADOW_IN); ac++;
 frame = XmCreateFrame(scrollwin, "frame" ,al, ac);
 XtManageChild(frame);

// Frame/DrawingArea
 ac = 0;
 XtSetArg(al[ac], XmNheight, PTCANVASHEIGHT); ac++;
 XtSetArg(al[ac], XmNwidth, PTCANVASWIDTH); ac++;
 mw->thePTcanvas = XmCreateDrawingArea(frame, "thePTcanvas", al, ac);
 XtManageChild(mw->thePTcanvas);
 PTinfo = new MenuData;
 PTinfo->mw = mw;
 PTinfo->ItemID = 0;
 mw->PTwininfo->Append(PTinfo);
 XtAddCallback(PTinfo->mw->thePTcanvas, XmNexposeCallback, PTExposeCB, XtPointer(PTinfo));
 PTinfo = new MenuData;
 PTinfo->mw = mw;
 PTinfo->ItemID = 0;
 mw->PTwininfo->Append(PTinfo);
 XtAddCallback(PTinfo->mw->thePTcanvas, XmNresizeCallback, PTResizeCB, XtPointer(PTinfo));



 } 

 

// Ein Modul-Window erzeugen
//--------------------------

static ModWinListElem* CreateModWin(Widget toplevel, char* name)
	// Neues Fenster
 {
 Arg  al[10];
 int  ac;
 Widget form,menubar,menu,pane;
 ModWinListElem *mw = new ModWinListElem;

 mw->name = new char[strlen(name)+1];
 strcpy(mw->name,name);

 // neue Shell fuer neues Window

 char *appname=new char[strlen(name)+strlen("tTLA+ Browser : ")+1];
 sprintf(appname,"tTLA+ Browser : %s",name);
 ac = 0;
 XtSetArg(al[ac],XmNy,ModWinList->GetLength()*80); ac++;
 mw->theshell = XtAppCreateShell(appname,"modwinshell",
                                 applicationShellWidgetClass,
                                 XtDisplay(toplevel),al,ac);
 delete appname;

 // shell->form(1/1)  (i/j) = i-ter von j Soehnen

 ac=0;
 form = XmCreateForm(mw->theshell,"form",al,ac);
 XtManageChild(form);

 // shell->form->menubar(1/2)
 // plus menuitems

 ac=0;
 XtSetArg(al[ac],XmNtopAttachment,XmATTACH_FORM); ac++;
 XtSetArg(al[ac],XmNleftAttachment,XmATTACH_FORM); ac++;
 XtSetArg(al[ac],XmNrightAttachment,XmATTACH_FORM); ac++;
 menubar = XmCreateMenuBar(form,"menubar",al,ac);
 XtManageChild(menubar);
 menu = CreateMenu(menubar,"Window",'W');
   CreateMenuItem(menu,"Max. Module Windows",'M',WMMAXMODWINS,mw);
   CreateMenuItem(menu,"Close",'C',WMCLOSE,mw);
 menu = CreateMenu(menubar,"Goto",'G');
   CreateMenuItem(menu,"Top Window",'T',GMGOTOTOP,mw);
   CreateMenuItem(menu,"Bookmarks",'B',GMBOOKMARKS,mw);
 menu = CreateMenu(menubar,"List",'L');
   CreateMenuItem(menu,"Display Contents",'C',LMCONTENTS,mw);
   CreateMenuItem(menu,"Display Errors",'E',LMERROR,mw);
   CreateMenuItem(menu,"Display Typing",'T',LMTYPING,mw);   // Typfehler-Anzeige

 menu = CreateMenu(menubar,"Parsetree",'P');
   CreateMenuItem(menu,"Display Parsetree",'D',PMDISPLAY, mw);  // Strukturbaum-Anzeige
 
 // shell->form->panedwindow(2/2)

 ac=0;
 XtSetArg(al[ac],XmNtopAttachment,XmATTACH_WIDGET);  ac++;
 XtSetArg(al[ac],XmNtopWidget,menubar); 	     ac++;
 XtSetArg(al[ac],XmNleftAttachment,XmATTACH_FORM);   ac++;
 XtSetArg(al[ac],XmNrightAttachment,XmATTACH_FORM);  ac++;
 XtSetArg(al[ac],XmNbottomAttachment,XmATTACH_FORM); ac++;
 pane = XmCreatePanedWindow(form,"pane",al,ac);
 XtManageChild(pane);

 // shell->form->panedwindow->list(1/2)

 ac=0;
 XtSetArg(al[ac],XmNvisibleItemCount,LISTVISIBLEITEMS); ac++;
 XtSetArg(al[ac],XmNselectionPolicy,XmSINGLE_SELECT); ac++;
 XtSetArg(al[ac],XmNlistSizePolicy,XmRESIZE_IF_POSSIBLE); ac++;
 mw->thelist = XmCreateScrolledList(pane,"modwinlist",al,ac);
 XtManageChild(mw->thelist);
 XtAddCallback(mw->thelist,XmNsingleSelectionCallback,ListCB,XtPointer(mw));
 
 // shell->form->panedwindow->scrolledtext(2/2)

 ac=0;
 XtSetArg(al[ac],XmNcolumns,80); 			ac++;
 XtSetArg(al[ac],XmNrows,20);    			ac++;
 XtSetArg(al[ac],XmNautoShowCursorPosition,False); 	ac++;
 XtSetArg(al[ac],XmNcursorPositionVisible,False); 	ac++;
 XtSetArg(al[ac],XmNeditable,False); 			ac++;
 XtSetArg(al[ac],XmNeditMode,XmMULTI_LINE_EDIT);	ac++;
 mw->thetext = XmCreateScrolledText(pane,"modwintext",al,ac);
 XtManageChild(mw->thetext);
 XtAddCallback(mw->thetext,XmNmotionVerifyCallback,TextCB,XtPointer(mw));

// Strukturbaum-Fenster erzeugen
 CreatePTWindow(mw,name); 

 // shell->MessageDialog

 XtRealizeWidget(mw->theshell);

 return mw;
 }

static ModWinListElem* ReCreateModWin(ModWinListElem* mw, char* name)
	// Inhalt eines alten Fensters austauschen
 {
 Arg  al[10];
 int  ac;
 XmString name_str;
 delete mw->name;
 mw->name = new char[strlen(name)+1];
 strcpy(mw->name,name);

 char *appname=new char[strlen(name)+strlen("tTLA+ Browser : ")+1];
 sprintf(appname,"tTLA+ Browser : %s",name);
 ac=0;
 XtSetArg(al[ac],XmNtitle,appname); ac++;
 XtSetValues(mw->theshell,al,ac);
 delete appname;
// Strukturbaumfenster verschnwinden lassen 
// und neue Titelzeile erzeugen
 if(XtIsManaged(mw->thePTdialogshell))
  XtUnmanageChild(mw->thePTdialogshell);
 char *wintitle = new char[strlen(name)+strlen("Parsetree : ") + 1];
 sprintf(wintitle,"Parsetree : %s",name);
 name_str = XmStringCreateSimple(wintitle);
 ac = 0;
 XtSetArg(al[ac],XmNdialogTitle,name_str);ac++;
 XtSetValues(mw->thePTdialogshell,al,ac);
 XmStringFree(name_str);
 delete wintitle;

 delete mw->lt;
 mw->lt=0;

 mw->SourcePos    = 0;
 mw->SourceLength = 0;

 return mw;
 }

// File in TextWidget einlesen
//----------------------------

static Boolean ReadFile(Widget text, char* dir, char* name, char* suffix)
 {
 char *filename,*libdir;
 char *filetext;
 FILE *fp;
 int  filelength,readchars;
 struct stat stat_val;

 filename = new char[strlen(dir)+strlen(name)+strlen(suffix)+3];
 sprintf(filename,"%s/%s.%s",dir,name,suffix);
 if (stat(filename,&stat_val)!=0)
   { 
     delete filename;
     libdir = getenv("TLALIB");
     if(libdir)
      { 
        filename = new char[strlen(libdir)+strlen(name)+strlen(suffix)+3];
        sprintf(filename,"%s/%s.%s",libdir,name,suffix);
        if(stat(filename,&stat_val)!=0)
         {
           delete filename;
           return False;
         }
      }
     else
      {
        return False;
      } 
   }
 filelength = stat_val.st_size;
 fp = fopen(filename,"r");
 delete filename;
 if (!fp)
   {
   return False;
   }
 filetext = new char[filelength+2];
 readchars = fread(filetext,sizeof(char),filelength,fp);
 filetext[readchars]   = ' ';
 filetext[readchars+1] = '\0';
 XmTextSetTopCharacter(text,0);
 XmTextSetString(text,filetext);
 XmTextSetHighlight(text,0,filelength-2,XmHIGHLIGHT_NORMAL);
 delete filetext;
 fclose(fp);
 return filelength==readchars;
 }

// Exporte
//--------

/*---------------------------------------------------------------------------*
 * Modul ModWin ueber neue Spezifikation informieren
 *---------------------------------------------------------------------------*/

void InitModWins(Widget toplevel, char* dir, char* suffix)
 {
 thetoplevel = toplevel;	// toplevel-widget merken
 if (!thedir)			// directory der Modul-Quelltexte merken
   {
   thedir = new char[strlen(dir)+1];
   strcpy(thedir,dir);
   }
 if (!thesuffix)		// suffix der Modul-Quelltexte merken
   {
   thesuffix = new char[strlen(suffix)+1];
   strcpy(thesuffix,suffix);
   }
 }

/*---------------------------------------------------------------------------*
 * Ein ModulWindow anzeigen (evtl. neu generieren)
 *---------------------------------------------------------------------------*/

Boolean OpenModWin(char* name)
 {
 Arg al[10];
 int ac;
 int ok;
 int openwindows;
 LinkTable* lt;
 ModWinListElem *mw, *mwMin;

				// pruefe, ob Window schon geoeffnet, suche
				// dabei aeltestes Modul
 mwMin = mw = (ModWinListElem*) ModWinList->GetFirstElem();
 while (mw)
   {
   if (strcmp(name,mw->name)==0)
     {
     break;
     }
   if (mw->touched < mwMin->touched)
     {
     mwMin = mw;
     }
   mw = (ModWinListElem*) ModWinList->GetNextElem();
   }
 if (!mw)		// Modul noch nicht angezeigt
   {
   openwindows=ModWinList->GetLength();
   if (openwindows<maxopenwindows)	// neues Fenster oeffnen
     {
     mw = CreateModWin(thetoplevel,name);
     if(openwindows == 0)    // falls Modwin erstes Fenster ist ...
      { 

       PTInitFontConstants(mw);  // Fontdaten initialisieren
       PTCreateGC(mw);           // GC fuer Parsetree-Anzeige initialisieren

      }
     
     ModWinList->Append(mw);
     }
   else			// aeltestes Modul loeschen
     {
     mw = ReCreateModWin(mwMin,name);
     DeleteHighlights();
     ClearNavWin();
     XRaiseWindow(XtDisplay(mw->theshell),XtWindow(mw->theshell));
     XtPopup(mw->theshell,XtGrabNone);
     }
   lt = new LinkTable;	// Fenster mit (neuem) Modul fuellen
   ok = ReadFile(mw->thetext,thedir,name,thesuffix);
   ok = ReadContents(mw,name) && ok;

   InitPTGraph(mw,name);

   ok = lt->ReadLinkFile(name) && ok;
   mw->lt=lt;
   touch++;		// Fenster wurde beruehrt
   mw->touched=touch;
   return ok;
   }
 else		// Modul wird bereits angezeigt
   {

   XRaiseWindow(XtDisplay(mw->theshell),XtWindow(mw->theshell));
   XtPopup(mw->theshell,XtGrabNone);
   touch++;
   mw->touched=touch;
   return(True);
   }
 }

/*---------------------------------------------------------------------------*
 * Modul ModName anzeigen und Cursor zu Position pos
 *---------------------------------------------------------------------------*/

void BeamToPosAtMod(long pos, char* ModName)
 {
 ModWinListElem *Destmw;
 XmTextPosition maxpos;
 int ok;

 ok = OpenModWin(ModName);	// Fenster fuer Modul oeffnen
 if (!ok)
   {
   DisplayMsg("Can't open modul window correctly.");
   return;
   }
 Destmw = (ModWinListElem*) ModWinList->GetFirstElem();	// Fensterdaten suchen
 while (Destmw)
   {
   if (strcmp(ModName,Destmw->name)==0)
     {
     break;
     }
   Destmw = (ModWinListElem*) ModWinList->GetNextElem();
   }
 assert(Destmw);
 XmTextShowPosition(Destmw->thetext,pos);	    // evtl. scrollen
 XmTextSetInsertionPosition(Destmw->thetext,pos);   // TextCB triggern
 maxpos = XmTextGetLastPosition(Destmw->thetext); 
 XmTextSetInsertionPosition(Destmw->thetext,maxpos);// Cursor wegraeumen
 }


/*---------------------------------------------------------------------------*
 * Alle ModulWindows loeschen
 *---------------------------------------------------------------------------*/

void DestroyModWins()
 {
 delete ModWinList;
 ModWinList = new List;
 if (thedir)    { delete thedir;    thedir=0;    }
 if (thesuffix) { delete thesuffix; thesuffix=0; }
 }
