/*---------------------------------------------------------------------------*
 * SemConWin.C
 *
 * Olaf Meier, 01.09.94
 *
 * Generierung und Verwaltung vom
 * - Systemvariablenfenster
 * - Prozessauswahlfenster
 * - Konzeptfenster
 * fuer die semantische Konzepterfassung
 *----------------------------------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>

#include <assert.h>

#include "list.h"
#include "pilayout.h"    // Layoutberechnung der Prozess/Interface Darstellung
#include "adjlistcxx.h"  // Adjazenzlisten-Darstellung der Prozess/Interface-Komponenten
                         // und des Variablenbaum     
#include "semconwin.h"
#include "semconhelp.h"  // Hilfroutinen fuer die Konzepterstellung
#include "dbinterface.h"


#include <Xm/Xm.h>
#include <Xm/CascadeB.h>
#include <Xm/FileSB.h>
#include <Xm/Form.h>
#include <Xm/Frame.h>
#include <Xm/DrawingA.h>
#include <Xm/Text.h>
#include <Xm/List.h>
#include <Xm/PanedW.h>
#include <Xm/PushB.h>
#include <Xm/SelectioB.h>
#include <Xm/RowColumn.h>
#include <Xm/ScrollBar.h>
#include <Xm/ScrolledW.h>
#include <X11/Shell.h>

// MenuItem Identifier
//--------------------

#define SCNEW     1
#define SCDELETE  2
#define SCDISPLAY 3
#define SCCLOSE   4

#define PICLOSE   1

#define SBOXDISCARD 1
#define SBOXADD     2
#define SBOXDELETE  3

// ScrollBar Identifier
//---------------------

#define SCSBHOR 1
#define SCSBVER 2

#define PISBHOR 1
#define PISBVER 2 


// Initiale Windowgroesse
//-----------------------


#define SCCANVASWIDTH           400
#define SCCANVASHEIGHT          400
#define SCSHELLWIDTH            500
#define SCSHELLHEIGHT           700
#define SCMSGLISTVISIBLEITEMS   5


#define PICANVASWIDTH    400
#define PICANVASHEIGHT   400
#define PIDIALOGWIDTH    400
#define PIDIALOGHEIGHT   600


// ALIAS - Name fuer einige AdjList Funktionen
//--------------------------------------------

#define SCGETXPOS()             GetPos()
#define SCSETXPOS(x)            SetPos(x)
#define SCGETYPOS()             GetTempPos()
#define SCSETYPOS(y)            SetTempPos(y)
#define SCGETWIDGETNR()         GetLevel()
#define SCSETWIDGETNR(nr)       SetLevel(nr)
#define SCGETNODEOFFSET()       GetCounter()
#define SCSETNODEOFFSET(offset) SetCounter(offset)



// Macros
//-------

#define max(a,b) ((a) > (b)) ? (a) : (b)
#define min(a,b) ((a) < (b)) ? (a) : (b)


      

   

// Modulglobale Variablen
//-----------------------



static List         *ActionVarsList = 0,*ProcessList = 0, *ActProcVars = 0;
               
static char         *filename=0;   
static AdjList      *SCtheGraph=0,*PItheGraph = 0;   
static GC           SCGCcopy,SCGCdash;
static Dimension    SCFontWidth,SCFontHeight,SCHalfFontHeight;
static int          SCGraphWidth=0, PIGraphWidth = 0;
static int          SCGraphHeight=0, PIGraphHeight = 0; 
static int          SCx_offset,SCy_offset, PIx_offset,PIy_offset;  // Offset fuer ScrollBar Darstellung
static Widget       *SCWidgetArray=0,*PIWidgetArray = 0;     // Buttons fuer Knoten im Graph
static Widget       *PIListArray = 0;    // List Widgets pro PI Button zur Anzeige der zugeordnenten Variablen
static int          SCWidgetArrayLength,PIWidgetArrayLength;  // Anzahl Knoten

static Widget       SCScrollBarHor, // global bekannte Widgets
                    SCScrollBarVer,
                    PIScrollBarHor,
                    PIScrollBarVer, 
                    SCcanvas,PIcanvas, SCMsgList,SCshell = 0, PIdialogshell = 0,
                    PIselectionBox = 0 ;








     
       

/*---------------------------------------------------------------------------*
 * Meldung im Message-Window anzeigen
 *---------------------------------------------------------------------------*/

void SCDisplayMsg(char* msg)
 {
 XmString msg_str;

 msg_str = XmStringCreateSimple(msg);
 XmListAddItem(SCMsgList,msg_str,0);
 XmStringFree(msg_str);
 XmListSetBottomPos(SCMsgList,0);
 XFlush(XtDisplay(SCMsgList));
 }

// Aktivierung bzw. Deaktivierung der angezeigten PushButtons im Systemvariablen-Baum
static void SensitiveCheck(AdjList* adjl)
 {
   assert(adjl);
   int widgetnr;
   Node* n = adjl->GetFirstNode();
   while(n)
    { 
      widgetnr = n->SCGETWIDGETNR();
      if(n->Isselected())  // Knoten ist selektiert
         XtSetSensitive(SCWidgetArray[widgetnr], False);
      else
         XtSetSensitive(SCWidgetArray[widgetnr], True);

      n = adjl->GetNextNode();
    }
 }
         
// Einlesen des Systemvariablen-Baums
static  int SCReadGraph(char* filename)
{
 char *name;
 int  x,y,topx,counter,value,ok,endofnodes,endofedges, endofactions;
 ActionVarsElem* actionel; 
 char t;
 Node *n,*m,*varnode;
 Arg  al[10];
 int  actionnr,ac;
 XmString  name_str;
 Dimension width,height,maxwidth;

 void SCNodeCB(Widget w, XtPointer client_data, XtPointer call_data);

 if (!OpenVarTreeFile(filename, "r", 0, 0))
   {
   
   return 0;
   } 

 SCtheGraph = new AdjList; 

 GetVarTreeExtensions(&SCGraphWidth,&SCGraphHeight);

 SCGraphWidth  = SCGraphWidth  * SCFontWidth; 
 SCGraphHeight = SCGraphHeight * SCFontHeight;

 counter = 0;
 topx    = -1;


 ok = ReadVarNode(&t, &name, &x, &y, &endofnodes);
 while(!endofnodes && ok)      
   {
   n = SCtheGraph->SCAddNode(name,0);         // Knoten in AdjList eintragen
   n->SetNodeType(t);
   n->SCSETXPOS(x * SCFontWidth);           // setze Knoten Position (Mittelpunkt)
   n->SCSETYPOS(y * SCFontHeight);
   counter++;         
   if (topx==-1)                        // erster Knoten ist TopModul
     {
     topx = n->SCGETXPOS();
     }
   ok = ok && ReadVarAction(&actionnr, &endofactions);
   while(!endofactions && ok)
     {

     n->AddActionNr(actionnr);  // korrespondierende Aktion-Nummer dem aktuellen Knoten hinzufuegen 
     ok = ok && ReadVarAction(&actionnr, &endofactions);
     }
   ok = ok && ReadVarEdge(&t, &name, &endofedges);
   while(!endofedges && ok)             // lese alle ausgehenden Kanten
     {

     m = SCtheGraph->SCAddNode(name,0);       // Knoten eintragen
     n->PTAddEdge(m,t);                   // Kante eintragen
     ok = ok && ReadVarEdge(&t, &name, &endofedges);
     }

   ok = ok && ReadVarNode(&t, &name, &x, &y, &endofnodes);
   }

 CloseVarTreeFile();
 if (! ok)                              // Wenn Lesefehler
   {
   delete SCtheGraph;
   SCtheGraph=0; SCGraphWidth=0; SCGraphHeight=0;

   return ok;
   } 

 SCWidgetArray = new Widget[counter];
 SCWidgetArrayLength = counter;   
 counter = 0;
 n = SCtheGraph->GetFirstNode();       
 while(n)
   {
   name = n->GetName();
   name_str = XmStringCreateSimple(name);
   ac=0;
   XtSetArg(al[ac],XmNlabelString,name_str); ac++;
   XtSetArg(al[ac],XmNx,10000); ac++;
   XtSetArg(al[ac],XmNy,10000); ac++;
   SCWidgetArray[counter] = XmCreatePushButton(SCcanvas,"name",al,ac);
   XtManageChild(SCWidgetArray[counter]);
   XtAddCallback(SCWidgetArray[counter],XmNactivateCallback,SCNodeCB,name);
   n->SCSETWIDGETNR(counter);
   counter++;
   XmStringFree(name_str);
   n->SCSETNODEOFFSET(SCHalfFontHeight);  // Abstand von Kantenansatzpunkten
   n = SCtheGraph->GetNextNode();

   }

 ac=0;
 XtSetArg(al[ac],XmNheight,&height); ac++;      // Zeichenflaeche und
 XtSetArg(al[ac],XmNwidth,&width); ac++;        // Scrollbars init
 XtGetValues(SCcanvas,al,ac);
 SCy_offset = 0;
 if (width>SCGraphWidth)
   SCx_offset = (SCGraphWidth-width)/2;
 else
   SCx_offset = topx-width/2;
 maxwidth=max(SCGraphWidth,width);
 value=min(SCx_offset,maxwidth-width);
 value=max(0,value);
 ac=0;                                  // ScrollBars auslegen
 XtSetArg(al[ac],XmNmaximum,maxwidth); ac++;
 XtSetArg(al[ac],XmNsliderSize,width); ac++;
 XtSetArg(al[ac],XmNpageIncrement,max(1,width/2)); ac++;
 XtSetArg(al[ac],XmNvalue,value); ac++;
 XtSetValues(SCScrollBarHor,al,ac);
 ac=0;
 XtSetArg(al[ac],XmNmaximum,max(SCGraphHeight,height)); ac++;
 XtSetArg(al[ac],XmNsliderSize,height); ac++;
 XtSetArg(al[ac],XmNpageIncrement,max(1,height/2)); ac++;
 XtSetArg(al[ac],XmNvalue,SCy_offset); ac++;
 XtSetValues(SCScrollBarVer,al,ac);


// Erstellung der Action-Liste mit der jeweiligen Variablenliste


 ActionVarsList = new List;
 ProcessList = new List; 
 varnode = SCtheGraph->GetFirstNode();
 
 while(varnode )
  {
    actionnr = varnode->GetFirstActionNr();
    while(actionnr != -1) 
     {
       if((actionnr > 0) && (varnode->SCGETWIDGETNR() > 0))
        AppendActionVarsElem(actionnr,varnode->SCGETWIDGETNR(), ActionVarsList); 
       actionnr = varnode->GetNextActionNr();
      }
    varnode = SCtheGraph->GetNextNode();
   }

 return 1;
 }

// Erzeugung des Konzeptfensters mit der Prozess/Interface Darstellung

static void CreatePIGraph()
 {

   char *name,*procname,t,msg[50];
   int  topx,counter,value;
   Node *n,*m;
   Arg  al[10];
   int  ac,varid;
   XmString  name_str,varname_str;
   Widget listw;
   Dimension width,height,maxwidth;
   List *InterfaceList, *Var_ProcessesList;
   void PINodeCB(Widget w, XtPointer client_data, XtPointer call_data);
   void SC_PIListCB(Widget w, XtPointer client_data, XtPointer call_data);

   InterfaceList = new List;
   Var_ProcessesList = new List;

// Darstellung Systemvariable --> Prozesskomponenten ( Ermittlung der Interfacekomponenten)
   CollectVar_ProcessesElems(ProcessList, Var_ProcessesList);
// Suche nach Interface-Komponenten
   SearchforInterfaces(Var_ProcessesList, InterfaceList );


// Adjazenzlisten-Darstellung der Sicht     
   PItheGraph = CreateProcInterfAdjList(ProcessList,InterfaceList);

   delete InterfaceList ;
   delete Var_ProcessesList;

   if(PItheGraph != NULL)
    {
// Errechnung des Layouts
      PILayout(PItheGraph,&PIGraphWidth,&PIGraphHeight);

      PIGraphWidth = PIGraphWidth * SCFontWidth;
      PIGraphHeight = PIGraphHeight * SCFontHeight;
      counter = 0;
      topx = -1;
      n = PItheGraph->GetFirstNode();
      while(n)
      { t = n->GetNodeType();
        n->SCSETXPOS(n->SCGETXPOS() * SCFontWidth);
        n->SCSETYPOS(n->SCGETYPOS() * SCFontHeight); 
        if (t != 'D') counter++;  
        if ((topx==-1) && (t != 'D'))                 // erster nicht Dummyknoten ist TopButton
         {
           topx = n->SCGETXPOS();
         }
        n = PItheGraph->GetNextNode();
       }
      PIWidgetArray = new Widget[counter];     // Array fuer Button-Darstellung der
      PIWidgetArrayLength = counter;           // Knoten
      PIListArray = new Widget[counter];       // Array List Widgets zur Aufnahme der Variablennamen 
 
      counter = 0;
      n = PItheGraph->GetFirstNode();          // Buttons fuer Knoten erzeugen
      while(n)
       {
         name = n->GetName();

         if (name[0] != '#')                  // Wenn kein dummy Knoten
           {

             
             name_str = XmStringCreateSimple(name);
             ac=0;
             XtSetArg(al[ac],XmNlabelString,name_str); ac++;
             XtSetArg(al[ac],XmNx,10000); ac++;
             XtSetArg(al[ac],XmNy,10000); ac++;
             PIWidgetArray[counter] = XmCreatePushButton(PIcanvas,"name",al,ac);
             XtManageChild(PIWidgetArray[counter]);
             XtAddCallback(PIWidgetArray[counter],XmNactivateCallback,PINodeCB,XtPointer(counter));

             ac = 0;
             XtSetArg(al[ac],XmNdialogTitle,name_str);ac++;
             XtSetArg(al[ac],XmNdefaultPosition,False);ac++;
             XtSetArg(al[ac],XmNx,0);ac++;
             XtSetArg(al[ac],XmNy,counter * 50);ac++;
             XtSetArg(al[ac],XmNheight,100);ac++;
             XtSetArg(al[ac],XmNwidth,200);ac++;
             PIListArray[counter] = XmCreateFormDialog(PIdialogshell,"dialoglist",al,ac);

             ac=0;
             XtSetArg(al[ac],XmNtopAttachment, XmATTACH_FORM);ac++;
             XtSetArg(al[ac],XmNbottomAttachment, XmATTACH_FORM);ac++;
             XtSetArg(al[ac],XmNleftAttachment, XmATTACH_FORM);ac++;
             XtSetArg(al[ac],XmNrightAttachment, XmATTACH_FORM);ac++;
             XtSetArg(al[ac],XmNvisibleItemCount,5); ac++;
             XtSetArg(al[ac],XmNselectionPolicy,XmSINGLE_SELECT); ac++;
             XtSetArg(al[ac],XmNlistSizePolicy,XmRESIZE_IF_POSSIBLE); ac++;
        
             listw =  XmCreateScrolledList(PIListArray[counter],"varlist",al,ac);
             XtManageChild(listw);
             XtAddCallback(listw,XmNsingleSelectionCallback,SC_PIListCB,NULL);
// Jede Prozess- bzw. Interfacekomponente erhaelt eine Liste der ihr zugeordnenten
// Systemvariablen
             varid = n->GetFirstActionNr();
             while(varid > 0)
              {
                 varname_str = XmStringCreateLtoR(GetNameofNodeNr(varid,SCtheGraph),
                                                  XmSTRING_DEFAULT_CHARSET);
                 XmListAddItem(listw,varname_str,0);
                 XmStringFree(varname_str);
                 varid = n->GetNextActionNr();
              }
             n->SCSETWIDGETNR(counter);
             counter++;
             XmStringFree(name_str);

             n->SCSETNODEOFFSET(SCHalfFontHeight);  // Abstand von Kantenansatzpunkten
                


           }                                  // zu Knotenmittelpunkt
         else // Wenn dummy Knoten
           {
            n->SCSETNODEOFFSET(0);
           }
         n = PItheGraph->GetNextNode();
     }

    ac=0;
    XtSetArg(al[ac],XmNheight,&height); ac++;      // Zeichenflaeche und
    XtSetArg(al[ac],XmNwidth,&width); ac++;        // Scrollbars init
    XtGetValues(PIcanvas,al,ac);
    PIy_offset = 0;
    if (width>PIGraphWidth)
     PIx_offset = (PIGraphWidth-width)/2;
    else
     PIx_offset = topx-width/2;
    maxwidth=max(PIGraphWidth,width);
    value=min(PIx_offset,maxwidth-width);
    value=max(0,value);
    ac=0;                                  // ScrollBars auslegen
    XtSetArg(al[ac],XmNmaximum,maxwidth); ac++;
    XtSetArg(al[ac],XmNsliderSize,width); ac++;
    XtSetArg(al[ac],XmNpageIncrement,max(1,width/2)); ac++;
    XtSetArg(al[ac],XmNvalue,value); ac++;
    XtSetValues(PIScrollBarHor,al,ac);

    ac=0;
    XtSetArg(al[ac],XmNmaximum,max(PIGraphHeight,height)); ac++;
    XtSetArg(al[ac],XmNsliderSize,height); ac++;
    XtSetArg(al[ac],XmNpageIncrement,max(1,height/2)); ac++;
    XtSetArg(al[ac],XmNvalue,PIy_offset); ac++;
    XtSetValues(PIScrollBarVer,al,ac);
    XClearArea(XtDisplay(PIcanvas), XtWindow(PIcanvas),0,0,0,0,1); 
  }
 }

// Callback Routinen

// Menu des Variablenfensters
static void SCMenuCB(Widget w, XtPointer client_data, XtPointer call_data)
 {
  int i,item = (int) client_data;

  switch (item)
   {
     case SCCLOSE: DestroySemConWin();   // Variablenfenster schliessen
                   break;
     case SCNEW : delete ProcessList;     // Neues Konzept erzeugen, d.h. Aufhebung aller Variablen-
                                          // selektionen, Prozessnamen loeschen. 
                  ProcessList = new List;
                  XmListDeleteAllItems(XmSelectionBoxGetChild(PIselectionBox,XmDIALOG_LIST));
                  UnMarkAllVariables(SCtheGraph);
                  for(i = 1 ; i < SCWidgetArrayLength ; i++)
                   {
                     XtSetSensitive(SCWidgetArray[i], True);
                   }
                  XmListDeleteAllItems(SCMsgList);  
                  break;
     
     case SCDELETE : XtManageChild(PIselectionBox); // Loeschung einer Prozesskomponte
                     XtSetSensitive(XmSelectionBoxGetChild(PIselectionBox, XmDIALOG_OK_BUTTON), False);
                     break;
     case SCDISPLAY : XtManageChild(PIdialogshell);  // Anzeigen der Konzeptsicht
                      CreatePIGraph();
                      break;

     default : assert(0);
               break;
   }
 }

static void SC_PIListCB(Widget w, XtPointer client_data, XtPointer call_data)
// Eine Variable in der Prozess-Variablenliste wurde angewaehlt
 {
   XmListCallbackStruct *cbp = (XmListCallbackStruct*) call_data;
   XmListDeselectPos(w,cbp->item_position); // highlight abschalten
                                                // (sonst keine Funktion)
 }
  
// Menu des Konzeptfensters
static void PIMenuCB(Widget w, XtPointer client_data, XtPointer call_data)
 {
  int item = (int) client_data;
  switch (item)
   {
    case PICLOSE:  XtUnmanageChild(PIdialogshell);
                   if(PItheGraph != NULL)
                    {
                      for(int i = 0; i < PIWidgetArrayLength; i++)
                       {
                         XtDestroyWidget(PIWidgetArray[i]);
                       }
                      delete [] PIWidgetArray;
                      delete [] PIListArray;
                      PIListArray = 0;
                      PIWidgetArray = 0;
                      PIGraphHeight = 0;
                      PIGraphWidth = 0;
                      PIWidgetArrayLength = 0;     
                      delete PItheGraph;
                      PItheGraph = 0;
                     }
                    break;
    default : assert(0);
   }
 }


// Menue des Prozessfensters
static void ProcessCB(Widget w, XtPointer client_data, XtPointer call_data)
 {
  Widget list = XmSelectionBoxGetChild(PIselectionBox, XmDIALOG_LIST) ,
         text = XmSelectionBoxGetChild(PIselectionBox, XmDIALOG_TEXT) ;
  int item = (int) client_data;
  char *procname,*msg; 
  XmString name_str;
  NumElem* numel;
  ProcessElem* pel;
  List* varlist;
  switch (item)
   {
    case SBOXADD : // Falls ein String im TextWidget :
                   // Prozessname schon vorhanden : Variablen hinzufuegen und SBox verschwinden lassen
                  //  Prozessname nicht vorhanden : Neues Prozesselement hinzufuegen und SBox verschwinden lassen
                  //  Falls kein String im TextWidget :
                  //  nichts unternehmen (auf Eingabe warten!)
                   procname = XmTextGetString(text);
                   if(strlen(procname) > 0)
                    { name_str = XmStringCreateLtoR(procname, XmSTRING_DEFAULT_CHARSET);
                      if(XmListItemExists(list,name_str) == True)
                        { 
                           pel = (ProcessElem*) ProcessList->GetFirstElem();
                           while(pel)
                            {
                               if(strcmp(pel->Getprocessname(), procname) == 0)
                                 { 
                                   
                                    SCDisplayMsg(" ");
                                    msg = new char[strlen(procname) + strlen("Add variables to process ''") +1];
                                    sprintf(msg,"Add variables to process '%s'",procname);
                                    SCDisplayMsg(msg);
                                    delete msg;  
                                    varlist = pel->Getprocessvars();
                                    UnMarkVariables(varlist,SCtheGraph);
                                    UnMarkVariables(ActProcVars,SCtheGraph);                                   
                                    ConcatenateNumLists(ActProcVars ,varlist);
                                    delete ActProcVars; ActProcVars = 0;
                                    MarkVariables(varlist,SCtheGraph);
                         
                                     
                                  }
                                pel = (ProcessElem*) ProcessList->GetNextElem();
                             }
                         }
                       else
                         {
                            SCDisplayMsg(" ");
                            msg = new char[strlen(procname) + strlen("Add variables to new process ''") +1];
                            sprintf(msg,"Add variables to new process '%s'",procname);
                            SCDisplayMsg(msg);
                            delete msg; 
                            XmListAddItem(list,name_str,0);
                            pel = new ProcessElem(procname, ActProcVars);
                            ProcessList->Append(pel);
                            ActProcVars = 0;
                           
                         }  
                      XtUnmanageChild(PIselectionBox);
                      XtSetSensitive(XmSelectionBoxGetChild(PIselectionBox, XmDIALOG_APPLY_BUTTON),True);
                      XmListDeselectAllItems(list); 
                      SensitiveCheck(SCtheGraph);
                      XmStringFree(name_str);
                      XmTextSetString(text,"");
                      } 
                   XtFree(procname);
                   break;
    case SBOXDELETE :
                   // Falls ein String im ListWidget ausgewaehlt wurde :
                   // String und Prozesselement loeschen und SBox verschwinden lassen
                   // Falls kein String im ListWidget ausgewaehlt wurde :
                   // Auf eine Auswahl warten.
                     procname = XmTextGetString(text);
                     if(strlen(procname) > 0)
                      {
                        name_str = XmStringCreateLtoR(procname,XmSTRING_DEFAULT_CHARSET);
                        if( XmListItemExists(list, name_str) == True)
                          { 
                             pel = (ProcessElem*) ProcessList->GetFirstElem();
                             while(pel)
                              {
                                if(strcmp(pel->Getprocessname(), procname) == 0)
                                  {
                                    XmListDeleteAllItems(SCMsgList);
                                    msg = new char[strlen(procname)+strlen("Delete process '' with variables:") + 1];
                                    sprintf(msg,"Delete process '%s' with variables:",procname);
                                    SCDisplayMsg(msg);
                                    delete msg;
                                    SCDisplayMsg("------------------------------------------------");
                                    varlist = pel->Getprocessvars();
                                    numel = (NumElem*) varlist->GetFirstElem();
                                    while(numel)
                                     {
                                       SCDisplayMsg(GetNameofNodeNr(numel->Getnumber(),SCtheGraph));
                                       numel = (NumElem*) varlist->GetNextElem();
                                     }
                                    UnMarkVariables(varlist,SCtheGraph);
                                    XmListDeleteItem(list,name_str);
                                    ProcessList->DelCurrentElem();
                                    SensitiveCheck(SCtheGraph);
                                  
                                    break;
                                  }
                                pel = (ProcessElem*) ProcessList->GetNextElem();
                               }    
                              XtFree(procname);
                              XmStringFree(name_str); 
                              XtUnmanageChild(PIselectionBox);
                              XtSetSensitive(XmSelectionBoxGetChild(PIselectionBox, XmDIALOG_OK_BUTTON), True);
                           }
                          XtFree(procname);
                          XmStringFree(name_str);
                          XmTextSetString(text,""); 
                         }    
                     break;
    case SBOXDISCARD :// Deselektieren der evtl. aktuell gewaehlten Variable
                     // und SBox verschwinden lassen
                    
                     if(ActProcVars)
                      {
                       UnMarkVariables(ActProcVars,SCtheGraph);
                       delete ActProcVars; ActProcVars = 0;
                      }
                     SensitiveCheck(SCtheGraph);
                     XtUnmanageChild(PIselectionBox);
                     XtSetSensitive(XmSelectionBoxGetChild(PIselectionBox, XmDIALOG_APPLY_BUTTON),True);
                     XtSetSensitive(XmSelectionBoxGetChild(PIselectionBox, XmDIALOG_OK_BUTTON), True);
                     XmListDeselectAllItems(list);
                     XmTextSetString(text,"");
                     break;
    default : assert(0);
   }
 }

 



static void SCExposeCB(Widget w, XtPointer client_data, XtPointer call_data)
        // Graph neu zeichnen entsprechend aktuellem Bildausschnitt
 {
 XmDrawingAreaCallbackStruct* cbp = (XmDrawingAreaCallbackStruct*) call_data;
 XExposeEvent* event = (XExposeEvent*) cbp->event;
 Node  *n,*m;
 char  t;
 Arg   al[10];
 int   ac;
 int   counter;
 Dimension NodeWidth,NodeHeight;

 if (event->count != 0) return;         // Wenn nicht letzter Callback einer
                                        // Kette

 if (! SCtheGraph) return;                // Wenn Graph existiert, wird er
                                        // gezeichnet (entsprechend dem
 counter=0;                             // aktuellen Bildausschnitt)
 n = SCtheGraph->GetFirstNode();
 while(n)
   {
   
     ac=0;
     XtSetArg(al[ac],XmNwidth,&NodeWidth); ac++;
     XtSetArg(al[ac],XmNheight,&NodeHeight); ac++;
     XtGetValues(SCWidgetArray[counter],al,ac);
     XMoveWindow(XtDisplay(SCWidgetArray[counter]),
                 XtWindow(SCWidgetArray[counter]),
                 n->SCGETXPOS()-SCx_offset-NodeWidth/2,
                 n->SCGETYPOS()-SCy_offset-NodeHeight/2);
     counter++;
 
     m = n->GetFirstEdge();               // Kanten zeichnen in Modus gemaess
     while(m)                             // Graph Menu
      {
        t = n->GetTypeOfCurrentEdge();
        if(t != 'E')
         {       
          XDrawLine(XtDisplay(SCcanvas),XtWindow(SCcanvas),SCGCcopy,
               n->SCGETXPOS()-SCx_offset,n->SCGETYPOS()-SCy_offset+n->SCGETNODEOFFSET(),
               m->SCGETXPOS()-SCx_offset,m->SCGETYPOS()-SCy_offset-m->SCGETNODEOFFSET());

         }
        else
          XDrawLine(XtDisplay(SCcanvas),XtWindow(SCcanvas),SCGCdash,
               n->SCGETXPOS()-SCx_offset,n->SCGETYPOS()-SCy_offset+n->SCGETNODEOFFSET(),
               m->SCGETXPOS()-SCx_offset,m->SCGETYPOS()-SCy_offset-m->SCGETNODEOFFSET());  
        m = n->GetNextEdge();
      }
     n = SCtheGraph->GetNextNode();
    }
 }

static void PIExposeCB(Widget w, XtPointer client_data, XtPointer call_data)
 {
 XmDrawingAreaCallbackStruct* cbp = (XmDrawingAreaCallbackStruct*) call_data;
 XExposeEvent* event = (XExposeEvent*) cbp->event;
 Node  *n,*m;
 char  t;
 Arg   al[10];
 int   ac;
 int   counter;
 Dimension NodeWidth,NodeHeight;

 if (event->count != 0) return;         // Wenn nicht letzter Callback einer
                                        // Kette

 if (! PItheGraph) return;                // Wenn Graph existiert, wird er
                                        // gezeichnet (entsprechend dem
 counter=0;                             // aktuellen Bildausschnitt)
 n = PItheGraph->GetFirstNode();
 while(n)
   { 
      t = n->GetNodeType();
      if (t != 'D')                        // Buttons fuer nicht-dummy Knoten
       {                                  // positionieren
        ac=0;
        XtSetArg(al[ac],XmNwidth,&NodeWidth); ac++;
        XtSetArg(al[ac],XmNheight,&NodeHeight); ac++;
        XtGetValues(PIWidgetArray[counter],al,ac);
        XMoveWindow(XtDisplay(PIWidgetArray[counter]),
                 XtWindow(PIWidgetArray[counter]),
                 n->SCGETXPOS()-PIx_offset-NodeWidth/2,
                 n->SCGETYPOS()-PIy_offset-NodeHeight/2);


        counter++;
       }
    
   m = n->GetFirstEdge();               // Kanten zeichnen 
   while(m)                             // Graph Menu
     {
     t = n->GetTypeOfCurrentEdge();
     if ( t == 'I')
       XDrawLine(XtDisplay(PIcanvas),XtWindow(PIcanvas),SCGCcopy,
               n->SCGETXPOS()-PIx_offset,n->SCGETYPOS()-PIy_offset+n->SCGETNODEOFFSET(),
               m->SCGETXPOS()-PIx_offset,m->SCGETYPOS()-PIy_offset-m->SCGETNODEOFFSET());

     m = n->GetNextEdge();
     }
   n = PItheGraph->GetNextNode();
   }
 }


static void SCResizeCB(Widget w, XtPointer client_data, XtPointer call_data)
        // Zeichenflaeche ist vergroessert worden
 {
 Arg al[10];
 int ac;
 int value;
 Dimension height,width,maxwidth,maxheight;
 Node *n;

 ac=0;                                  // Scrollbars an neuen Bildausschnitt
 XtSetArg(al[ac],XmNheight,&height); ac++;      // anpassen
 XtSetArg(al[ac],XmNwidth,&width); ac++;
 XtGetValues(SCcanvas,al,ac);
 maxwidth=max(SCGraphWidth,width);
 if (width>SCGraphWidth)                  // Wenn Zeichenflaeche ganze Graph-
   {                                    // breite anzeigen kann, soll das
   SCx_offset = (SCGraphWidth-width)/2;     // der Knoten des Hauptmodul zentral
   }                                    // liegen
 value=min(SCx_offset,maxwidth-width);
 value=max(0,value);
 if (width<=SCGraphWidth)
   {
   SCx_offset = value;
   }
 ac=0;                                  // Scrollbars neu ausrichten
 XtSetArg(al[ac],XmNsliderSize,width); ac++;
 XtSetArg(al[ac],XmNpageIncrement,max(1,width/2)); ac++;
 XtSetArg(al[ac],XmNmaximum,maxwidth); ac++;
 XtSetArg(al[ac],XmNvalue,value); ac++;
 XtSetValues(SCScrollBarHor,al,ac);
 maxheight=max(SCGraphHeight,height);
 value=min(SCy_offset,maxheight-height);
 value=max(0,value);
 ac=0;
 XtSetArg(al[ac],XmNsliderSize,height); ac++;
 XtSetArg(al[ac],XmNpageIncrement,max(1,height/2)); ac++;
 XtSetArg(al[ac],XmNmaximum,max(SCGraphHeight,height)); ac++;
 XtSetArg(al[ac],XmNvalue,value); ac++;
 XtSetValues(SCScrollBarVer,al,ac);
 if(XtWindow(SCcanvas) != NULL)
   XClearArea(XtDisplay(SCcanvas),XtWindow(SCcanvas),0,0,0,0,1);
                                        // erzeuge expose event
 }


static void PIResizeCB(Widget w, XtPointer client_data, XtPointer call_data)
 
 {
 Arg al[10];
 int ac;
 int value;
 Dimension height,width,maxwidth,maxheight;
 Node *n;

 ac=0;                                  // Scrollbars an neuen Bildausschnitt
 XtSetArg(al[ac],XmNheight,&height); ac++;      // anpassen
 XtSetArg(al[ac],XmNwidth,&width); ac++;
 XtGetValues(PIcanvas,al,ac);
 maxwidth=max(PIGraphWidth,width);
 if (width>PIGraphWidth)                  // Wenn Zeichenflaeche ganze Graph-
   {                                    // breite anzeigen kann, soll das
   PIx_offset = (PIGraphWidth-width)/2;     // der Knoten des Hauptmodul zentral
   }                                    // liegen
 value=min(PIx_offset,maxwidth-width);
 value=max(0,value);
 if (width<=PIGraphWidth)
   {
   SCx_offset = value;
   }
 ac=0;                                  // Scrollbars neu ausrichten
 XtSetArg(al[ac],XmNsliderSize,width); ac++;
 XtSetArg(al[ac],XmNpageIncrement,max(1,width/2)); ac++;
 XtSetArg(al[ac],XmNmaximum,maxwidth); ac++;
 XtSetArg(al[ac],XmNvalue,value); ac++;
 XtSetValues(PIScrollBarHor,al,ac);
 maxheight=max(PIGraphHeight,height);
 value=min(PIy_offset,maxheight-height);
 value=max(0,value);
 ac=0;
 XtSetArg(al[ac],XmNsliderSize,height); ac++;
 XtSetArg(al[ac],XmNpageIncrement,max(1,height/2)); ac++;
 XtSetArg(al[ac],XmNmaximum,max(PIGraphHeight,height)); ac++;
 XtSetArg(al[ac],XmNvalue,value); ac++;
 XtSetValues(PIScrollBarVer,al,ac);
 if(XtWindow(PIcanvas) != NULL)
   XClearArea(XtDisplay(PIcanvas),XtWindow(PIcanvas),0,0,0,0,1);
                                        // erzeuge expose event
 }

static void SCScrollCB(Widget w, XtPointer client_data, XtPointer call_data)
        // Scrollbars wurden bewegt
 {
 Arg al[10];
 int ac;
 int value;
 Dimension width;
 int which = (int) client_data;

 ac=0;                                  // neuen Bildausschnitt bestimmen
 XtSetArg(al[ac],XmNvalue,&value); ac++;// und zeichnen lassen
 XtGetValues(w,al,ac);
 if (which==SCSBHOR)                      // horizontaler Scrollbar wurde bewegt
   {
   ac=0;
   XtSetArg(al[ac],XmNwidth,&width); ac++;
   XtGetValues(SCcanvas,al,ac);
   if (width<=SCGraphWidth)
     {
     SCx_offset=value;
     }
   }
 else                                   // vertikaler Scrollbar wurde bewegt
   {
   SCy_offset=value;
   }
 XClearArea(XtDisplay(SCcanvas),XtWindow(SCcanvas),0,0,0,0,1);
                                        // erzeuge expose event
 }


static void PIScrollCB(Widget w, XtPointer client_data, XtPointer call_data)
        // Scrollbars wurden bewegt
 {
 Arg al[10];
 int ac;
 int value;
 Dimension width;
 int which = (int) client_data;

 ac=0;                                  // neuen Bildausschnitt bestimmen
 XtSetArg(al[ac],XmNvalue,&value); ac++;// und zeichnen lassen
 XtGetValues(w,al,ac);
 if (which==PISBHOR)                      // horizontaler Scrollbar wurde bewegt
   {
   ac=0;
   XtSetArg(al[ac],XmNwidth,&width); ac++;
   XtGetValues(PIcanvas,al,ac);
   if (width<=PIGraphWidth)
     {
     PIx_offset=value;
     }
   }
 else                                   // vertikaler Scrollbar wurde bewegt
   {
   PIy_offset=value;
   }
 XClearArea(XtDisplay(PIcanvas),XtWindow(PIcanvas),0,0,0,0,1);
                                        // erzeuge expose event
 }
                
static void PINodeCB(Widget w, XtPointer client_data, XtPointer call_data)
 {
   int button = (int) client_data;
   if(XtIsManaged(PIListArray[button]))
    { 
      XtUnmanageChild(PIListArray[button]);
    }
   else
    {
      XtManageChild(PIListArray[button]);
    }
       
 }

static void SCNodeCB(Widget w, XtPointer client_data, XtPointer call_data)
        // Ein Knoten im Variablenbaum wurde angewaehlt
{

  Node* n;
  int i,varnum;
  NumElem* numel;
  List *ActselectVars = 0;
  char *NodeName = (char*) client_data;
  n = SCtheGraph->GetFirstNode();
  while(n)
   {
     if(strcmp(NodeName, n->GetName()) == 0)
      break;
     n = SCtheGraph->GetNextNode();
   }

 if(n && (n->SCGETWIDGETNR() > 0))
  {  
     ActProcVars = new List;
     ActselectVars = new List;
     while(n && (n->GetNodeType() == 'A' ) && n->HasPTpredecessor())  // falls ARRAY-Variable
  
      {
         n = n->GetPTpredecessor();
      }
     CollectActionvars( n,  ActionVarsList,ActProcVars, ActselectVars );
     CollectSelectedVars(ActProcVars, ActselectVars, SCtheGraph);
     ConcatenateNumLists(ActselectVars,ActProcVars);
     delete ActselectVars; ActselectVars = 0;
     XmListDeleteAllItems(SCMsgList);
     SCDisplayMsg("Choosen variables:");
     SCDisplayMsg("__________________");
     numel = (NumElem*) ActProcVars->GetFirstElem();
     while(numel)
      {
        SCDisplayMsg(GetNameofNodeNr(numel->Getnumber(),SCtheGraph));
        numel = (NumElem*) ActProcVars->GetNextElem(); 
      } 
     MarkVariables(ActProcVars,SCtheGraph);
     SensitiveCheck(SCtheGraph);   
     XtManageChild(PIselectionBox);
     XtSetSensitive(XmSelectionBoxGetChild(PIselectionBox, XmDIALOG_APPLY_BUTTON),False); 
 
  }
 }









// Hilfsfunktionen zur Window Generierung
//---------------------------------------

static Widget CreateMenu(Widget menubar, char* name, char mnemonic)
        // Menu mit Name und Abkuerzung generieren
 {
 Arg al[10];
 int ac;
 Widget   menu,button;
 XmString name_str;

 ac=0;                                  // menupane erzeugen
 menu = XmCreatePulldownMenu(menubar,name,al,ac);
 name_str = XmStringCreateSimple(name);
 ac=0;                                  // Button fuer Menu
 XtSetArg(al[ac],XmNsubMenuId,menu); ac++;
 XtSetArg(al[ac],XmNlabelString,name_str); ac++;
 XtSetArg(al[ac],XmNmnemonic,mnemonic); ac++;
 button = XmCreateCascadeButton(menubar,name,al,ac);
 XtManageChild(button);
 XmStringFree(name_str);
 return(menu);
 }



static void CreateMenuItem(Widget menupane, char* name, char mnemonic,
                           XtPointer client_data )
        // MenuItem mit Name, Abkuerzung und client_data fuer MenuCB erzeugen
 {
 Arg al[10];
 int ac;
 Widget   item;
 XmString name_str;



 name_str = XmStringCreateSimple(name); // MenuItem erzeugen
 ac=0;
 XtSetArg(al[ac],XmNlabelString,name_str); ac++;
 XtSetArg(al[ac],XmNmnemonic,mnemonic); ac++;
 item = XmCreatePushButton(menupane,name,al,ac);
 XtManageChild(item);
 XtAddCallback(item,XmNactivateCallback,SCMenuCB,client_data);
 XmStringFree(name_str);
 
 }



static void SCCreateGC()
  // Graphic Context fuer das Variablen- Anzeigefenster erzeugen

 {
 Arg al[10];
 int ac;
 Pixel fg, bg;
 XGCValues vals;

 ac = 0;
 XtSetArg(al[ac],XmNforeground,&fg); ac++;
 XtSetArg(al[ac],XmNbackground,&bg); ac++;
 XtGetValues(SCcanvas, al,ac);
 vals.foreground = fg;
 vals.background = bg;
 // Durchgezogene Linie fuer korrekte Systemvariablen
 SCGCcopy = XtGetGC(SCcanvas, GCForeground | GCBackground, &vals);

 vals.foreground=fg;
 vals.background=bg;
 vals.line_style=LineOnOffDash;
 // gestrichelte Linie fuer Systemvariablen, die nicht in '[Sysvar1,...]' stehen
 SCGCdash = XtGetGC(SCcanvas,GCForeground | GCBackground | GCLineStyle,&vals);
 }


static void SCInitFontConstants()
 // Fonthoehe und Breite merken

 {
  Arg al[10];
  int ac;
  XmString test_str;
  XmFontList fontlist;
  // Fontdaten aus Message-Window
  ac = 0;
  XtSetArg(al[ac], XmNfontList, &fontlist);ac++;
  XtGetValues(SCMsgList, al,ac);
  test_str =  XmStringCreateSimple("M");
  SCFontHeight = XmStringHeight(fontlist, test_str);
  SCFontWidth =  XmStringWidth(fontlist, test_str);
  XmStringFree(test_str);
  SCHalfFontHeight = SCFontHeight / 2;
 }

// Erzeugung des Konzeptfensters
static void CreatePIGraphWindow()

{
  Arg al[10];
  int ac;
  Widget commands, frame, scrollwin, closebutton;
  XmString name_str;
  char* wintitle = new char[strlen("process / interface view") + 1];

  strcpy(wintitle, "process / interface view");

  ac = 0;
  XtSetArg(al[ac], XmNheight, PIDIALOGHEIGHT);ac++;
  XtSetArg(al[ac], XmNwidth , PIDIALOGWIDTH);ac++;
  XtSetArg(al[ac], XmNtitle, wintitle);ac++;
  XtSetArg(al[ac], XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL);ac++;
  XtSetArg(al[ac], XmNresizePolicy, XmRESIZE_NONE);ac++;

// FormDialog

  PIdialogshell = XmCreateFormDialog(SCshell, "PIdialog", al,ac);

  delete wintitle;

// FormDialog/RowColumn

  ac = 0;
  XtSetArg(al[ac], XmNtopAttachment, XmATTACH_FORM); ac++;
  XtSetArg(al[ac], XmNleftAttachment, XmATTACH_FORM); ac++;
  commands = XmCreateRowColumn(PIdialogshell,"commands",al,ac);
  XtManageChild(commands);

// RowColumn/PushButton
  name_str = XmStringCreateSimple("Close");
  ac = 0;
  XtSetArg(al[ac] , XmNlabelString , name_str); ac++;
  XtSetArg(al[ac] , XmNmnemonic,'C'); ac++;
  closebutton = XmCreatePushButton(commands, "closebutton" , al, ac);
  XtAddCallback(closebutton, XmNactivateCallback,PIMenuCB, XtPointer(PICLOSE));
  XtManageChild(closebutton);
  XmStringFree(name_str);

// FormDialog/ScrolledWindow
  ac = 0;
  XtSetArg(al[ac], XmNtopAttachment, XmATTACH_FORM);ac++;
  XtSetArg(al[ac], XmNbottomAttachment,XmATTACH_FORM);ac++;
  XtSetArg(al[ac], XmNleftAttachment, XmATTACH_WIDGET);ac++;
  XtSetArg(al[ac], XmNleftWidget, commands);ac++;
  XtSetArg(al[ac], XmNrightAttachment, XmATTACH_FORM); ac++;
  XtSetArg(al[ac], XmNscrollBarDisplayPolicy,XmSTATIC); ac++;
  scrollwin = XmCreateScrolledWindow(PIdialogshell,"scrollwin",al,ac);
  XtManageChild(scrollwin);

// ScrolledWindow/ScrollBar
  ac = 0;
  XtSetArg(al[ac], XmNorientation, XmHORIZONTAL);ac++;
  XtSetArg(al[ac], XmNmaximum, PICANVASWIDTH);ac++;
  XtSetArg(al[ac], XmNsliderSize, PICANVASWIDTH); ac++;
  XtSetArg(al[ac], XmNpageIncrement , PICANVASWIDTH / 2); ac++;
  PIScrollBarHor = XmCreateScrollBar(scrollwin,"PIScrollBarHor",al, ac);
  XtManageChild(PIScrollBarHor);

  XtAddCallback(PIScrollBarHor,XmNvalueChangedCallback, PIScrollCB,
                                                        XtPointer(PISBHOR));
// ScrolledWindow/ScrollBar
  ac = 0;
  XtSetArg(al[ac], XmNorientation, XmVERTICAL); ac++;
  XtSetArg(al[ac], XmNmaximum, PICANVASHEIGHT); ac++;
  XtSetArg(al[ac], XmNsliderSize, PICANVASHEIGHT);ac++;
  XtSetArg(al[ac], XmNpageIncrement , PICANVASHEIGHT / 2); ac++;
  PIScrollBarVer = XmCreateScrollBar(scrollwin,"PIScrollBarVer", al ,ac);
  XtManageChild(PIScrollBarVer);

  XtAddCallback(PIScrollBarVer,XmNvalueChangedCallback, PIScrollCB,
                                                        XtPointer(PISBVER));
// ScrolledWindow/Frame
  ac = 0;
  XtSetArg(al[ac], XmNshadowType , XmSHADOW_IN); ac++;
  frame = XmCreateFrame(scrollwin, "frame" ,al, ac);
  XtManageChild(frame);

// Frame/DrawingArea
  ac = 0;
  XtSetArg(al[ac], XmNheight, PICANVASHEIGHT); ac++;
  XtSetArg(al[ac], XmNwidth, PICANVASWIDTH); ac++;
  PIcanvas = XmCreateDrawingArea(frame, "PIcanvas", al, ac);
  XtManageChild(PIcanvas);

  XtAddCallback(PIcanvas, XmNexposeCallback, PIExposeCB,0);
  XtAddCallback(PIcanvas, XmNresizeCallback, PIResizeCB,0);

 }

// Erzeugung des Prozessfensters
static void CreateProcessListWindow()
 {
   Arg  al[10];
   int  ac;
   Widget button,list;
   XmString label_str,title_str,list_label_str,name_str;
   int  i,ok;
   char *name;

   label_str = XmStringCreateSimple("Select a processname :");
   list_label_str = XmStringCreateSimple("Processes");
   title_str = XmStringCreateSimple("Processlist");
   ac=0;
   XtSetArg(al[ac],XmNautoUnmanage,False); ac++;
   XtSetArg(al[ac],XmNselectionLabelString,label_str); ac++;
   XtSetArg(al[ac],XmNlistLabelString,list_label_str); ac++;
   XtSetArg(al[ac],XmNdialogTitle,title_str); ac++;
   XtSetArg(al[ac],XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL);ac++;
   PIselectionBox = XmCreateSelectionDialog(SCshell,"PIselectionBox",al,ac);
   XtAddCallback(PIselectionBox,XmNokCallback,ProcessCB,XtPointer(SBOXADD));
   XtAddCallback(PIselectionBox,XmNapplyCallback,ProcessCB,XtPointer(SBOXDELETE));
   XtAddCallback(PIselectionBox,XmNcancelCallback,ProcessCB,XtPointer(SBOXDISCARD));
   XtUnmanageChild(XmSelectionBoxGetChild(PIselectionBox,XmDIALOG_HELP_BUTTON));
   XmStringFree(label_str);
   XmStringFree(list_label_str);
   XmStringFree(title_str);

 // Add Button

   button = XmSelectionBoxGetChild(PIselectionBox,XmDIALOG_OK_BUTTON);
   label_str = XmStringCreateSimple("Add");
   ac=0;
   XtSetArg(al[ac],XmNlabelString,label_str); ac++;
   XtSetValues(button,al,ac);
   XmStringFree(label_str);

 // Delete Button

    button = XmSelectionBoxGetChild(PIselectionBox,XmDIALOG_APPLY_BUTTON);
    label_str = XmStringCreateSimple("Delete");
    ac=0;
    XtSetArg(al[ac],XmNlabelString,label_str); ac++;
    XtSetValues(button,al,ac);
    XmStringFree(label_str);

 // Cancel Button

    button = XmSelectionBoxGetChild(PIselectionBox,XmDIALOG_CANCEL_BUTTON);
    label_str = XmStringCreateSimple("Discard");
    ac=0;
    XtSetArg(al[ac],XmNlabelString,label_str); ac++;
    XtSetValues(button,al,ac);
    XmStringFree(label_str);

 }

// Exporte
//--------

// Erzeugung des Variablenfensters. Lesen der Systemvariablen-Datei 'specname'.
// Bei Erfolg wird 1, sonst 0 zurueckgeliefert.
int CreateSemConWin(Widget toplevel, char* specname)
 {

  Arg al[10];
  int ok,ac  ;
 
 Widget menubar,menu,pane,scrollwin,frame,form; 
 
 if(SCshell != NULL)
  {
    DisplaySemConWin();
    return 1;
  } 
 char *wintitle = new char[strlen(specname) + strlen("Semantic concept of Spec ") + 1];
 sprintf(wintitle,"Semantic concept of Spec %s",specname);

 ac = 0;
 XtSetArg(al[ac],XmNheight,SCSHELLHEIGHT); ac++;
 XtSetArg(al[ac],XmNwidth,SCSHELLWIDTH); ac++;
 SCshell = XtAppCreateShell(wintitle,"SemConShell",
                                 applicationShellWidgetClass,
                                 XtDisplay(toplevel),al,ac);

 delete wintitle;


 // shell->form(1/1)  (i/j) = i-ter von j Soehnen
 
 ac = 0;

 form = XmCreateForm(SCshell,"form",al,ac);
 XtManageChild(form);
 delete wintitle;
 


 // shell->form->menubar(1/2)
 // plus menuitems

 ac=0;
 XtSetArg(al[ac],XmNtopAttachment,XmATTACH_FORM); ac++;
 XtSetArg(al[ac],XmNleftAttachment,XmATTACH_FORM); ac++;
 XtSetArg(al[ac],XmNrightAttachment,XmATTACH_FORM); ac++;
 menubar = XmCreateMenuBar(form,"menubar",al,ac);
 XtManageChild(menubar);


 menu = CreateMenu(menubar,"Window",'W');
  CreateMenuItem(menu,"Close",'C',XtPointer(SCCLOSE));
 menu = CreateMenu(menubar,"Semantic Concept",'S');
  CreateMenuItem(menu,"New Concept",'N',XtPointer(SCNEW));
  CreateMenuItem(menu,"Delete Process",'P',XtPointer(SCDELETE));
  CreateMenuItem(menu,"Display Concept",'D',XtPointer(SCDISPLAY));

 // shell->form->panedwindow(2/2)


 ac=0;
 XtSetArg(al[ac],XmNtopAttachment,XmATTACH_WIDGET);  ac++;
 XtSetArg(al[ac],XmNtopWidget,menubar);              ac++;
 XtSetArg(al[ac],XmNleftAttachment,XmATTACH_FORM);   ac++;
 XtSetArg(al[ac],XmNrightAttachment,XmATTACH_FORM);  ac++;
 XtSetArg(al[ac],XmNbottomAttachment,XmATTACH_FORM); ac++;
 pane = XmCreatePanedWindow(form,"pane",al,ac);
 XtManageChild(pane);



 // shell->form->panedwindow->scrolledwindow(1/2)

 ac=0;
 XtSetArg(al[ac],XmNscrollBarDisplayPolicy,XmSTATIC); ac++;
 XtSetArg(al[ac],XmNallowResize, False);ac++;
 XtSetArg(al[ac],XmNheight,SCSHELLHEIGHT - 200);ac++;
 XtSetArg(al[ac],XmNwidth,SCSHELLWIDTH);ac++;
 scrollwin = XmCreateScrolledWindow(pane,"scrollwin",al,ac);
 XtManageChild(scrollwin);



 ac=0;
 XtSetArg(al[ac],XmNorientation,XmHORIZONTAL); ac++;
 XtSetArg(al[ac],XmNmaximum,SCCANVASWIDTH); ac++;
 XtSetArg(al[ac],XmNsliderSize,SCCANVASWIDTH); ac++;
 XtSetArg(al[ac],XmNpageIncrement,SCCANVASWIDTH/2); ac++;
 SCScrollBarHor = XmCreateScrollBar(scrollwin,"SCScrollBarHor",al,ac);
 XtManageChild(SCScrollBarHor);
 XtAddCallback(SCScrollBarHor,XmNvalueChangedCallback,SCScrollCB,XtPointer(SCSBHOR));



 ac=0;
 XtSetArg(al[ac],XmNorientation,XmVERTICAL); ac++;
 XtSetArg(al[ac],XmNmaximum,SCCANVASHEIGHT); ac++;
 XtSetArg(al[ac],XmNsliderSize,SCCANVASHEIGHT); ac++;
 XtSetArg(al[ac],XmNpageIncrement,SCCANVASHEIGHT/2); ac++;
 SCScrollBarVer = XmCreateScrollBar(scrollwin,"SCScrollBarVer",al,ac);
 XtManageChild(SCScrollBarVer);
 XtAddCallback(SCScrollBarVer,XmNvalueChangedCallback,SCScrollCB,XtPointer(SCSBVER));



 ac=0;
 XtSetArg(al[ac],XmNshadowType,XmSHADOW_IN); ac++;
 frame = XmCreateFrame(scrollwin,"frame",al,ac);
 XtManageChild(frame);



 ac=0;
 XtSetArg(al[ac],XmNheight,SCCANVASHEIGHT); ac++;
 XtSetArg(al[ac],XmNwidth,SCCANVASWIDTH);  ac++;
 SCcanvas = XmCreateDrawingArea(frame,"SCcanvas",al,ac);
 XtManageChild(SCcanvas);
 XtAddCallback(SCcanvas,XmNexposeCallback,SCExposeCB,0);
 XtAddCallback(SCcanvas,XmNresizeCallback,SCResizeCB,0);



 XmScrolledWindowSetAreas(scrollwin,SCScrollBarHor,SCScrollBarVer,frame);


 // shell->form->panedwindow->list(2/2)

 ac=0;
 XtSetArg(al[ac],XmNvisibleItemCount,SCMSGLISTVISIBLEITEMS); ac++;
 XtSetArg(al[ac],XmNselectionPolicy,XmSINGLE_SELECT); ac++;
 XtSetArg(al[ac],XmNlistSizePolicy,XmRESIZE_IF_POSSIBLE); ac++;
 XtSetArg(al[ac],XmNallowResize,False);ac++;
 SCMsgList = XmCreateScrolledList(pane,"SCMsgList",al,ac);
 XtManageChild(SCMsgList);
 XtAddCallback(SCMsgList,XmNsingleSelectionCallback,SC_PIListCB,NULL);

 SCCreateGC();
 SCInitFontConstants();
 CreateProcessListWindow();
 CreatePIGraphWindow(); 

// Einlesen der Systemvariablen-Datei.
 ok = SCReadGraph(specname);

 if(!ok)
  {
    XtDestroyWidget(SCshell);
    return 0;
  }
 else
  { 
     XtRealizeWidget(SCshell);
     XClearArea(XtDisplay(SCcanvas),XtWindow(SCcanvas),0,0,0,0,1);   // erzeuge expose event
     return 1;
  }
 

 }


// Anzeige des Variablenfensters auf dem Bildschirm
void DisplaySemConWin()
 {
     XRaiseWindow(XtDisplay(SCshell),XtWindow(SCshell));
     XtPopup(SCshell,XtGrabNone);
 }

// Loeschen des Variablenfensters
void DestroySemConWin()
 {

  
  if(SCWidgetArray)
   {
    for (int i=0; i<SCWidgetArrayLength; i++)
     {
     XtDestroyWidget(SCWidgetArray[i]);
     }
   delete [] SCWidgetArray;
   SCWidgetArray = 0;
   }

  SCGraphHeight = 0;
  SCGraphWidth = 0;
  SCWidgetArrayLength = 0;
  if(ActionVarsList) { delete ActionVarsList; ActionVarsList = 0;}
  if(ProcessList) { delete ProcessList; ProcessList = 0;}
  if(SCtheGraph) { delete SCtheGraph; SCtheGraph = 0;}
  if(PIdialogshell != NULL)
    {
     XtDestroyWidget(PIdialogshell);
     PIdialogshell = NULL;
    }   

  if(SCshell != NULL)
   {
     XtDestroyWidget(SCshell);
     SCshell = NULL;
   }


 }
