/*----------------------------------------------------------------------*
 * HiGraph.h
 *
 * Michael Geikowski, 27.3.93
 *
 * Definiert die Klasse HiGraph zur Verwaltung von Hierarchie-Graphen.
 *----------------------------------------------------------------------*/

#ifndef HIGRAPH_H
#define HIGRAPH_H

#include "adjlistgnu.h"
#include "pqueuegnu.h"

class HiGraph
//===========
	// Speichert einen Graphen als Hierarchie-Graphen, d.h. die Knoten
        // sind in Level 0..n eingeteilt. Knoten auf Level 0 haben keine
        // Eingangskanten. Knoten auf Level i (i>0) haben nur Eingangskanten,
        // die von Knoten der Level 0..i-1 ausgehen. Die Einteilung in Level
        // wird nicht(!) durch diese Klasse vorgenommen. Es werden nur Zeiger
        // auf die Knoten verwaltet. Speicherplatz fuer die Knoten muss
        // anderswo zur Verfuegung gestellt werden. Die Hilfsvariablen
        // Je Level existiert eine Priority Queue, die alle Knoten des Levels
        // aufnehmen kann.
 {
 public:

   HiGraph(AdjList *al, int NrOfLevels);
 //-------------------------------------
 	// Generiert aus einer Adjazenzlisten-Darstellung al eines Graphen
        // eine neue Hierarchie-Graphen Darstellung. Die Knoten des Graphen
        // muessen bereits in Level eingeteilt sein (durch Node::SetLevel()).
        // NrOfLevels gibt die Anzahl der zugewiesenen Level an. Die Level
        // muessen von 0 bis NrOfLevel-1 durchnummeriert sein.
        // Pos und TempPos je Knoten erhalten die Nummer des Knotens auf
        // seinem Level (1..NrOfNodes).

   ~HiGraph();
 //-----------
 	// Loescht die HiGraph Datenstruktur, aber keinen Knoten.

   Node* FirstNodeOfLevel(int Level);
 //----------------------------------
 	// Startet neuen Durchlauf durch die Knoten eines Levels `Level' und
        // liefert ersten Knoten des Levels. Es muss mindestens ein Knoten
        // des Levels existieren.

   Node* NextNodeOfLevel(int Level);
 //---------------------------------
 	// Liefert naechsten Knoten des aktuellen Durchlaufs auf Level `Level'
        // bzw. Null, wenn alle Knoten durchlaufen sind.

   Node* GetNodeOfLevel(int Level, int nr);
 //----------------------------------------
 	// Liefert `nr'-ten Knoten auf Level `Level'

   void SetNodeOfLevel(Node *n, int Level, int nr);
 //------------------------------------------------
  	// Setzt `nr'-ten Knoten des Levels `Level' auf `n'

   PQueue* GetQueueOfLevel(int Level)	// Gibt Queue zu Level `Level'
 //----------------------------------
	{ assert(the_Graph && 0<=Level && Level<Levels);
	  return the_Graph[Level].queue; }

   int GetNodesOnLevel(int Level)	// Gibt Anzahl Knoten auf Level `Level'
 //------------------------------
	{ assert(the_Graph && 0<=Level && Level<Levels);
	  return the_Graph[Level].NrOfNodes; }

   int GetNrOfLevels()			// Gibt Anzahl der Level
 //-------------------
	{ return Levels; }

 private:

   class LevelInfo	// Hilfsklasse zur Verwaltung aller Knoten eines Levels
   //=============
    {
    public:

      LevelInfo();
    //------------
 
      ~LevelInfo();
    //-------------

      int    NrOfNodes;		// Anzahl Knoten des Levels
      int    current;		// Merker fuer Durchlauf durch Knoten des Levels
      Node   **NodeArray;	// Array aller Knoten des Levels
      PQueue *queue;		// Priority Queue fuer NrOfNodes viele Knoten
    };

   LevelInfo* the_Graph;	// Array aller Level
   int        Levels;		// Anzahl Level
 };

#endif
