/* SATM REGION START */ 
/*****************************************************************************

	TLA+ Modul - BNF

	(E) Version 2.0         9.7.92		Mester
   Alle Rechte gem. dt. UrhG und WissHG vorbehalten.
   Jedwede Verwendung nur mit schriftlicher Genehmigung des Autors.
	design decisions are appended as comment to the 
	end of this file

	2.0: - adaption to 'a grammar for TLA+' as of May, 26.  
	     - modifications to remove reduce/reduce-errors
	     - adapting Leslies comments
	     THESE STEPS HAVE NOT BEEN DOCUMENTED HITHERTO !

 *****************************************************************************/
/* SATM REGION END */

/*****************************************************************************
 * TLA+ Browser : Praeprozessor - Parser Pass 1
 *
 * Michael Geikowski, 8.7.93
 *
 * ( without design decisions )
 *****************************************************************************/

%{

/***
 *** C declarations section
 ***/

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include "adjlistgnu.h"
#include "symlist.h"

// Ersatz des Headers fuer scanner.l

extern SymList *sList;

extern void reinitialize_lex (FILE* ifp,char* filename);
extern int  yylex();

// Globale Variablen

AdjList*  the_AdjList;
NodeList* NodeQueue;
Node*     this_module;

// einige forward Deklarationen

extern void yyerror (char* s);
static void HandleNewNode(SymListElem* selem, char EdgeType);

#define YYSTYPE SymListElem*

%}

/***
 *** Bison declarations section
 ***/

%expect 2
/* Zwei s/r Konflikte:
 * 1)
 *      ( SetExpr . )  >>>reduce>>>  ( expr . ) 
 *                     >>>shift>>>>  ( SetExpr ) .
 *   Hier ist shiften gewuenscht : Parser verhaelt sich somit korrekt.
 *
 * 2)
 *      ( expr . )  >>>reduce>>>  ( constExpr_csl . ) 
 *                  >>>shift>>>>  ( expr ) .
 *   Hier ist shiften gewuenscht : Parser verhaelt sich somit korrekt.
 * Im zweiten Fall drueckt sich der Konflikt zwischen gklammerten Ausdruecken
 * (shift) und einelementigen Tupeln (reduce) aus. Nach Lamport sind aber
 * einelementige Tupel nicht erlaubt. 
 */

		/* TLA+ reserved names */
%token
    ANY
    NAT

/* TLA+ directives */

    INFIX		/* left-associatve infix   */
    INFIXR		/* right-associative infix */

/* Relatoren */

    EQ		NE	LT	GT 	LE  	GE
		
/* :: Ersatzdarstellungen mathematischer Symbole */

    _ALL_QUANTOR_	_EXISTS_QUANTOR_	_BOLD_EXISTS_
    _L_NOT_		_L_OR_			_L_AND_
    PREF_AND		PREF_OR		/* for use by lexical analysis	*/
    IBL			IBR		/* dito, invisible braces	*/
    PRIME
    _ELEMENT_OF_	_NOT_ELEMENT_OF_	
    _SUBSET_OF_ 	_SET_DIFFERENCE_	_UNION_		_INTERSECT_
    _EMPTYSET_		
    _IMPLIES_		_EQUIV_		
    _LEADSTO_		_HENCEFORTH_		_EVENTUALLY_	
    _CROSS_   				/* X */
    _MAPSTO_		_RIGHTARROW_

/* TLA+ keywords */

    MODULE
    PARAMETERS	AS			 
    EXPORT	OPAQUE	IMPORT	INCLUDE
    END						/* of one module */
    IF		THEN	ELSE 	
    CASE	
    FALSE	TRUE
    SF		WF		
    IN		SUBSET
    ENABLED	LET		
    CHOOSE	UNCHANGED
    UNION
    DOMAINt	/* read 'domain token';	DOMAIN is already defined in math.h */
    REPLACE		WITH
    DEFINED_AS		MINUS		PLUS
    DIV			MOD		POWER
    DIVIDED		TIMES
   
/* the sorts */

    CONSTANT
    BOOLEAN
    PREDICATE
    FUNCTION				/* for STATE / TRANSITION */
    ACTION
    TEMPORAL 

/* the 'sections' */

    CONSTANTS	BOOLEANS		
    PREDICATES  STATE	TRANSITION	FUNCTIONS 	ACTIONS	
    TEMPORALS
    ASSUMPTIONS		THEOREMS

/* Identifier */

	normalID	/* usual identifier w upper/lower case  */
			/* (also used as filename)		*/ 

/* Basisdatentypen */

	STRING
	NATNUM 

		/* Trenner */

	DOT
	DOUBLECOLON
	COLON
	SEMICOLON
	COMMA
	RBR_ON		RBR_OFF		/* RBR round brackets ()             */
	DSBR_ON		DSBR_OFF	/* D/SBR (double) square brackets [] */

/* syntactic sugar / additional tokens to avoid ambiguity */

	 CHANGE
	 SET
	 TUPELSELECTOR

/* associativity and precedence */

%left	_LEADSTO_
%left	_IMPLIES_ _EQUIV_
%left	_L_AND_	_L_OR_
%left	EQ NE LT GT LE GE _SUBSET_OF_
%left	PLUS MINUS TIMES DIVIDED DIV MOD POWER
%left	_SET_DIFFERENCE_ _UNION_ _INTERSECT_ _CROSS_
%left	_ELEMENT_OF_ _NOT_ELEMENT_OF_
%left   UNARYMINUS
%left   PRECLOWERTHANDOT
%left	TUPELSELECTOR DOT
%left   PRIME
%%

/***
 *** grammar rules section
 ***
 *** Explanations:
 *** The name of the transitions inform the reader about what they do.
 *** Names given in the singular form represent one occurrence in the BNF.
 *** Transitions ending in an plural 's' mean arbitrary occurrence, while
 *** an added '_' denotes at least one occurrence of the transition. An
 *** added '01' means zero or one occurrence of the according transition.
 *** 
 *** a 'csl' should denote a comma separated list / am
 *** a 'ssl' denotes a semicolon separated list / am
 *** a 'bsl' denotes a blank separated list / am
 ***/

OneModule:
	ModuleType MODULE ModuleName01
	HeaderParts 
	BodyParts
	ModuleEnd
;

ModuleType:
		/* Knotentyp festlegen */
	/* empty */
        { this_module->SetNodeType('T'); }
     |  CONSTANT
        { this_module->SetNodeType('C'); }
     | 	TEMPORAL
        { this_module->SetNodeType('T'); }
     ;

ModuleName01:
	/* empty */
     |	ModuleName
     ;

HeaderParts:
	/* empty */
     |	HeaderParts HeaderPart
     |  HeaderParts error SEMICOLON
     ;

HeaderPart: 
    	IMPORT ImportList SEMICOLON  
    |	EXPORT OPAQUE Restexport
    |	EXPORT Export_ID_csl OPAQUE Restexport
    |	EXPORT Export_ID_csl SEMICOLON
    |	PARAMETERS formalPar_ssl SEMICOLON
;

Export_ID:
	Export_qualifierPrefixes normalID
;

Export_qualifierPrefixes: 
	/* empty */
   |	Export_qualifierPrefixes normalID DOT
   |	Export_qualifierPrefixes normalID RBR_ON normalID_csl RBR_OFF DOT
;

Export_ID_csl:
	Export_ID
    |	Export_ID_csl COMMA Export_ID
;

Restexport: 
	Export_ID_csl SEMICOLON
    |	SEMICOLON
    ;

BodyParts:
	/* empty */
    |	BodyParts BodyPart
;

BodyPart:
    	IncludeEntry
    |   CM_Definition
    |	TM_Definition
    |	OneModule
;

CM_Definition: 
	CONSTANTS ConstantDefinitions
    |	BOOLEANS BooleanDefinitions
    |	CONSTANT ASSUMPTIONS AssumptionDefinitions
    |	ASSUMPTIONS AssumptionDefinitions
    |	CONSTANT THEOREMS TheoremDefinitions
    |	THEOREMS TheoremDefinitions
;

TM_Definition: 
    	PREDICATES PredicateDefinitions
    |	STATE FUNCTIONS	StateFunctionDefinitions
    |	ACTIONS	ActionDefinitions
    |	TRANSITION FUNCTIONS TransitionFunctionDefinitions
    |	TEMPORALS TemporalDefinitions
    |	ACTION ASSUMPTIONS AssumptionDefinitions
    |	ACTION THEOREMS TheoremDefinitions
    |	TEMPORAL THEOREMS TheoremDefinitions
;

ModuleEnd: 
	END
    |	END ModuleName
;

   /* 
    * definitions for the import-mechanism in a module header, ok
    */ 

ImportList: 
		/* Nachfolgerknoten mit Typ `IMPORT' */
	ModuleName
        { HandleNewNode($1,'M'); }
    |	ModuleName AS qualifyingID
        { HandleNewNode($1,'M'); }
    | 	ImportList COMMA ModuleName
        { HandleNewNode($3,'M'); }
    |	ImportList COMMA ModuleName AS qualifyingID
        { HandleNewNode($3,'M'); }
;

IncludeEntry:
	INCLUDE IncludeParts
;

IncludeParts:
	IncludePart SubstitutionPart01 SEMICOLON
    |	IncludeParts IncludePart SubstitutionPart01 SEMICOLON
    ;

IncludePart: 
		/* Nachfolgerknoten mit Typ `INCLUDE' */
    	ModuleName
        { HandleNewNode($1,'N'); }
    |   ModuleName AS qualifyingID IncludeParameters01
        { HandleNewNode($1,'N'); }
;

IncludeParameters01: 
	/* empty */
    |	RBR_ON formalPar_ssl RBR_OFF
;

SubstitutionPart01:
	/*empty */
    |	WITH Substitution_ 
;

Substitution_:
	OneSubstitution_
    |	Substitution_ COMMA OneSubstitution_ 
;

OneSubstitution_:
	normalID REPLACE expr
;

/*
 * transitions for the various symbol-definition parts within a module body
 */

ConstantDefinitions:
    	/* empty */
    |   ConstantDefinitions ConstantDefinition SEMICOLON
    |   ConstantDefinitions error SEMICOLON
;

BooleanDefinitions:
	/* empty */
    |   BooleanDefinitions BooleanDefinition SEMICOLON
    |   BooleanDefinitions error SEMICOLON
;

PredicateDefinitions:
	/* empty */
    |   PredicateDefinitions PredicateDefinition SEMICOLON
    |   PredicateDefinitions error SEMICOLON
;

StateFunctionDefinitions:
	/* empty */
    |   StateFunctionDefinitions StateFunctionDefinition SEMICOLON
    |   StateFunctionDefinitions error SEMICOLON
;

ActionDefinitions:
	/* empty */
    |   ActionDefinitions ActionDefinition SEMICOLON
    |   ActionDefinitions error SEMICOLON
;

TransitionFunctionDefinitions:
	/* empty */
    |   TransitionFunctionDefinitions TransitionFunctionDefinition SEMICOLON
    |   TransitionFunctionDefinitions error SEMICOLON
;

AssumptionDefinitions:
	/* empty */
    |   AssumptionDefinitions AssumptionDefinition SEMICOLON
    |   AssumptionDefinitions error SEMICOLON
;

TheoremDefinitions:
	/* empty */
    |   TheoremDefinitions TheoremDefinition SEMICOLON
    |   TheoremDefinitions error SEMICOLON
;

TemporalDefinitions:
	/* empty */
    |   TemporalDefinitions TemporalDefinition SEMICOLON
    |   TemporalDefinitions error SEMICOLON
;

/*
 * the real definitions
 */

ConstantDefinition:
	SymbolDef_LHS DEFINED_AS ConstantExpr
;

BooleanDefinition:
	SymbolDef_LHS DEFINED_AS BooleanExpr 
;

PredicateDefinition:
	SymbolDef_LHS DEFINED_AS PredExpr 
;

StateFunctionDefinition:
	SymbolDef_LHS DEFINED_AS StateFcnExpr 
;

ActionDefinition:
	SymbolDef_LHS DEFINED_AS ActionExpr 
;

TransitionFunctionDefinition:
	SymbolDef_LHS DEFINED_AS TransFctExpr 
;

AssumptionDefinition: 
	SymbolDef_LHS DEFINED_AS expr 	
;

TheoremDefinition:
	SymbolDef_LHS DEFINED_AS expr 
;

TemporalDefinition:
	SymbolDef_LHS DEFINED_AS expr 
;

/*
 * expressions for the various classes
 * (we now get a confluent system ...)
 */

ConstantExpr:	
	expr
;

BooleanExpr:	
	expr 
;

StateFcnExpr: 
	expr
;

PredExpr: 
	expr
;

TransFctExpr: 
	expr
;

ActionExpr: 
	expr
;

/*
 * the various MAJOR expressions we need to construct the
 * symbol definitions within the various sections
 */

constExpr_csl: 
   	expr
   |	constExpr_csl COMMA expr
   ;

expr_csl: 
   	expr
   |	expr_csl COMMA expr
   ;

expr: 
	/* LCexpr parts */ 

    	RBR_ON expr RBR_OFF
    |	RBR_ON constExpr_csl RBR_OFF
    |	expr TUPELSELECTOR
    |  	TRUE 
    |	FALSE
    |	expr boolRel expr %prec _L_AND_
    |	expr boolRelIB IBL expr IBR %prec _IMPLIES_
    | 	_L_NOT_ IBL expr IBR
    |	_ALL_QUANTOR_ isElementOfSetListQB DOUBLECOLON IBL expr IBR
    |	_EXISTS_QUANTOR_ isElementOfSetListQB DOUBLECOLON IBL expr IBR
    |	SetExpr setRel SetExpr %prec _SUBSET_OF_
    |	expr BinRel expr %prec EQ
    |  	expr isOrNotElementOf SetExpr %prec _ELEMENT_OF_
    |	prefixLines
    |  	CHOOSE RBR_ON normalID _ELEMENT_OF_ SetExpr COLON expr RBR_OFF
    |  	'[' FUNCTION isElementOfSetListQB _MAPSTO_ expr ']'
    |	'[' CHANGE ID SEMICOLON expr _MAPSTO_ expr ']'
    |  	DSBR_ON recdef_csl DSBR_OFF
    |	DSBR_ON CHANGE ID SEMICOLON componentID _MAPSTO_ expr DSBR_OFF
    | 	SetExpr
    | 	expr BinOp expr %prec PLUS
    |   MINUS expr %prec UNARYMINUS
    | 	NATNUM
    |	StringExpr
    |  	expr DOT componentID 
    |	IF IBL expr IBR THEN IBL expr IBR ELSE IBL expr IBR
    |	CASE IBL CaseRA_list IBR
    |	LET IBL expr_list IBR IN IBL expr IBR
 
    /* LAexpr parts */

    | 	expr PRIME
    |	ENABLED RBR_ON ActionExpr RBR_OFF
    | 	UNCHANGED RBR_ON ID_csl RBR_OFF
    |	'[' ActionExpr ']' '_' RBR_ON ID_csl RBR_OFF
    |   LT expr GT '_' RBR_ON ID_csl RBR_OFF

    /* TLAexpr parts */

    |	_HENCEFORTH_ '[' StateFcnExpr ']' RBR_ON ActionExpr RBR_OFF
    |	_HENCEFORTH_ RBR_ON expr RBR_OFF
    |	_EVENTUALLY_ '[' StateFcnExpr ']' RBR_ON ActionExpr RBR_OFF
    |	_EVENTUALLY_ RBR_ON expr RBR_OFF
    |	_BOLD_EXISTS_ normalID_csl DOUBLECOLON IBL expr IBR
    |	expr _LEADSTO_ expr
    | 	WF '[' StateFcnExpr ']' RBR_ON ActionExpr RBR_OFF
    | 	SF '[' StateFcnExpr ']' RBR_ON ActionExpr RBR_OFF 
;

SetExpr: 
	ID
    |   OperatorApplication
    |   ID '[' constExpr_csl ']'
    |	ANY
    | 	RBR_ON SetExpr RBR_OFF
    |	SetExpr setOp SetExpr %prec _UNION_
    |	beginSet constExpr_csl '}'
    |	beginSet expr '|' isElementOfSetListQB COLON expr '}'
    |	beginSet expr '|' isElementOfSetListQB '}'
/*    |	beginSet isElementOfSetListQB COLON expr '}' ???          */
/*    |	beginSet expr COLON isElementOfSetListQB '}' alte Version */
    |	SUBSET RBR_ON SetExpr RBR_OFF
    | 	UNION RBR_ON SetExpr RBR_OFF
    | 	_EMPTYSET_
    |	predefSet
    |	SET '[' SetExpr _RIGHTARROW_ SetExpr ']'
    |	'[' CHANGE ID SEMICOLON expr COLON SetExpr ']'
    |	DOMAINt RBR_ON ID RBR_OFF
    |	DSBR_ON rctype_ssl DSBR_OFF
    |	DSBR_ON CHANGE ID SEMICOLON componentID COLON SetExpr DSBR_OFF
;

/*
 * the various MINOR expressions we need to construct the
 * symbol definitions within the various sections
 */

beginSet:
	'{'
;

expr_list:
	SymbolDef_LHS DEFINED_AS expr SEMICOLON
    |	expr_list SymbolDef_LHS DEFINED_AS expr SEMICOLON
;

predefSet: 
	NAT
;

setRel: 
    	_SUBSET_OF_
;

setOp: 
     	_UNION_
    | 	_INTERSECT_
    |	_SET_DIFFERENCE_
    |	_CROSS_
;

BinOp: 
	PLUS
    |	MINUS
    |	TIMES
    |   DIVIDED
    |   DIV
    |   MOD
    |   POWER
;

StringExpr: 
	STRING
;

isOrNotElementOf: 
	_ELEMENT_OF_
    |	_NOT_ELEMENT_OF_
;

recdef_csl: 
	recdef
    |	recdef_csl COMMA recdef
;

recdef:	
	componentID _MAPSTO_ expr
;

rctype_ssl: 
    	rctype
    |	rctype_ssl SEMICOLON rctype
;

rctype: 
    	normalID_csl COLON SetExpr
;

boolRel: 
	_L_AND_
    |	_L_OR_
;

boolRelIB: 
    	_IMPLIES_
    |	_EQUIV_
;

BinRel: 
	  EQ
        | NE
        | LT
        | GT
        | LE
        | GE
;

OperatorApplication: 
	qualifierPrefixes normalID RBR_ON expr_csl RBR_OFF
                                                        %prec PRECLOWERTHANDOT
;

prefixLines: 
	conjunctLines
    |	disjunctLines
;

conjunctLines: 
	PREF_AND IBL expr IBR	%prec PREF_AND
    | 	conjunctLines PREF_AND IBL expr IBR
;

disjunctLines: 
	PREF_OR IBL expr IBR     %prec PREF_OR
    | 	disjunctLines PREF_OR IBL expr IBR
;

isElementOfSetListQB:
    	normalID_csl _ELEMENT_OF_ SetExpr
    |	isElementOfSetListQB SEMICOLON normalID_csl _ELEMENT_OF_ SetExpr
;

CaseRA_list: 
    	CaseRA
    |	CaseRA_list CaseRA
;

CaseRA:	
	expr _RIGHTARROW_ expr SEMICOLON
;

/*
 * function and operator definition 
 */

SymbolDef_LHS: 
        normalID TypeAndSortDecl01
    |	normalID RBR_ON formalPar_ssl RBR_OFF TypeAndSortDecl01
    |  	normalID '[' formalPar_ssl ']' TypeAndSortDecl01
    |	infix '_' formalPar '_' normalID '_' formalPar '_' TypeAndSortDecl01
;

formalPar_ssl: 
        formalPar
    |	formalPar_ssl SEMICOLON formalPar
;

formalPar: 
	normalID_csl TypeAndSortDecl01
    |  	normalID RBR_ON formalPar_ssl RBR_OFF TypeAndSortDecl01
    |	infix '_' formalPar '_' normalID '_' formalPar '_' TypeAndSortDecl01
;

infix: 
	INFIX
    |	INFIXR
;

/*
 * Typen/Sortenangaben
 */

TypeAndSortDecl01: 
	/* empty */
    |	COLON TypeAndSortDecl
;

TypeAndSortDecl: 
	TypeDecl
    |	TypeDecl Sort
    |	Sort
;

TypeDecl:
    	SetExpr
;

Sort: 
	CONSTANT
    |   BOOLEAN
    |   PREDICATE
    |   STATE FUNCTION
    |   ACTION
    |   TRANSITION FUNCTION
    |   TEMPORAL
;

/*
 *  Bezeichner
 */

ModuleName: 
	normalID
;
qualifyingID: 
	normalID
;
componentID: 
	normalID
;

ID: 
	qualifierPrefixes normalID %prec PRECLOWERTHANDOT
;

qualifierPrefixes: 
	/* empty */
   |	qualifierPrefixes normalID DOT				%prec DOT
   |	qualifierPrefixes normalID RBR_ON expr_csl RBR_OFF DOT	%prec DOT
;

normalID_csl: 
	normalID
    |	normalID_csl COMMA normalID
;

ID_csl: 
	ID
    |	ID_csl COMMA ID
;

%%

static void HandleNewNode(SymListElem *selem, char EdgeType)
	/* Neuen Knoten mit Typ in Adjazenzliste eintragen */
 {
 Node* n = the_AdjList->AddNode(selem->Getlexem());	// evt. neuer Knoten
 this_module->AddEdge(n,EdgeType);			// neue Kante
 if (n->GetCounter()==0)
   {
   n->SetCounter(1);		// Wenn Knoten neu: als bekannt markieren
   NodeQueue->Append(n);	// und in Warteschlange eintragen
   }
 }

// Exporte

void yyerror (char* s)	// Fehler in Pass 1 ignorieren
 {
 }

void parse_pass1 (char* filename)
	/* Pass-1-Parser Hauptfunktion */
 {
 FILE* 	ifp = 0;
 char *dir,*libpath = NULL; 
 

 ifp = fopen(filename, "r");
 if (ifp == NULL)
   { 
   dir = getenv("TLALIB");

   if(dir)
    {
      libpath = new char[strlen(dir) + strlen(filename) + 2];

      sprintf(libpath,"%s/%s",dir,filename);

      ifp = fopen(libpath, "r");
      if(ifp == NULL)
       {
          delete libpath;  
          fprintf(stderr, "Can't open specification file '%s'.\n",filename);
          fflush(stderr);
          return;
       }
    }
   else
    {
     fprintf(stderr, "Can't open specification file '%s'.\n",filename);
     fflush(stderr);
     return;
    }
   }
 reinitialize_lex(ifp,NULL);
 yyparse();
 delete sList;

 }
