/* SATM REGION START */ 
/*****************************************************************************

	TLA+ Modul - BNF

	(E) Version 2.0         9.7.92		Mester
   Alle Rechte gem. dt. UrhG und WissHG vorbehalten.
   Jedwede Verwendung nur mit schriftlicher Genehmigung des Autors.
	design decisions are appended as comment to the 
	end of this file

	2.0: - adaption to 'a grammar for TLA+' as of May, 26.  
	     - modifications to remove reduce/reduce-errors
	     - adapting Leslies comments
	     THESE STEPS HAVE NOT BEEN DOCUMENTED HITHERTO !

 *****************************************************************************/
/* SATM REGION END */

/*****************************************************************************
 * TLA+ Browser : Praeprozessor - Parser Pass 2
 *
 * Michael Geikowski, 19.7.93
 *
 * ( without design decisions )
 *****************************************************************************/

%{

/***
 *** C declarations section
 ***/

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include "symlist.h"
#include "sort.h"
#include "env.h"
#include "crossref.h"
#include "dbinterface.h"
// om->Parsetree erzeugen
#include "parsetree.h"
// om->Typueberpruefung 
#include "typecheck.h"
// om->Parsetree in Adj.-Listenform darstellen
#include "adjlistgnu.h"
#include "semconcept.h"
// Ersatz des Headers fuer scanner.l



extern char* yytext;
extern int   writeit;
extern SymList* sList;
extern SymListElem* lastElem;

extern void reinitialize_lex (FILE* ifp, char* filename);
extern int  yylex();




// Globale Variablen

CrossRef* crossref;
ParseExpr *Parsetree; // om->Zeiger auf den Parsetree eines ge'parse'ten Moduls
// Modulglobale Variablen

static Environment* Env;
static Environment* UnknownIDEnv;
static int    ModuleDepth;
static int    isTypeDecl;
static int    isHeader;
static int    UnknownID_allowed;
static SymList*     ExportList;
static SymList*     IncludeList;
static SymListElem* lastIncludeName;
static SymListElem* dummy_selem;
static char* actfilename;

// om->...fuer den Parsetree:

static SmStack *SemanticStack; // Zeiger auf Semantischen Stack



// einige forward Deklarationen

extern void yyerror (char* msg);
static void startframe(int depth);
static void endframe(SymListElem* dest);
static void writeframe(int depth, char* s1, char* s2, SymListElem* dest);
static void act_Recover();
static SymList* act_Import(SymListElem *ModName, SymListElem *qualifier); // Rueckgabe von SymList->om
static void act_Include(SymListElem *ModName, SymListElem *qualifier);
static Sort* SortOfID(SymListElem* ID);

%}

/***
 *** Bison declarations section
 ***/

%expect 2
/* Zwei s/r Konflikte:
 * 1)
 *      ( SetExpr . )  >>>reduce>>>  ( expr . ) 
 *                     >>>shift>>>>  ( SetExpr ) .
 *   Hier ist shiften gewuenscht : Parser verhaelt sich somit korrekt.
 *
 * 2)
 *      ( expr . )  >>>reduce>>>  ( constExpr_csl . ) 
 *                  >>>shift>>>>  ( expr ) .
 *   Hier ist shiften gewuenscht : Parser verhaelt sich somit korrekt.
 * Im zweiten Fall drueckt sich der Konflikt zwischen gklammerten Ausdruecken
 * (shift) und einelementigen Tupeln (reduce) aus. Nach Lamport sind aber
 * einelementige Tupel nicht erlaubt. 
 */

%union
{ SymListElem *selem;
  SymList     *slist;
  Sort        *genSort;
  Sort_Type   primSort;
}

		/* TLA+ reserved names */
%token <selem>
    ANY
    NAT

/* TLA+ directives */

    INFIX		/* left-associatve infix   */
    INFIXR		/* right-associative infix */

/* Relatoren */

    EQ		NE	LT	GT 	LE  	GE
		
/* :: Ersatzdarstellungen mathematischer Symbole */

    _ALL_QUANTOR_	_EXISTS_QUANTOR_	_BOLD_EXISTS_
    _L_NOT_		_L_OR_			_L_AND_
    PREF_AND		PREF_OR		/* for use by lexical analysis	*/
    IBL			IBR		/* dito, invisible braces	*/
    PRIME
    _ELEMENT_OF_	_NOT_ELEMENT_OF_	
    _SUBSET_OF_ 	_SET_DIFFERENCE_	_UNION_		_INTERSECT_
    _EMPTYSET_		
    _IMPLIES_		_EQUIV_		
    _LEADSTO_		_HENCEFORTH_		_EVENTUALLY_	
    _CROSS_   				/* X */
    _MAPSTO_		_RIGHTARROW_

/* TLA+ keywords */

    MODULE
    PARAMETERS	AS			 
    EXPORT	OPAQUE	IMPORT	INCLUDE
    END						/* of one module */
    IF		THEN	ELSE 	
    CASE	
    FALSE	TRUE
    SF		WF		
    IN		SUBSET
    ENABLED	LET		
    CHOOSE	UNCHANGED
    UNION
    DOMAINt	/* read 'domain token';	DOMAIN is already defined in math.h */
    REPLACE		WITH
    DEFINED_AS		MINUS		PLUS
    DIV			MOD		POWER
    DIVIDED		TIMES
   
/* the sorts */

    CONSTANT
    BOOLEAN
    PREDICATE
    FUNCTION				/* for STATE / TRANSITION */
    ACTION
    TEMPORAL 

/* the 'sections' */

    CONSTANTS	BOOLEANS		
    PREDICATES  STATE	TRANSITION	FUNCTIONS 	ACTIONS	
    TEMPORALS
    ASSUMPTIONS		THEOREMS

/* Identifier */

	normalID	/* usual identifier w upper/lower case  */
			/* (also used as filename)		*/ 

/* Basisdatentypen */

	STRING
	NATNUM 

		/* Trenner */

	DOT
	DOUBLECOLON
	COLON
	SEMICOLON
	COMMA
	RBR_ON		RBR_OFF		/* RBR round brackets ()             */
	DSBR_ON		DSBR_OFF	/* D/SBR (double) square brackets [] */

/* syntactic sugar / additional tokens to avoid ambiguity */

	 CHANGE
	 SET
	 TUPELSELECTOR

/* associativity and precedence */

%left	_LEADSTO_
%left	_IMPLIES_ _EQUIV_
%left	_L_AND_	_L_OR_
%left	EQ NE LT GT LE GE _SUBSET_OF_
%left	PLUS MINUS TIMES DIVIDED DIV MOD POWER
%left	_SET_DIFFERENCE_ _UNION_ _INTERSECT_ _CROSS_
%left	_ELEMENT_OF_ _NOT_ELEMENT_OF_
%left   UNARYMINUS
%left   PRECLOWERTHANDOT
%left	TUPELSELECTOR DOT
%left   PRIME

/* typ-Definitionen */

%token	<selem>	'_'	'['	']'	'{'	'}'	'|'

%type   <selem>
	ModuleName	componentID	qualifyingID	ID
	CM_Definition	TM_Definition	ModuleEnd
	infix		boolRel		boolRelIB	BinRel 
	setOp		BinOp		StringExpr	predefSet
	setRel
	isOrNotElementOf		SymbolDef_LHS
	IncludePart	HeaderPart	Restexport	IncludeEntry
	SubstitutionPart01		BodyPart
%type   <slist>
	normalID_csl	ID_csl		qualifierPrefixes
	Export_qualifierPrefixes
%type	<primSort>
	Sort
%type	<genSort>
	TypeAndSortDecl		TypeAndSortDecl01
	formalPar		formalPar_ssl
	SetExpr			expr			expr_csl
	ConstantExpr		constExpr_csl
	BooleanExpr		StateFcnExpr		PredExpr
	TransFctExpr		ActionExpr
	prefixLines		conjunctLines		disjunctLines
	rctype_ssl		rctype
	recdef_csl		recdef
	CaseRA_list		CaseRA
	OperatorApplication	isElementOfSetListQB
	IncludeParameters01
%%

/***
 *** grammar rules section
 ***
 *** Explanations:
 *** The name of the transitions inform the reader about what they do.
 *** Names given in the singular form represent one occurrence in the BNF.
 *** Transitions ending in an plural 's' mean arbitrary occurrence, while
 *** an added '_' denotes at least one occurrence of the transition. An
 *** added '01' means zero or one occurrence of the according transition.
 *** 
 *** a 'csl' should denote a comma separated list / am
 *** a 'ssl' denotes a semicolon separated list / am
 *** a 'bsl' denotes a blank separated list / am
 ***/

OneModule:
	ModuleType MODULE ModuleName01
        { char* s=0;	// Kopf ins Inhaltsverzeichnis
          if (Env->GetModuleName())
            s=Env->GetModuleName()->Getlexem();
          if (Env->GetModuleType()==mt_temporal)
            writeframe(0,"TEMPORAL MODULE",s,$2);
          else
            writeframe(0,"CONSTANT MODULE",s,$2);
	  isHeader = 1;}
	HeaderParts 
	BodyParts
	ModuleEnd
        { Parsetree = new ParseExpr(tOneModule,"OneModule",6);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,6);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,5);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,4);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,3);
          AppendtoList($2,Parsetree,"MODULE",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));
        }
;

ModuleType:
	/* empty */
	{ ModuleDepth++;
	  if (ModuleDepth == 1)
	    Env = new Environment(s_statefct, mt_temporal, actfilename);
	  else {					// Untermodul beginnt
            if (Env->GetModuleType() == mt_const)
              SimpleError("temporal submodule in constant module not allowed",
                          lastElem->pos,lastElem->leng);
	    Env->StartSubModule(s_statefct, mt_temporal);
            UnknownIDEnv->Push(); }
          Parsetree = new ParseExpr(tModuleType,"ModuleType",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree))}
     |  CONSTANT
	{ ModuleDepth++;
	  if (ModuleDepth == 1)
	    Env = new Environment(s_const, mt_const, actfilename);
	  else {					// Untermodul beginnt
	    Env->StartSubModule(s_const, mt_const);
            UnknownIDEnv->Push(); }
          Parsetree = new ParseExpr(tModuleType,"ModuleType",1);
          AppendtoList($1,Parsetree,"CONSTANT",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
     |	TEMPORAL
	{ ModuleDepth++;
	  if (ModuleDepth == 1)
	    Env = new Environment(s_statefct, mt_temporal, actfilename);
	  else {					// Untermodul beginnt
            if (Env->GetModuleType() == mt_const)
              SimpleError("temporal submodule in constant module not allowed",
                          lastElem->pos,lastElem->leng);
	    Env->StartSubModule(s_statefct, mt_temporal);
            UnknownIDEnv->Push(); } 
          Parsetree = new ParseExpr(tModuleType,"ModuleType",1);
          AppendtoList($1,Parsetree,"TEMPORAL",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));}
;

ModuleName01:
	/* empty */
	{ if (ModuleDepth == 1)
	    SimpleError("missing module name",lastElem->pos,lastElem->leng);
	  Env->SetModuleName(0);
          Parsetree = new ParseExpr(tModuleName01,"ModuleName01",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
     |	ModuleName
	{ $1->Set_nidt(nidt_moduleName);
	  Env->SetModuleName($1);
          Parsetree = new ParseExpr(tModuleName01,"ModuleName01",1);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
     ;

HeaderParts:
	/* empty */
       {Parsetree = new ParseExpr(tHeaderParts,"HeaderParts", 1);
        SemanticStack->SmStackPush(new SmStackElem(Parsetree));}
     |	HeaderParts HeaderPart
        { Parsetree = new ParseExpr(tHeaderParts,"HeaderParts",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));}
     |  HeaderParts error SEMICOLON
	{ act_Recover();
          Parsetree = new ParseExpr(tHeaderParts,"HeaderParts",3);
          AppendtoList($3,Parsetree,"SEMICOLON",3);
          AppendtoList(new SymListElem(lastElem->pos,lastElem->leng,lineno, 999) ,Parsetree,"error",2); // error -> tokenID = 999
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
;

HeaderPart: /* <selem> (dummy) */ 
    	IMPORT ImportList SEMICOLON  
        { Parsetree = new ParseExpr(tHeaderPart,"HeaderPart",3);
          AppendtoList($3,Parsetree,"SEMICOLON",3);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,2);
          AppendtoList($1,Parsetree,"IMPORT",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));}
    |	EXPORT OPAQUE Restexport
	{ if (ModuleDepth > 1)
	    SimpleError("export in submodule not allowed",$1->pos,$1->leng);
          Parsetree = new ParseExpr(tHeaderPart,"HeaderPart",3);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,3);
          AppendtoList($2,Parsetree,"OPAQUE",2);
          AppendtoList($1,Parsetree,"EXPORT",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));}
    |	EXPORT Export_ID_csl OPAQUE Restexport
	{ if (ModuleDepth > 1)
	    SimpleError("export in submodule not allowed",$1->pos,$1->leng);
          Parsetree = new ParseExpr(tHeaderPart,"HeaderPart",4);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,4);
           AppendtoList($3, Parsetree,"OPAQUE",3);
           PopandAppend(SemanticStack->SmStackPop(),Parsetree,2);
           AppendtoList($1, Parsetree,"EXPORT",1);
           SemanticStack->SmStackPush(new SmStackElem(Parsetree));}
    |	EXPORT Export_ID_csl SEMICOLON
	{ if (ModuleDepth > 1)
	    SimpleError("export in submodule not allowed",$1->pos,$1->leng);
          Parsetree = new ParseExpr(tHeaderPart,"HeaderPart",3);
          AppendtoList($3, Parsetree,"SEMICOLON",3);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,2);
          AppendtoList($1, Parsetree,"EXPORT",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));}
    |	PARAMETERS formalPar_ssl SEMICOLON
	{ delete $2;
          Parsetree = new ParseExpr(tHeaderPart,"HeaderPart",3);
          AppendtoList($3, Parsetree,"SEMICOLON",3);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,2);
          AppendtoList($1, Parsetree,"PARAMETERS",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));}
;

Export_ID:
	Export_qualifierPrefixes normalID
	{ if ($2->is_primed())
	    SimpleError("primed identifier not allowed here",$2->pos,$2->leng);
	  if (ModuleDepth==1) {
	    $2->SetQualifier($1);
	    ExportList->Append($2); }
          else				// Export in Untermodul nicht erlaubt
            delete $1;
          Parsetree = new ParseExpr(tExport_ID,"Export_ID",2);
          AppendtoList($2, Parsetree,"normalID",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));}
;

Export_qualifierPrefixes: /* <slist> (Liste der Qualifizierer) */
	/* empty */
	{ $$ = new SymList;
         Parsetree = new ParseExpr(tExport_qualifierPrefixes,"Export_qualifierPrefixes",1);
         SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
   |	Export_qualifierPrefixes normalID DOT
	{ if ($2->is_primed())
	    SimpleError("primed identifier not allowed here",$2->pos,$2->leng);
	  $1->Append($2);
	  $$ = $1;
          Parsetree = new ParseExpr(tExport_qualifierPrefixes,"Export_qualifierPrefixes",3);
          AppendtoList($3, Parsetree,"DOT",3);
          AppendtoList($2, Parsetree,"normalID",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));} 
   |	Export_qualifierPrefixes normalID RBR_ON normalID_csl RBR_OFF DOT
	{ if ($2->is_primed())
	    SimpleError("primed identifier not allowed here",$2->pos,$2->leng);
	  $1->Append($2);
	  $2->SetQualifier($4);
	  $$ = $1;
          Parsetree = new ParseExpr(tExport_qualifierPrefixes,"Export_qualifierPrefixes",6);
          AppendtoList($6, Parsetree,"DOT",6);
          AppendtoList($5, Parsetree,"RBR_OFF",5);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,4);
          AppendtoList($3, Parsetree,"RBR_ON",3);
          AppendtoList($2, Parsetree,"normalID",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));}
;

Export_ID_csl:
	Export_ID
        { Parsetree = new ParseExpr(tExport_ID_csl,"Export_ID_csl",1);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));}
    |	Export_ID_csl COMMA Export_ID
        { Parsetree = new ParseExpr(tExport_ID_csl,"Export_ID_csl",3);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,3);
          AppendtoList($2, Parsetree,"COMMA",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));}
;

Restexport: /* <selem> */
	Export_ID_csl SEMICOLON
        { $$ = $2;
          Parsetree = new ParseExpr(tRestexport,"Restexport",2);
          AppendtoList($2, Parsetree,"SEMICOLON",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));}
    |	SEMICOLON
        { Parsetree = new ParseExpr(tRestexport,"Restexport",1);
          AppendtoList($1, Parsetree,"SEMICOLON",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));}
;

BodyParts:
	/* empty */
        { isHeader = 0;
          Parsetree = new ParseExpr(tBodyParts,"BodyParts",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |	BodyParts BodyPart
        { Parsetree = new ParseExpr(tBodyParts,"BodyParts",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));}
;

BodyPart: /* <selem> (dummy) */
    	IncludeEntry
	{ $$ = (SymListElem*) NULL;
          Parsetree = new ParseExpr(tBodyPart,"BodyPart",1);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));}
    |   CM_Definition
	{ $$ = (SymListElem*) NULL;
          Parsetree = new ParseExpr(tBodyPart,"BodyPart",1);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));}
    |	TM_Definition
	{ if (Env->GetModuleType() == mt_const)
	    SimpleError("only constant definitions allowed in constant module"
                        ,$1->pos,$1->leng);  
	  $$ = (SymListElem*) NULL;
          Parsetree = new ParseExpr(tBodyPart,"BodyPart",1);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |	OneModule
	{ ModuleDepth--;			// Ende eines Untermoduls
	  Env->AppendAndPopSubModule();
          UnknownIDEnv->AppendRestUnknownIDs(Env);
          UnknownIDEnv->Pop(); 
	  $$ = (SymListElem*) NULL;
          Parsetree = new ParseExpr(tBodyPart,"BodyPart",1);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
;

CM_Definition: /* <selem> (dummy) */
		/* Neuen Abschnitt in Inhaltsverzeichnis vermerken und      */
                /* voreingestellte Sorten abhaengig vom Abschnitt bestimmen */
	CONSTANTS
	{ writeframe(1,"CONSTANTS",0,$1);
          Env->SetDefaultOpSort(s_const);
	  Env->SetDefaultParSort(s_const); }
	ConstantDefinitions
        { Parsetree = new ParseExpr(tCM_Definition,"CM_Definition",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,2);
          AppendtoList($1, Parsetree,"CONSTANTS",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |	BOOLEANS
	{ writeframe(1,"BOOLEANS",0,$1);
	  Env->SetDefaultOpSort(s_bool);
	  Env->SetDefaultParSort(s_const); }
	BooleanDefinitions
        {Parsetree = new ParseExpr(tCM_Definition,"CM_Definition",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,2);
          AppendtoList($1, Parsetree,"BOOLEANS",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |	CONSTANT ASSUMPTIONS
	{ writeframe(1,"CONSTANT ASSUMPTIONS",NULL,$1);
	  Env->SetDefaultOpSort(s_bool);
	  Env->SetDefaultParSort(s_const); }
	AssumptionDefinitions
	{ if (ModuleDepth > 1)
	    SimpleError("assumptions in submodule not allowed",
                        $2->pos,$2->leng);
          Parsetree = new ParseExpr(tCM_Definition,"CM_Definition",3);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,3);
          AppendtoList($2, Parsetree,"ASSUMPTIONS",2);
          AppendtoList($1, Parsetree,"CONSTANT",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
    |	ASSUMPTIONS
	{ writeframe(1,"ASSUMPTIONS",NULL,$1);
	  if (Env->GetModuleType() == mt_const) {
            Env->SetDefaultOpSort(s_bool);
	    Env->SetDefaultParSort(s_const); }
          else {
            Env->SetDefaultOpSort(s_action);
  	    Env->SetDefaultParSort(s_statefct); } }
	AssumptionDefinitions
	{ if (ModuleDepth > 1)
	    SimpleError("assumptions in submodule not allowed",
                         $1->pos,$1->leng);
          Parsetree = new ParseExpr(tCM_Definition,"CM_Definition",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,2);
          AppendtoList($1, Parsetree,"ASSUMPTIONS",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |	CONSTANT THEOREMS
	{ writeframe(1,"CONSTANT THEOREMS",NULL,$1);
 	  Env->SetDefaultOpSort(s_bool);
	  Env->SetDefaultParSort(s_const); }
	TheoremDefinitions
        { Parsetree = new ParseExpr(tCM_Definition,"CM_Definition",3);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,3);
          AppendtoList($2, Parsetree,"THEOREMS",2);
          AppendtoList($1, Parsetree,"CONSTANT",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |	THEOREMS
	{ writeframe(1,"THEOREMS",NULL,$1);
	  if (Env->GetModuleType() == mt_const) {
            Env->SetDefaultOpSort(s_bool);
	    Env->SetDefaultParSort(s_const); }
          else {
            Env->SetDefaultOpSort(s_action);
	    Env->SetDefaultParSort(s_statefct); } }
	TheoremDefinitions
        { Parsetree = new ParseExpr(tCM_Definition,"CM_Definition",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,2);
          AppendtoList($1, Parsetree,"THEOREMS",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
;

TM_Definition: /* <selem> (dummy) */
		/* Neuen Abschnitt in Inhaltsverzeichnis vermerken und      */
                /* voreingestellte Sorten abhaengig vom Abschnitt bestimmen */
    	PREDICATES
	{ writeframe(1,"PREDICATES",NULL,$1);
	  Env->SetDefaultOpSort(s_pred);
	  Env->SetDefaultParSort(s_statefct); }
	PredicateDefinitions
	{ $$ = $1;
          Parsetree = new ParseExpr(tTM_Definition,"TM_Definition",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,2);
          AppendtoList($1, Parsetree,"PREDICATES",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |	STATE FUNCTIONS
	{ writeframe(1,"STATE FUNCTIONS",NULL,$1);
	  Env->SetDefaultOpSort(s_statefct);
	  Env->SetDefaultParSort(s_statefct); }
	StateFunctionDefinitions
	{ $$ = $1;
          Parsetree = new ParseExpr(tTM_Definition,"TM_Definition",3);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,3);
          AppendtoList($2, Parsetree,"FUNCTIONS",2);
          AppendtoList($1, Parsetree,"STATE",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |	ACTIONS
	{ writeframe(1,"ACTIONS",NULL,$1);
          Env->SetDefaultOpSort(s_action);
	  Env->SetDefaultParSort(s_statefct); }
	ActionDefinitions
	{ $$ = $1;
          Parsetree = new ParseExpr(tTM_Definition,"TM_Definition",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,2);
          AppendtoList($1, Parsetree,"ACTIONS",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |	TRANSITION FUNCTIONS
	{ writeframe(1,"TRANSITION FUNCTIONS",NULL,$1);
	  Env->SetDefaultOpSort(s_transfct);
	  Env->SetDefaultParSort(s_statefct); }
	TransitionFunctionDefinitions
	{ $$ = $1;
          Parsetree = new ParseExpr(tTM_Definition,"TM_Definition",3);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,3);
          AppendtoList($2, Parsetree,"FUNCTIONS",2);
          AppendtoList($1, Parsetree,"TRANSITION",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |	TEMPORALS
	{ writeframe(1,"TEMPORALS",NULL,$1);
	  Env->SetDefaultOpSort(s_temporal);
	  Env->SetDefaultParSort(s_statefct); }
	TemporalDefinitions
	{ $$ = $1;
          Parsetree = new ParseExpr(tTM_Definition,"TM_Definition",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,2);
          AppendtoList($1, Parsetree,"TEMPORALS",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |	ACTION ASSUMPTIONS
	{ writeframe(1,"ACTION ASSUMPTIONS",NULL,$1);
	  Env->SetDefaultOpSort(s_action);
	  Env->SetDefaultParSort(s_statefct); }
	AssumptionDefinitions
	{ if (ModuleDepth > 1)
	    SimpleError("assumptions in submodule not allowed",
                         $2->pos,$2->leng);
	  $$ = $1;
          Parsetree = new ParseExpr(tTM_Definition,"TM_Definition",3);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,3);
          AppendtoList($2, Parsetree,"ASSUMTIONS",2);
          AppendtoList($1, Parsetree,"ACTION",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |	ACTION THEOREMS
	{ writeframe(1,"ACTION THEOREMS",NULL,$1);
	  Env->SetDefaultOpSort(s_action);
	  Env->SetDefaultParSort(s_statefct); }
	TheoremDefinitions
	{ $$ = $1;
          Parsetree = new ParseExpr(tTM_Definition,"TM_Definition",3);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,3);
          AppendtoList($2, Parsetree,"THEOREMS",2);
          AppendtoList($1, Parsetree,"ACTION",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |	TEMPORAL THEOREMS
	{ writeframe(1,"TEMPORAL THEOREMS",NULL,$1);
	  Env->SetDefaultOpSort(s_temporal);
	  Env->SetDefaultParSort(s_statefct); }
	TheoremDefinitions
	{ $$ = $1;
          Parsetree = new ParseExpr(tTM_Definition,"TM_Definition",3);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,3);
          AppendtoList($2, Parsetree,"THEOREMS",2);
          AppendtoList($1, Parsetree,"TEMPORAL",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
;

ModuleEnd: /* <selem> */
		/* Eintrag ins Inhaltsverzeichnis und Vergleich von */
		/* Modulname an Kopf und Ende			    */
	END
        { writeframe(0,"END",NULL,$1);
          $$ = $1;
          Parsetree = new ParseExpr(tModuleEnd,"ModuleEnd",1);
          AppendtoList($1, Parsetree,"END",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |	END ModuleName
	{ writeframe(0,"END",$2->Getlexem(),$1);
          $2->Set_nidt(nidt_moduleName);
	  if ( Env->GetModuleName() ) {
            if (! $2->lexemcmp(Env->GetModuleName()) )
  	      SimpleError("different module names at header and end",
                          $2->pos,$2->leng);
            else
              $2->SetDefinedAt(Env->GetModuleName()); }
          $$ = $1;
          Parsetree = new ParseExpr(tModuleEnd,"ModuleEnd",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,2);
          AppendtoList($1, Parsetree,"END",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
;

   /* 
    * definitions for the import-mechanism in a module header, ok
    */ 

ImportList:
	ModuleName
	{ SymList *liste = act_Import($1,0);

          Parsetree = new ParseExpr(tImportList,"ImportList",1);
          Parsetree->AddIncludes(liste);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
    |	ModuleName AS qualifyingID
	{ SymList *liste = act_Import($1,$3); 

          Parsetree = new ParseExpr(tImportList,"ImportList",3);
          Parsetree->AddIncludes(liste);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,3);
          AppendtoList($2, Parsetree,"AS",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    | 	ImportList COMMA ModuleName
	{ SymList *liste = act_Import($3,0);

          Parsetree = new ParseExpr(tImportList,"ImportList",3);
          Parsetree->AddIncludes(liste);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,3);
          AppendtoList($2, Parsetree,"COMMA",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |	ImportList COMMA ModuleName AS qualifyingID
	{ SymList *liste = act_Import($3,$5);

          Parsetree = new ParseExpr(tImportList,"ImportList",5);
          Parsetree->AddIncludes(liste);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,5);
          AppendtoList($4, Parsetree,"AS",4);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,3);
          AppendtoList($2, Parsetree,"COMMA",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
;

IncludeEntry: /* <selem> (dummy) */
	INCLUDE IncludeParts
        { Parsetree = new ParseExpr(tIncludeEntry,"IncludeEntry",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,2);
          AppendtoList($1, Parsetree,"INCLUDE",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
;

IncludeParts:
	IncludePart SubstitutionPart01 SEMICOLON
        { UnknownIDEnv->AppendAndPopUnknownIDs(isHeader,Env);
          Env->Pop();
          if (IncludeList) {	// Neue Bezeichner in Umgebung eintragen
            if ($1) Env->Append($1);
	    Env->Append(IncludeList,1,lastIncludeName); }
          Parsetree = new ParseExpr(tIncludeParts,"InculdeParts",3);
          AppendtoList($3, Parsetree,"SEMICOLON",3);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |	IncludeParts IncludePart SubstitutionPart01 SEMICOLON
	{ UnknownIDEnv->AppendAndPopUnknownIDs(isHeader,Env);
	  Env->Pop();
          if (IncludeList) {	// Neue Bezeichner in Umgebung eintragen
            if ($2) Env->Append($2);
	    Env->Append(IncludeList,1,lastIncludeName); }
          Parsetree = new ParseExpr(tIncludeParts,"IncludeParts",4);
          AppendtoList($4, Parsetree,"SEMICOLON",4);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,3);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
    ;

IncludePart: /* <selem> (eventuell Qualifizierer) */
    	ModuleName
	{ Env->Push();		// lokale Umgebung fuer Substitutionsteil
          UnknownIDEnv->Push();
          act_Include($1,0);
          lastIncludeName=$1;
	  $$ = 0;

          Parsetree = new ParseExpr(tIncludePart,"InculdePart",1);
          Parsetree->AddIncludes(IncludeList);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));     }
    |   ModuleName AS qualifyingID IncludeParameters01
	{ act_Include($1,$3);
          lastIncludeName=$1;
	  $3->SetSort($4);
          $3->Set_nidt(nidt_qualifier);
	  $$ = $3;

          Parsetree = new ParseExpr(tIncludePart,"IncludePart",4);
          Parsetree->AddIncludes(IncludeList);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,4);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,3);
          AppendtoList($2, Parsetree,"AS",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
;

IncludeParameters01: /* <genSort> (Liste der Parametersoreten des optionalen*/
		     /*            Qualifizierers abgesclossen mit s_const  */
	/* empty */
	{ Env->Push();		// lokale Umgebung fuer Substitutionsteil
          UnknownIDEnv->Push();
	  $$ = new Sort(s_const);
          Parsetree = new ParseExpr(tIncludeParameters01,"IncludeParameters01",1);
          SemanticStack->SmStackPush( new SmStackElem(Parsetree)); }
    |	RBR_ON
	{ Env->Push();		// lokale Umgebung fuer Substitutionsteil
          UnknownIDEnv->Push();
	  Env->SetDefaultSort(Env->GetDefaultParSort()); }
	formalPar_ssl RBR_OFF
	{ $3->Append(new Sort(s_const));
	  $$ = $3;
          Parsetree = new ParseExpr(tIncludeParameters01,"IncludeParameters01",3);
          AppendtoList($4, Parsetree,"RBR_OFF",3);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,2);
          AppendtoList($1, Parsetree,"RBR_ON",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
;

SubstitutionPart01: /* <selem> (dummy) */
	/*empty */
	{ $$ = (SymListElem*) NULL;
          Parsetree = new ParseExpr(tSubstitutionPart01,"SubstitutionPart01",1);
          SemanticStack->SmStackPush( new SmStackElem(Parsetree)); }
    |	WITH Substitution_ 
	{ $$ = (SymListElem*) NULL;
          Parsetree = new ParseExpr(tSubstitutionPart01,"SubstitutionPart01",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,2);
          AppendtoList($1, Parsetree,"WITH",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
;

Substitution_:
	OneSubstitution_
        { Parsetree = new ParseExpr(tSubstitution_,"Substitution_",1);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |	Substitution_ COMMA OneSubstitution_ 
        { Parsetree = new ParseExpr(tSubstitution_,"Substitution_",3);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,3);
          AppendtoList($2, Parsetree,"COMMA",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
;

OneSubstitution_:
	normalID REPLACE expr
	/* Test auf korrekte Substitution */
	{ if ($1->is_primed())
            SimpleError("prime not allowed here",$1->pos,$1->leng);
          if (IncludeList) {
          SymList *dummy = new SymList;
	  SymListElem *e = IncludeList->Lookup($1,dummy);
	  delete dummy;
	  if (!e) { SimpleError("unknown identifier",$1->pos,$1->leng);
		    $1->SetSort(new Sort(s_const)); }
	  else if (e->Get_nidt()==nidt_substPar) {
		  SimpleError("parameter substituted twice",$1->pos,$1->leng);
		  $1->SetSort(new Sort(s_const)); }
	  else if ( (e->Get_nidt()!=nidt_formalPar) &&
		    (e->Get_nidt()!=nidt_formalglobPar)  ) {
		  SimpleError("only parameters can be substituted",
                              $1->pos,$1->leng);
		  $1->SetSort(new Sort(s_const)); }
	  else {
            if (! ( *($3) <= *(e->GetSort()) ) ) {
              SimpleError("replacing expression has illegal sort",
                          $2->pos,$2->leng);
              SimpleErrorCont("  used     : ");
              $3->ErrorOutPut();
              SimpleErrorCont("\n  expected : ");
              e->GetSort()->ErrorOutPut();
              SimpleErrorCont("\n"); }
	    $1->SetSort(e->GetSort()->copy());
            $1->SetDefinedAt(e);
	    e->Set_nidt(nidt_substPar); }
	  $1->Set_nidt(nidt_substPar);
	  delete $3; }
          Parsetree = new ParseExpr(tOneSubstitution_,"OneSubstitution_",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          AppendtoList($2,Parsetree,"REPLACE",2);
          AppendtoList($1,Parsetree,"normalID",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
;

/*
 * transitions for the various symbol-definition parts within a module body
 */

ConstantDefinitions:
    	/* empty */
        {Parsetree = new ParseExpr(tConstantDefinitions,"ConstantDefinitions", 1);
         SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |   ConstantDefinitions ConstantDefinition SEMICOLON
        { Parsetree = new ParseExpr(tConstantDefinitions,"ConstantDefinitions",3);
          AppendtoList($3,Parsetree,"SEMICOLON",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
    |   ConstantDefinitions error SEMICOLON
	{ act_Recover();
          Parsetree = new ParseExpr(tConstantDefinitions,"ConstantDefinitions",3);
          AppendtoList($3,Parsetree,"SEMICOLON",3);
          AppendtoList(new SymListElem(lastElem->pos,lastElem->leng,lineno,999),Parsetree,"error",2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
;

BooleanDefinitions:
	/* empty */
       {Parsetree = new ParseExpr(tBooleanDefinitions,"BooleanDefinitions", 1);
        SemanticStack->SmStackPush(new SmStackElem(Parsetree));}
    |   BooleanDefinitions BooleanDefinition SEMICOLON
        { Parsetree = new ParseExpr(tBooleanDefinitions,"BooleanDefinitions",3);
          AppendtoList($3,Parsetree,"SEMICOLON",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
    |   BooleanDefinitions error SEMICOLON
	{ act_Recover();
          Parsetree = new ParseExpr(tBooleanDefinitions,"BooleanDefinitions",3);
          AppendtoList($3,Parsetree,"SEMICOLON",3);
          AppendtoList(new SymListElem(lastElem->pos,lastElem->leng,lineno,999),Parsetree,"error",2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
;

PredicateDefinitions:
	/* empty */
        {Parsetree = new ParseExpr(tPredicateDefinitions,"PredicateDefinitions", 1);
         SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |   PredicateDefinitions PredicateDefinition SEMICOLON
        { Parsetree = new ParseExpr(tPredicateDefinitions,"PredicateDefinitions",3);
          AppendtoList($3,Parsetree,"SEMICOLON",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
    |   PredicateDefinitions error SEMICOLON
	{ act_Recover();
          Parsetree = new ParseExpr(tPredicateDefinitions,"PredicateDefinitions",3);
          AppendtoList($3,Parsetree,"SEMICOLON",3);
          AppendtoList(new SymListElem(lastElem->pos,lastElem->leng,lineno,999),Parsetree,"error",2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
;

StateFunctionDefinitions:
	/* empty */
        {Parsetree = new ParseExpr(tStateFunctionDefinitions,"StateFunctionDefinitions", 1);
         SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |   StateFunctionDefinitions StateFunctionDefinition SEMICOLON
        { Parsetree = new ParseExpr(tStateFunctionDefinitions,"StateFunctionDefinitions",3);
          AppendtoList($3,Parsetree,"SEMICOLON",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
    |   StateFunctionDefinitions error SEMICOLON
	{ act_Recover();
          Parsetree = new ParseExpr(tStateFunctionDefinitions,"StateFunctionDefinitions",3);
          AppendtoList($3,Parsetree,"SEMICOLON",3);
          AppendtoList(new SymListElem(lastElem->pos,lastElem->leng,lineno,999),Parsetree,"error",2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
;

ActionDefinitions:
	/* empty */
        {Parsetree = new ParseExpr(tActionDefinitions,"ActionDefinitions", 1);
         SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |   ActionDefinitions ActionDefinition SEMICOLON
        { Parsetree = new ParseExpr(tActionDefinitions,"ActionDefinitions",3);
          AppendtoList($3,Parsetree,"SEMICOLON",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
    |   ActionDefinitions error SEMICOLON
	{ act_Recover();
          Parsetree = new ParseExpr(tActionDefinitions,"ActionDefinitions",3);
          AppendtoList($3,Parsetree,"SEMICOLON",3);
          AppendtoList(new SymListElem(lastElem->pos,lastElem->leng,lineno,999),Parsetree,"error",2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
;

TransitionFunctionDefinitions:
	/* empty */
        {Parsetree = new ParseExpr(tTransitionFunctionDefinitions,"TransitionFunctionDefinitions",1);
         SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |   TransitionFunctionDefinitions TransitionFunctionDefinition SEMICOLON
        { Parsetree = new ParseExpr(tTransitionFunctionDefinitions,"TransitionFunctionDefinitions",3);
          AppendtoList($3,Parsetree,"SEMICOLON",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
    |   TransitionFunctionDefinitions error SEMICOLON
	{ act_Recover();
          Parsetree = new ParseExpr(tTransitionFunctionDefinitions,"TransitionFunctionDefinitions",3);
          AppendtoList($3,Parsetree,"SEMICOLON",3);
          AppendtoList(new SymListElem(lastElem->pos,lastElem->leng,lineno,999),Parsetree,"error",2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));    }
;

AssumptionDefinitions:
	/* empty */
        { Parsetree = new ParseExpr(tAssumptionDefinitions,"AssumptionDefinitions",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));}
    |   AssumptionDefinitions AssumptionDefinition SEMICOLON
        { Parsetree = new ParseExpr(tAssumptionDefinitions,"AssumptionDefinitions",3);
          AppendtoList($3, Parsetree,"SEMICOLON",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));}
    |   AssumptionDefinitions error SEMICOLON
	{ act_Recover();
          Parsetree = new ParseExpr(tAssumptionDefinitions,"AssumptionDefinitions",3);
          AppendtoList($3, Parsetree,"SEMICOLON",3);
          AppendtoList(new SymListElem(lastElem->pos,lastElem->leng,lineno,999),Parsetree,"error",2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));    }
;

TheoremDefinitions:
	/* empty */
        { Parsetree = new ParseExpr(tTheoremDefinitions,"TheoremDefinitions",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); } 
    |   TheoremDefinitions TheoremDefinition SEMICOLON
        { Parsetree = new ParseExpr(tTheoremDefinitions,"TheoremDefinitions",3);
          AppendtoList($3,Parsetree,"SEMICOLON",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
    |   TheoremDefinitions error SEMICOLON
	{ act_Recover();
          Parsetree = new ParseExpr(tTheoremDefinitions,"TheoremDefinitions",3);
          AppendtoList($3,Parsetree,"SEMICOLON",3);
          AppendtoList(new SymListElem(lastElem->pos,lastElem->leng,lineno,999),Parsetree,"error",2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));      }
;

TemporalDefinitions:
	/* empty */
        { Parsetree = new ParseExpr(tTemporalDefinitions,"TemporalDefinitions",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |   TemporalDefinitions TemporalDefinition SEMICOLON
        { Parsetree = new ParseExpr(tTemporalDefinitions,"TemporalDefinitions",3);
          AppendtoList($3,Parsetree,"SEMICOLON",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
    |   TemporalDefinitions error SEMICOLON
	{ act_Recover();
          Parsetree = new ParseExpr(tTemporalDefinitions,"TemporalDefinitions",3);
          AppendtoList($3,Parsetree,"SEMICOLON",3);
          AppendtoList(new SymListElem(lastElem->pos,lastElem->leng,lineno,999),Parsetree,"error",2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
;

/*
 * the real definitions
 *
 * Jeweils Inhaltsverzeichniseintrag und Sortenpruefung
 */

ConstantDefinition:
	{ startframe(2);} SymbolDef_LHS 
        { endframe($2); } DEFINED_AS ConstantExpr
	{ if (($2)->GetSort()->GetLastElem()->GetSort() != s_const)
	    SimpleError("defined symbol of illegal sort",$4->pos,$4->leng);
	  UnaryExactMatch($5,s_const,$4->pos,$4->leng);
	  delete $5;
          UnknownIDEnv->AppendAndPopUnknownIDs(isHeader,Env);
	  Env->Pop();
	  if (($2)->Get_nidt() != nidt_fctOperator) Env->Append($2);
          Parsetree = new ParseExpr(tConstantDefinition,"ConstantDefinition",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          AppendtoList($4,Parsetree,"DEFINED_AS",2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
;

BooleanDefinition:
	{ startframe(2);} SymbolDef_LHS 
	{ endframe($2); } DEFINED_AS BooleanExpr 
	{ if (($2)->GetSort()->GetLastElem()->GetSort() != s_bool)
	    SimpleError("defined symbol of illegal sort",$4->pos,$4->leng);
	  UnaryExactMatch($5,s_bool,$4->pos,$4->leng);
	  delete $5;
          UnknownIDEnv->AppendAndPopUnknownIDs(isHeader,Env);
	  Env->Pop();
	  if (($2)->Get_nidt() != nidt_fctOperator) Env->Append($2);
          Parsetree = new ParseExpr(tBooleanDefinition,"BooleanDefinition",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          AppendtoList($4,Parsetree,"DEFINED_AS",2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
;

PredicateDefinition:
	{ startframe(2);} SymbolDef_LHS 
	{ endframe($2); } DEFINED_AS PredExpr 
	{ if (($2)->GetSort()->GetLastElem()->GetSort() != s_pred)
	    SimpleError("defined symbol of illegal sort",$4->pos,$4->leng);
	  UnaryExactMatch($5,s_pred,$4->pos,$4->leng);
	  delete $5;
          UnknownIDEnv->AppendAndPopUnknownIDs(isHeader,Env);
	  Env->Pop();
	  if (($2)->Get_nidt() != nidt_fctOperator) Env->Append($2);
          Parsetree = new ParseExpr(tPredicateDefinition,"PredicateDefinition",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          AppendtoList($4,Parsetree,"DEFINED_AS",2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
;

StateFunctionDefinition:
	{ startframe(2);} SymbolDef_LHS 
	{ endframe($2); } DEFINED_AS StateFcnExpr 
	{ if (($2)->GetSort()->GetLastElem()->GetSort() != s_statefct)
	    SimpleError("defined symbol of illegal sort",$4->pos,$4->leng);
	  UnaryExactMatch($5,s_statefct,$4->pos,$4->leng);
	  delete $5;
          UnknownIDEnv->AppendAndPopUnknownIDs(isHeader,Env);
	  Env->Pop();
	  if (($2)->Get_nidt() != nidt_fctOperator) Env->Append($2);
          Parsetree = new ParseExpr(tStateFunctionDefinition,"StateFunctionDefinition",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          AppendtoList($4,Parsetree,"DEFINED_AS",2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
;

ActionDefinition:
	{ startframe(2);} SymbolDef_LHS 
	{ endframe($2); } DEFINED_AS ActionExpr 
	{ if (($2)->GetSort()->GetLastElem()->GetSort() != s_action)
	    SimpleError("defined symbol of illegal sort",$4->pos,$4->leng);
	  UnaryExactMatch($5,s_action,$4->pos,$4->leng);
	  delete $5;
          UnknownIDEnv->AppendAndPopUnknownIDs(isHeader,Env);
	  Env->Pop();
	  if (($2)->Get_nidt() != nidt_fctOperator) Env->Append($2);
          Parsetree = new ParseExpr(tActionDefinition,"ActionDefinition",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          AppendtoList($4,Parsetree,"DEFINED_AS",2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
;

TransitionFunctionDefinition:
	{ startframe(2);} SymbolDef_LHS 
	{ endframe($2); } DEFINED_AS TransFctExpr 
	{ if (($2)->GetSort()->GetLastElem()->GetSort() != s_transfct)
	    SimpleError("defined symbol of illegal sort",$4->pos,$4->leng);
	  UnaryExactMatch($5,s_transfct,$4->pos,$4->leng);
	  delete $5;
          UnknownIDEnv->AppendAndPopUnknownIDs(isHeader,Env);
	  Env->Pop();
	  if (($2)->Get_nidt() != nidt_fctOperator) Env->Append($2);
          Parsetree = new ParseExpr(tTransitionFunctionDefinition,"TransitionFunctionDefinition",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          AppendtoList($4,Parsetree,"DEFINED_AS",2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
;

AssumptionDefinition: 
	{ startframe(2);} SymbolDef_LHS 
	{ endframe($2); } DEFINED_AS expr
	{ if ( (Env->GetDefaultOpSort() == s_action) &&
               (($5)->GetLastElem()->GetSort()==s_pred)) {
             Env->SetDefaultOpSort(s_pred);
             ($2)->GetSort()->GetLastElem()->SetSort(s_pred); }
          if (($2)->GetSort()->GetLastElem()->GetSort()
	      != Env->GetDefaultOpSort())
	    SimpleError("defined symbol of illegal sort",$4->pos,$4->leng);
	  UnaryExactMatch($5,Env->GetDefaultOpSort(),$4->pos,$4->leng);
	  delete $5;
          UnknownIDEnv->AppendAndPopUnknownIDs(isHeader,Env);
	  Env->Pop();
	  if (($2)->Get_nidt() != nidt_fctOperator) Env->Append($2);
	  $2->Set_nidt(nidt_assumption); 
          Parsetree = new ParseExpr(tAssumptionDefinition,"AssumptionDefinition",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          AppendtoList($4,Parsetree,"DEFINED_AS",2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }  
;

TheoremDefinition:
	{ startframe(2);} SymbolDef_LHS 
	{ endframe($2); } DEFINED_AS expr 
	{ if ( (Env->GetDefaultOpSort() == s_action) &&
               (($5)->GetLastElem()->GetSort()==s_pred)) {
             Env->SetDefaultOpSort(s_pred);
             ($2)->GetSort()->GetLastElem()->SetSort(s_pred); }
          if (($2)->GetSort()->GetLastElem()->GetSort()
	      != Env->GetDefaultOpSort())
	    SimpleError("defined symbol of illegal sort",$4->pos,$4->leng);
	  UnaryExactMatch($5,Env->GetDefaultOpSort(),$4->pos,$4->leng);
	  delete $5;
          UnknownIDEnv->AppendAndPopUnknownIDs(isHeader,Env);
	  Env->Pop();
	  if (($2)->Get_nidt() != nidt_fctOperator) Env->Append($2);
	  $2->Set_nidt(nidt_theorem);
          Parsetree = new ParseExpr(tTheoremDefinition,"TheoremDefinition",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          AppendtoList($4,Parsetree,"DEFINED_AS",2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));    }
;

TemporalDefinition:
	{ startframe(2);} SymbolDef_LHS 
	{ endframe($2); } DEFINED_AS expr 
	{ if (($2)->GetSort()->GetLastElem()->GetSort()
	      != Env->GetDefaultOpSort())
	    SimpleError("defined symbol of illegal sort",$4->pos,$4->leng);
	  UnaryExactMatch($5,s_temporal,$4->pos,$4->leng);
	  delete $5;
          UnknownIDEnv->AppendAndPopUnknownIDs(isHeader,Env);
	  Env->Pop();
	  if (($2)->Get_nidt() != nidt_fctOperator) Env->Append($2);
          Parsetree = new ParseExpr(tTemporalDefinition,"TemporalDefinition",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          AppendtoList($4,Parsetree,"DEFINED_AS",2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
;

/*
 * expressions for the various classes
 * (we now get a confluent system ...)
 */

ConstantExpr:	/* <genSort> */
	expr	{ $$ = $1;
                  Parsetree = new ParseExpr(tConstantExpr,"ConstantExpr",1);
                  PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
                  SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
;

BooleanExpr:	/* <genSort> */
	expr 	{ $$ = $1; 
                  Parsetree = new ParseExpr(tBooleanExpr,"BooleanExpr",1);
                  PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
                  SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
;

StateFcnExpr: /* <genSort> */
	expr	{ $$ = $1; 
                  Parsetree = new ParseExpr(tStateFcnExpr,"StateFcnExpr",1);
                  PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
                  SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
;

PredExpr: /* <genSort> */
	expr	{ $$ = $1;
                  Parsetree = new ParseExpr(tPredExpr,"PredExpr",1);
                  PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
                  SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
;

TransFctExpr: /* <genSort> */
	expr	{ $$ = $1; 
                  Parsetree = new ParseExpr(tTransFctExpr,"TransFctExpr",1);
                  PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
                  SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
;

ActionExpr: /* <genSort> */
	expr	{ $$ = $1; 
                  Parsetree = new ParseExpr(tActionExpr,"ActionExpr",1);
                  PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
                  SemanticStack->SmStackPush(new SmStackElem(Parsetree));}
;

/*
 * the various MAJOR expressions we need to construct the
 * symbol definitions within the various sections
 */

constExpr_csl: /* <genSort> (Maximum aller Sorten) */
   	expr
	{ $$ = UnaryMaxMatch($1,s_transfct,
                             lastElem->pos,lastElem->leng,s_const);
          Parsetree = new ParseExpr(tconstExpr_csl,"constExpr_csl",1);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
   |	constExpr_csl COMMA expr
	{ $$ = MaxMatch($1,UnaryMaxMatch($3,s_transfct,lastElem->pos,
                                         lastElem->leng,s_const));
          Parsetree = new ParseExpr(tconstExpr_csl,"constExpr_csl",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          AppendtoList($2,Parsetree,"COMMA",2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
   ;

expr_csl: /* <genSort> (Liste aller Sorten) */
   	expr
	{ $$ = $1;
          Parsetree = new ParseExpr(texpr_csl,"expr_csl",1);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
   |	expr_csl COMMA expr
	{ $1->Append($3);
	  $$ = $1;
          Parsetree = new ParseExpr(texpr_csl,"expr_csl",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          AppendtoList($2, Parsetree,"COMMA",2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
   ;

expr: /* <genSort> (Sorte eines (Teil-)ausdrucks) */ 
	/* Sortenpruefung gemaess Abschnitt 2.2.3 der Ausarbeitung */
        /* Hilfsfunktionen UnaryMaxMatch(), ... siehe sort.h	   */
	/* Seiteneffekt: Behandlung Def/Use von lokalen Bezeichnern*/
	/* LCexpr parts */ 

    	RBR_ON expr RBR_OFF
	{ $$ = $2;
          Parsetree = new ParseExpr(texpr,"expr",3);
          AppendtoList($3, Parsetree,"RBR_OFF",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,2);
          AppendtoList($1, Parsetree,"RBR_ON",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));    }
    |	RBR_ON constExpr_csl RBR_OFF
	{ $$ = $2;
          Parsetree = new ParseExpr(texpr,"expr",3);
          AppendtoList($3, Parsetree,"RBR_OFF",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,2);
          AppendtoList($1, Parsetree,"RBR_ON",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |	expr TUPELSELECTOR
	{ $$ = UnaryMaxMatch($1,s_transfct,$2->pos,$2->leng,s_const);
          Parsetree = new ParseExpr(texpr,"expr",2);
          AppendtoList($2, Parsetree,"TUPLESELECTOR",2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
    |  	TRUE 
	{ $$ = new Sort (s_bool);
          Parsetree = new ParseExpr(texpr,"expr",1);
          AppendtoList($1, Parsetree,"TRUE",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |	FALSE
	{ $$ = new Sort (s_bool);
          Parsetree = new ParseExpr(texpr,"expr",1);
          AppendtoList($1, Parsetree,"FALSE",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
    |	expr boolRel expr %prec _L_AND_
        { $$ = BinaryMaxMatch($1,$3,s_temporal,$2->pos,$2->leng,s_bool);
          Parsetree = new ParseExpr(texpr,"expr",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
    |	expr boolRelIB IBL expr IBR %prec _IMPLIES_
        { $$ = BinaryMaxMatch($1,$4,s_temporal,$2->pos,$2->leng,s_bool);
          Parsetree = new ParseExpr(texpr,"expr",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
    | 	_L_NOT_ IBL expr IBR
	{ $$ = UnaryMaxMatch($3,s_temporal,$1->pos,$1->leng,s_bool);
          Parsetree = new ParseExpr(texpr,"expr",2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,2);
          AppendtoList($1, Parsetree,"_L_NOT_",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
    |	_ALL_QUANTOR_
        { UnknownIDEnv->Push(); }
        isElementOfSetListQB DOUBLECOLON IBL expr IBR
        { UnknownIDEnv->AppendAndPopUnknownIDs(isHeader,Env);
	  Env->Pop();
	  $$ = BoolConstToBool(
                 UnaryMaxMatch($6,s_temporal,lastElem->pos,
                               lastElem->leng,s_bool       ),
                 $3);
          Parsetree = new ParseExpr(texpr,"expr",4);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,4);
          AppendtoList($4, Parsetree,"DOUBLECOLON",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,2);
          AppendtoList($1, Parsetree,"_ALL_QUANTOR_",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
    |	_EXISTS_QUANTOR_
        { UnknownIDEnv->Push(); }
        isElementOfSetListQB DOUBLECOLON IBL expr IBR
	{ UnknownIDEnv->AppendAndPopUnknownIDs(isHeader,Env);
	  Env->Pop();
	  $$ = BoolConstToBool(
                 UnaryMaxMatch($6,s_temporal,lastElem->pos,
                               lastElem->leng,s_bool       ),
                 $3);
          Parsetree = new ParseExpr(texpr,"expr",4);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,4);
          AppendtoList($4, Parsetree,"DOUBLECOLON",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,2);
          AppendtoList($1, Parsetree,"_EXISTS_QUANTOR_",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
    |	SetExpr setRel SetExpr %prec _SUBSET_OF_
        { $$ = BinaryMaxBoolMatch($1,$3,$2->pos,$2->leng);
          Parsetree = new ParseExpr(texpr,"expr",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
    |	expr BinRel expr %prec EQ
        { $$ = BinaryMaxBoolMatch($1,$3,$2->pos,$2->leng);
          Parsetree = new ParseExpr(texpr,"expr",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
    |   expr isOrNotElementOf SetExpr %prec _ELEMENT_OF_
        { $$ = BinaryMaxBoolMatch($1,$3,$2->pos,$2->leng);
          Parsetree = new ParseExpr(texpr,"expr",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));    }
    |	prefixLines
	{ $$ = $1; 
          Parsetree = new ParseExpr(texpr,"expr",1);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));    }
    |  	CHOOSE RBR_ON normalID _ELEMENT_OF_ SetExpr
	{ $5 = UnaryMaxMatch($5,s_transfct,lastElem->pos,
                             lastElem->leng,s_const);
	  if ($3->is_primed())
            SimpleError("prime not allowed here",$3->pos,$3->leng);
	  $3->Set_nidt(nidt_boundVar);
	  $3->SetSort(new Sort(s_const));
	  Env->Push();
          UnknownIDEnv->Push();
	  Env->Append($3); }
	COLON expr RBR_OFF
	{ UnknownIDEnv->AppendAndPopUnknownIDs(isHeader,Env);
	  Env->Pop();
	  $$ = BoolConstToConst(
                 UnaryMaxMatch($8,s_action,lastElem->pos,
                               lastElem->leng,s_bool     ),
                 $5);
          Parsetree = new ParseExpr(texpr,"expr",8);
          AppendtoList($9, Parsetree,"RBR_OFF",8);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,7);
          AppendtoList($7, Parsetree,"COLON",6);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,5);
          AppendtoList($4, Parsetree,"_ELEMENT_OF_",4);
          AppendtoList($3, Parsetree,"normalID",3);
          AppendtoList($2, Parsetree,"RBR_ON",2);
          AppendtoList($1, Parsetree,"CHOOSE",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
    |  	'[' FUNCTION
        { UnknownIDEnv->Push(); }
        isElementOfSetListQB _MAPSTO_ expr ']'
	{ UnknownIDEnv->AppendAndPopUnknownIDs(isHeader,Env);
	  Env->Pop();
          $$ = BinaryMaxMatch($4,$6,s_transfct,$5->pos,$5->leng,s_const);
          Parsetree = new ParseExpr(texpr,"expr",6);
          AppendtoList($7, Parsetree,"]",6);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,5);
          AppendtoList($5, Parsetree,"_MAPSTO_",4);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          AppendtoList($2, Parsetree,"FUNCTION",2);
          AppendtoList($1, Parsetree,"[",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));     }
    |	'[' CHANGE ID SEMICOLON expr _MAPSTO_ expr ']'
	{ Sort *s = SortOfID($3);
          $$ = BinaryMaxMatch(
            UnaryMaxMatch(s,s_transfct,$3->pos,$3->leng,s_const),
            BinaryMaxMatch($5,$7,s_transfct,$6->pos,$6->leng,s_const),
            s_transfct,$1->pos,$1->leng,s_const);
          Parsetree = new ParseExpr(texpr,"expr",8);
          AppendtoList($8, Parsetree,"]",8);  
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,7);
          AppendtoList($6, Parsetree,"_MAPSTO_",6);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,5);
          AppendtoList($4, Parsetree,"SEMICOLON",4);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          AppendtoList($2, Parsetree,"CHANGE",2);
          AppendtoList($1, Parsetree,"[",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));    }
    |  	DSBR_ON recdef_csl DSBR_OFF
	{ $$ = $2; 
          Parsetree = new ParseExpr(texpr,"expr",3);
          AppendtoList($3, Parsetree,"DSBR_OFF",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,2);
          AppendtoList($1, Parsetree,"DSBR_ON",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));    }
    |	DSBR_ON CHANGE ID SEMICOLON componentID _MAPSTO_ expr DSBR_OFF
        { Sort *s = SortOfID($3);
          $$ = BinaryMaxMatch(s,$7,s_transfct,$1->pos,$1->leng,s_const);
          Parsetree = new ParseExpr(texpr,"expr",8);
          AppendtoList($8, Parsetree,"DSBR_OFF",8);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,7);
          AppendtoList($6, Parsetree,"_MAPSTO_",6);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,5);
          AppendtoList($4, Parsetree,"SEMICOLON",4);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          AppendtoList($2, Parsetree,"CHANGE",2);
          AppendtoList($1, Parsetree,"DSBR_ON",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
    | 	SetExpr
	{ $$ = $1;
          Parsetree = new ParseExpr(texpr,"expr",1);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
    | 	expr BinOp expr %prec PLUS
        { $$ = BinaryMaxMatch($1,$3,s_transfct,$2->pos,$2->leng,s_const);
          Parsetree = new ParseExpr(texpr,"expr",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
    |   MINUS expr %prec UNARYMINUS
       	{ $$ = UnaryMaxMatch($2,s_transfct,$1->pos,$1->leng,s_const);
          Parsetree = new ParseExpr(texpr,"expr",2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,2);
          AppendtoList($1, Parsetree,"MINUS",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));    }
    | 	NATNUM
	{ $$ = new Sort(s_const);
          Parsetree = new ParseExpr(texpr,"expr",1);
          AppendtoList($1, Parsetree,"NATNUM",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
    |	StringExpr
	{ $$ = new Sort(s_const);
          Parsetree = new ParseExpr(texpr,"expr",1);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));    }
    |  	expr DOT componentID
	{ $$ = UnaryMaxMatch($1,s_transfct,$2->pos,$2->leng,s_const);
          Parsetree = new ParseExpr(texpr,"expr",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          AppendtoList($2, Parsetree,"DOT",2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  } 
    |	IF IBL expr IBR THEN IBL expr IBR ELSE IBL expr IBR
	{ $$ = BoolConstToConst(
                 UnaryMaxMatch($3,s_action,$5->pos,$5->leng,s_bool),
                 BinaryMaxMatch(
                   UnaryMaxMatch($7,s_transfct,$9->pos,$9->leng,s_const),
                   UnaryMaxMatch($11,s_transfct,lastElem->pos,
                                                lastElem->leng,s_const),
                   s_transfct,lastElem->pos,lastElem->leng,s_const));
           Parsetree = new ParseExpr(texpr,"expr",6);
           PopandAppend(SemanticStack->SmStackPop(), Parsetree,6);
           AppendtoList($9, Parsetree,"ELSE",5);
           PopandAppend(SemanticStack->SmStackPop(), Parsetree,4);
           AppendtoList($5, Parsetree,"THEN",3);
           PopandAppend(SemanticStack->SmStackPop(), Parsetree,2);
           AppendtoList($1, Parsetree,"IF",1);
           SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
    |	CASE IBL CaseRA_list IBR
	{ $$ = $3;
          Parsetree = new ParseExpr(texpr,"expr",2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,2);
          AppendtoList($1, Parsetree,"CASE",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));    }
    |	LET
	{ Env->Push();
          UnknownIDEnv->Push();
	  Env->SetDefaultOpSort(Env->GetDefaultParSort()); }
	IBL expr_list IBR IN IBL expr IBR
	{ UnknownIDEnv->AppendAndPopUnknownIDs(isHeader,Env);
	  Env->Pop();
	  $$ = $8;
          Parsetree = new ParseExpr(texpr,"expr",4);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,4);
          AppendtoList($6, Parsetree,"IN",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,2);
          AppendtoList($1, Parsetree,"LET",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
 
    /* LAexpr parts */

    | 	expr PRIME
	{ Sort_Type t = $1->GetFirstElem()->GetSort();
          if ( ($1->GetLength() > 1) || (t == s_nested) ) {
	      SimpleError("prime not allowed here",$2->pos,$2->leng);
              $$ = $1; }
          else {
            if (Env->GetModuleType() == mt_const)
	      SimpleError("prime not allowed in CONSTANT MODULE",
                          $2->pos,$2->leng);
	    if      (t == s_pred)     t = s_action;
	    else if (t == s_statefct) t = s_transfct;
	    else if ( (t != s_const) && (t != s_bool) && (t != s_joker) )
              SimpleError("prime not allowed here",$2->pos,$2->leng);
            delete $1;
	    $$ = new Sort(t); }
          Parsetree = new ParseExpr(texpr,"expr",2);
          AppendtoList($2, Parsetree,"PRIME",2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
    |	ENABLED RBR_ON ActionExpr RBR_OFF
	{ Sort *dummy = UnaryMaxMatch($3,s_action,$1->pos,$1->leng,s_bool);
	  delete dummy;
	  $$ = new Sort (s_pred);
          Parsetree = new ParseExpr(texpr,"expr",4);
          AppendtoList($4, Parsetree,"RBR_OFF",4);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          AppendtoList($2, Parsetree,"RBR_ON",2);
          AppendtoList($1, Parsetree,"ENABLED",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   } 
    | 	UNCHANGED RBR_ON ID_csl RBR_OFF
	{ SymListElem *e = $3->GetFirstElem();
	  while (e)
	    {
	    if (! (*(e->GetSort()) <= s_statefct))
	      SimpleError("symbol of sort STATE FCN expected",e->pos,e->leng);
	    e = $3->GetNextElem();
	    }
	  delete $3;
	  $$ = new Sort (s_action);
          Parsetree = new ParseExpr(texpr,"expr",4);
          AppendtoList($4, Parsetree,"RBR_OFF",4);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          AppendtoList($2, Parsetree,"RBR_ON",2);
          AppendtoList($1, Parsetree,"UNCHANGED",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));    }		
    |	'[' ActionExpr ']' '_' RBR_ON ID_csl RBR_OFF
	{ Sort *dummy = UnaryMaxMatch($2,s_action,$3->pos,$3->leng,s_bool);
	  delete dummy;
	  SymListElem *e = $6->GetFirstElem();
	  while (e)
	    {
	    if (! (*(e->GetSort()) <= s_statefct))
	      SimpleError("symbol of sort STATE FCN expected",e->pos,e->leng);
	    e = $6->GetNextElem();
	    }
	  delete $6;
	  $$ = new Sort (s_action);
          Parsetree = new ParseExpr(texpr,"expr",7);
          AppendtoList($7, Parsetree,"RBR_OFF",7);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,6);
          AppendtoList($5, Parsetree,"RBR_ON",5); 
          AppendtoList($4, Parsetree,"_",4);
          AppendtoList($3, Parsetree,"]",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,2);
          AppendtoList($1, Parsetree,"[",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));    }
    |   LT expr GT '_' RBR_ON ID_csl RBR_OFF
	{ Sort *dummy = UnaryMaxMatch($2,s_action,$3->pos,$3->leng,s_bool);
	  delete dummy;
	  SymListElem *e = $6->GetFirstElem();
	  while (e)
	    {
	    if (! (*(e->GetSort()) <= s_statefct))
	      SimpleError("symbol of sort STATE FCN expected",e->pos,e->leng);
	    e = $6->GetNextElem();
	    }
	  delete $6;
	  $$ = new Sort (s_action); 
          Parsetree = new ParseExpr(texpr,"expr",7);
          AppendtoList($7, Parsetree,"RBR_OFF",7);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,6);
          AppendtoList($5, Parsetree,"RBR_ON",5); 
          AppendtoList($4, Parsetree,"_",4);
          AppendtoList($3, Parsetree,"GT",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,2);
          AppendtoList($1, Parsetree,"LT",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }

    /* TLAexpr parts */

    |	_HENCEFORTH_ '[' StateFcnExpr ']' RBR_ON ActionExpr RBR_OFF
	{ Sort *dummy = UnaryMaxMatch($3,s_statefct,$4->pos,$4->leng,s_const);
	  delete dummy;
	  dummy = UnaryMaxMatch($6,s_action,$7->pos,$7->leng,s_bool);
	  delete dummy;
	  $$ = new Sort (s_temporal);
          Parsetree = new ParseExpr(texpr,"expr",7);
          AppendtoList($7, Parsetree,"RBR_OFF",7);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,6);
          AppendtoList($5, Parsetree,"RBR_ON",5);
          AppendtoList($4, Parsetree,"]",4);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          AppendtoList($2, Parsetree,"[",2); 
          AppendtoList($1, Parsetree,"_HENCEFORTH_",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
    |	_HENCEFORTH_ RBR_ON expr RBR_OFF
	/* { if (!( (*($3) <= s_pred) || (*($3) == s_temporal) ))  */
	/*     SimpleError("sort PREDICATE or TEMPORAL expected",  */
        /*                 $1->pos,$1->leng);                      */
	/*   delete $3;                                            */
	/*   $$ = new Sort (s_temporal); }                         */
        /* Wie soll []( [A]_f ) und [] [f] (A) korrekt behandelt   */
        /* werden ? Hier nicht ganz korrekte Notloesung:           */
	{ Sort *dummy = UnaryMaxMatch($3,s_temporal,$1->pos,$1->leng,s_bool);
	  delete dummy;
	  $$ = new Sort (s_temporal); 
          Parsetree = new ParseExpr(texpr,"expr",4);
          AppendtoList($4, Parsetree,"RBR_OFF",4);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          AppendtoList($2, Parsetree,"RBR_ON",2); 
          AppendtoList($1, Parsetree,"_HENCEFORTH_",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |	_EVENTUALLY_ '[' StateFcnExpr ']' RBR_ON ActionExpr RBR_OFF
	{ Sort *dummy = UnaryMaxMatch($3,s_statefct,$4->pos,$4->leng,s_const);
	  delete dummy;
	  dummy = UnaryMaxMatch($6,s_action,$7->pos,$7->leng,s_bool);
	  delete dummy;
	  $$ = new Sort (s_temporal);
          Parsetree = new ParseExpr(texpr,"expr",7);
          AppendtoList($7, Parsetree,"RBR_OFF",7);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,6);
          AppendtoList($5, Parsetree,"RBR_ON",5);
          AppendtoList($4, Parsetree,"]",4);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          AppendtoList($2, Parsetree,"[",2);
          AppendtoList($1, Parsetree,"_EVENTUALLY_",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
    |	_EVENTUALLY_ RBR_ON expr RBR_OFF
	/* { if (!( (*($3) <= s_pred) || (*($3) == s_temporal) ))  */
	/*     SimpleError("sort PREDICATE or TEMPORAL expected",  */
        /*                 $1->pos,$1->leng);                      */
	/*   delete $3;                                            */
	/*   $$ = new Sort (s_temporal); }                         */
        /* Wie soll <>( <A>_f ) und <> [f] (A) korrekt behandelt   */
        /* werden ? Hier nicht ganz korrekte Notloesung:           */
	{ Sort *dummy = UnaryMaxMatch($3,s_temporal,$1->pos,$1->leng,s_bool);
	  delete dummy;
	  $$ = new Sort (s_temporal);
          Parsetree = new ParseExpr(texpr,"expr",4);
          AppendtoList($4, Parsetree,"RBR_OFF",4);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          AppendtoList($2, Parsetree,"RBR_ON",2);
          AppendtoList($1, Parsetree,"_EVENTUALLY_",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));    }
    |	_BOLD_EXISTS_ normalID_csl DOUBLECOLON
	{ SortListElem *se = new SortListElem(s_statefct);
          SymList* dummy = new SymList;
	  SymListElem *e = $2->GetFirstElem();
	  while (e)
	    {
	    e->Set_nidt(nidt_boundVar);
	    e->SetSort(se);
            
	    e = $2->GetNextElem();
	    }
	  Env->Push();
          UnknownIDEnv->Push();
	  Env->Append($2); }
	IBL expr IBR
	{ Sort *dummy = UnaryMaxMatch($6,s_temporal,lastElem->pos,
                                      lastElem->leng,s_bool);
	  delete dummy;
	  UnknownIDEnv->AppendAndPopUnknownIDs(isHeader,Env);
	  Env->Pop();
	  $$ = new Sort (s_temporal);
          Parsetree = new ParseExpr(texpr,"expr",4);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,4);
          AppendtoList($3, Parsetree,"DOUBLECOLON",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,2);
          AppendtoList($1, Parsetree,"_BOLD_EXISTS_",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
    |	expr _LEADSTO_ expr
	{ Sort *dummy = BinaryMaxMatch($1,$3,s_temporal,$2->pos,
                                       $2->leng,s_bool);
	  delete dummy;
	  $$ = new Sort (s_temporal);
          Parsetree = new ParseExpr(texpr,"expr",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3); 
          AppendtoList($2, Parsetree,"_LEADSTO_",2); 
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
    | 	WF '[' StateFcnExpr ']' RBR_ON ActionExpr RBR_OFF
	{ Sort *dummy = UnaryMaxMatch($3,s_statefct,$4->pos,$4->leng,s_const);
	  delete dummy;
	  dummy = UnaryMaxMatch($6,s_action,$7->pos,$7->leng,s_bool);
	  delete dummy;
	  $$ = new Sort (s_temporal);
          Parsetree = new ParseExpr(texpr,"expr",7);
          AppendtoList($7, Parsetree,"RBR_OFF",7);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,6);
          AppendtoList($5, Parsetree,"RBR_ON",5);
          AppendtoList($4, Parsetree,"]",4);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          AppendtoList($2, Parsetree,"[",2);
          AppendtoList($1, Parsetree,"WF",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));    }
    | 	SF '[' StateFcnExpr ']' RBR_ON ActionExpr RBR_OFF 
	{ Sort *dummy = UnaryMaxMatch($3,s_statefct,$4->pos,$4->leng,s_const);
	  delete dummy;
	  dummy = UnaryMaxMatch($6,s_action,$7->pos,$7->leng,s_bool);
	  delete dummy;
	  $$ = new Sort (s_temporal);
          Parsetree = new ParseExpr(texpr,"expr",7);
          AppendtoList($7, Parsetree,"RBR_OFF",7);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,6);
          AppendtoList($5, Parsetree,"RBR_ON",5);
          AppendtoList($4, Parsetree,"]",4);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          AppendtoList($2, Parsetree,"[",2);
          AppendtoList($1, Parsetree,"SF",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));     }
;

SetExpr: /* genSort  (Sorte eines (Teil-)ausdrucks) */ 
	/* Sortenpruefung gemaess Abschnitt 2.2.3 der Ausarbeitung */
        /* Hilfsfunktionen UnaryMaxMatch(), ... siehe sort.h	   */
	/* Seiteneffekt: Behandlung Def/Use von lokalen Bezeichnern*/

	ID
        { $$ = SortOfID($1);
          Parsetree = new ParseExpr(tSetExpr,"SetExpr",1);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
    |   OperatorApplication
	{ $$ = $1;
          Parsetree = new ParseExpr(tSetExpr,"SetExpr",1);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
    |   ID '[' constExpr_csl ']'
        { Sort *s = SortOfID($1);
	  $$ = MaxMatch(
                 UnaryMaxMatch(s,s_transfct,$1->pos,$1->leng,s_const),$3);
          Parsetree = new ParseExpr(tSetExpr,"SetExpr",4);
          AppendtoList($4, Parsetree,"]",4);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          AppendtoList($2, Parsetree,"[",2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
    |	ANY
	{ if (! isTypeDecl)
            SimpleError("ANY only allowed in type declarations",
                        $1->pos,$1->leng);
          $$ = new Sort(s_const);
          Parsetree = new ParseExpr(tSetExpr,"SetExpr",1);
          AppendtoList($1, Parsetree,"ANY",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));    }
    | 	RBR_ON SetExpr RBR_OFF
	{ $$ = $2;
          Parsetree = new ParseExpr(tSetExpr,"SetExpr",3);
          AppendtoList($3, Parsetree,"RBR_OFF",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,2);
          AppendtoList($1, Parsetree,"RBR_ON",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));    }
    |	SetExpr setOp SetExpr %prec _UNION_
        { $$ = BinaryMaxMatch($1,$3,s_transfct,$2->pos,$2->leng,s_const);
          Parsetree = new ParseExpr(tSetExpr,"SetExpr",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));     }
    |	beginSet constExpr_csl '}'
	{ UnknownID_allowed--;
          UnknownIDEnv->PopUnknownIDs();
          $$ = MaxMatch($2,new Sort(s_const));
                                /* Keinen s_joker aus {...} rauslassen */
          Parsetree = new ParseExpr(tSetExpr,"SetExpr",3);
          AppendtoList($3, Parsetree,"}",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));     }
    |	beginSet expr '|' isElementOfSetListQB COLON expr '}'
        { UnknownID_allowed--;
          UnknownIDEnv->AppendAndPopUnknownIDs(isHeader,Env);
	  Env->Pop();
	  $$ = MaxMatch(
                MaxMatch(
                 UnaryMaxMatch($2,s_transfct,$3->pos,$3->leng,s_const),
                 BoolConstToConst(
                   UnaryMaxMatch($6,s_action,$7->pos,$7->leng,s_bool),
                   $4)),
                new Sort(s_const)); /* Keinen s_joker aus {...} rauslassen */
           Parsetree = new ParseExpr(tSetExpr,"SetExpr",7);
           AppendtoList($7, Parsetree,"}",7);
           PopandAppend(SemanticStack->SmStackPop(), Parsetree,6);
           AppendtoList($5, Parsetree,"COLON",5);
           PopandAppend(SemanticStack->SmStackPop(), Parsetree,4);
           AppendtoList($3, Parsetree,"|",3);
           PopandAppend(SemanticStack->SmStackPop(), Parsetree,2);
           PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
           SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
/*	  $$ = BoolConstToConst(
                 UnaryMaxMatch($4,s_action,$5->pos,$5->leng,s_bool),
                 $2);     } */
    |	beginSet expr '|' isElementOfSetListQB '}'
        { UnknownID_allowed--;
	  UnknownIDEnv->AppendAndPopUnknownIDs(isHeader,Env);
	  Env->Pop();
          $$ = MaxMatch(
                BinaryMaxMatch(
                 UnaryMaxMatch($2,s_transfct,$3->pos,$3->leng,s_const),
                 $4,s_transfct,$5->pos,$5->leng,s_const),
               new Sort(s_const)); /* Keinen s_joker aus {...} rauslassen */
          Parsetree = new ParseExpr(tSetExpr,"SetExpr",5);
          AppendtoList($5, Parsetree,"}",5);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,4);
          AppendtoList($3, Parsetree,"|",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));       }
    |	SUBSET RBR_ON SetExpr RBR_OFF
	{ $$ = UnaryMaxMatch($3,s_transfct,$1->pos,$1->leng,s_const);
          Parsetree = new ParseExpr(tSetExpr,"SetExpr",4);
          AppendtoList($4, Parsetree,"RBR_OFF",4); 
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          AppendtoList($2, Parsetree,"RBR_ON",2);
          AppendtoList($1, Parsetree,"SUBSET",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   } 
    | 	UNION RBR_ON SetExpr RBR_OFF
	{ $$ = UnaryMaxMatch($3,s_transfct,$1->pos,$1->leng,s_const);
          Parsetree = new ParseExpr(tSetExpr,"SetExpr",4);
          AppendtoList($4, Parsetree,"RBR_OFF",4);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          AppendtoList($2, Parsetree,"RBR_ON",2);
          AppendtoList($1, Parsetree,"UNION",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));    } 
    | 	_EMPTYSET_
	{ $$ = new Sort(s_const);
          Parsetree = new ParseExpr(tSetExpr,"SetExpr",1);
          AppendtoList($1, Parsetree,"_EMPTYSET_",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));     }
    |	predefSet
	{ $$ = new Sort(s_const);
          Parsetree = new ParseExpr(tSetExpr,"SetExpr",1);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));       }
    |	SET '[' SetExpr _RIGHTARROW_ SetExpr ']'
        { $$ = BinaryMaxMatch($3,$5,s_transfct,$4->pos,$4->leng,s_const);
          Parsetree = new ParseExpr(tSetExpr,"SetExpr",6);
          AppendtoList($6, Parsetree,"]",6);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,5);
          AppendtoList($4, Parsetree,"_RIGHTARROW_",4);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          AppendtoList($2, Parsetree,"[",2);
          AppendtoList($1, Parsetree,"SET",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));    }
    |	'[' CHANGE ID SEMICOLON expr COLON SetExpr ']'
	{ Sort *s = SortOfID($3);
          $$ = BinaryMaxMatch(
                 BinaryMaxMatch(
                   UnaryMaxMatch($5,s_transfct,$6->pos,$6->leng,s_const), 
                   UnaryMaxMatch($7,s_transfct,$8->pos,$8->leng,s_const), 
                   s_transfct,$8->pos,$8->leng,s_const),
                 UnaryMaxMatch(s,s_transfct,$3->pos,$3->leng,s_const),
                 s_transfct,$8->pos,$8->leng,s_const);
          Parsetree = new ParseExpr(tSetExpr,"SetExpr",8);
          AppendtoList($8, Parsetree,"]",8);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,7);
          AppendtoList($6, Parsetree,"COLON",6);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,5);
          AppendtoList($4, Parsetree,"SEMICOLON",4);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          AppendtoList($2, Parsetree,"CHANGE",2);
          AppendtoList($1, Parsetree,"[",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   } 
    |	DOMAINt RBR_ON ID RBR_OFF
	{ Sort *s = SortOfID($3);
          $$ = UnaryMaxMatch(s,s_transfct,$1->pos,$1->leng,s_const);
          Parsetree = new ParseExpr(tSetExpr,"SetExpr",4);
          AppendtoList($4, Parsetree,"RBR_OFF",4); 
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          AppendtoList($2, Parsetree,"RBR_ON",2);
          AppendtoList($1, Parsetree,"DOMAINt",1); 
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); } 
    |	DSBR_ON rctype_ssl DSBR_OFF
	{ $$ = $2;
          Parsetree = new ParseExpr(tSetExpr,"SetExpr",3);
          AppendtoList($3, Parsetree,"DSBR_OFF",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,2);
          AppendtoList($1, Parsetree,"DSBR_ON",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
    |	DSBR_ON CHANGE ID SEMICOLON componentID COLON SetExpr DSBR_OFF
        { Sort *s = SortOfID($3);
          $$ = BinaryMaxMatch(s,$7,s_transfct,$2->pos,$2->leng,s_const);
          Parsetree = new ParseExpr(tSetExpr,"SetExpr",8);
          AppendtoList($8, Parsetree,"DSBR_OFF",8);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,7);
          AppendtoList($6, Parsetree,"COLON",6);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,5);
          AppendtoList($4, Parsetree,"SEMICOLON",4);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          AppendtoList($2, Parsetree,"CHANGE",2);
          AppendtoList($1, Parsetree,"DSBR_ON",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
;

/*
 * the various MINOR expressions we need to construct the
 * symbol definitions within the various sections
 */

beginSet:
	'{'
	{ UnknownIDEnv->Push();
          UnknownID_allowed++;
          Parsetree = new ParseExpr(tbeginSet,"beginSet",1);
          AppendtoList($1, Parsetree,"{",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
;

expr_list:
	/* Def. lokale Bezeichner + Sortenpruefung */
	SymbolDef_LHS DEFINED_AS expr SEMICOLON
	{ if (!(*($3) == s_joker))
  	    UnaryExactMatch($3,$1->GetSort()->GetLastElem()->GetSort(),
                            $2->pos,$2->leng);
	  delete $3;
          UnknownIDEnv->AppendAndPopUnknownIDs(isHeader,Env);
	  Env->Pop();
	  if (($1)->Get_nidt() != nidt_fctOperator) Env->Append($1);
          Parsetree = new ParseExpr(texpr_list,"expr_list",4);
          AppendtoList($4, Parsetree,"SEMICOLON",4);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,3);
          AppendtoList($2, Parsetree,"DEFINED_AS",2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |	expr_list SymbolDef_LHS DEFINED_AS expr SEMICOLON
	{ if (!(*($4) == s_joker))
	    UnaryExactMatch($4,$2->GetSort()->GetLastElem()->GetSort(),
                            $3->pos,$3->leng);
	  delete $4;
          UnknownIDEnv->AppendAndPopUnknownIDs(isHeader,Env);
	  Env->Pop();
	  if (($2)->Get_nidt() != nidt_fctOperator) Env->Append($2);
          Parsetree = new ParseExpr(texpr_list,"expr_list",5);
          AppendtoList($5, Parsetree,"SEMICOLON",5);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,4);
          AppendtoList($3, Parsetree,"DEFINED_AS",3);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,2);
          PopandAppend(SemanticStack->SmStackPop(), Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));    }
;

predefSet: /* <selem> */
	NAT	{ $$ = $1;
                  Parsetree = new ParseExpr(tpredefSet,"predefSet",1);
                  AppendtoList($1, Parsetree,"NAT",1);
                  SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
;

setRel: /* <selem> */ 
    	_SUBSET_OF_	{ $$ = $1;
                          Parsetree = new ParseExpr(tsetRel,"setRel",1);
                          AppendtoList($1, Parsetree,"_SUBSET_OF_",1);
                          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
;

setOp: /* <selem> */ 
     	_UNION_		 { $$ = $1;
                           Parsetree = new ParseExpr(tsetOp,"setOp",1);
                          AppendtoList($1, Parsetree,"_UNION_",1);
                          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    | 	_INTERSECT_	 { $$ = $1;
                          Parsetree = new ParseExpr(tsetOp,"setOp",1);
                          AppendtoList($1, Parsetree,"_INTERSECT_",1);
                          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |	_SET_DIFFERENCE_ { $$ = $1;
                           Parsetree = new ParseExpr(tsetOp,"setOp",1);
                          AppendtoList($1, Parsetree,"_SET_DIFFERENCE_",1);
                          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |	_CROSS_		 { $$ = $1;
                           Parsetree = new ParseExpr(tsetOp,"setOp",1);
                          AppendtoList($1, Parsetree,"_CROSS_",1);
                          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
;

BinOp: /* <selem> */ 
	PLUS	{ $$ = $1;
                  Parsetree = new ParseExpr(tBinOp,"BinOp",1);
                  AppendtoList($1, Parsetree,"PLUS",1);
                  SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |	MINUS	{ $$ = $1;
                  Parsetree = new ParseExpr(tBinOp,"BinOp",1);
                  AppendtoList($1, Parsetree,"MINUS",1);
                  SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |	TIMES	{ $$ = $1;
                  Parsetree = new ParseExpr(tBinOp,"BinOp",1);
                  AppendtoList($1, Parsetree,"TIMES",1);
                  SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |   DIVIDED	{ $$ = $1;
                  Parsetree = new ParseExpr(tBinOp,"BinOp",1);
                  AppendtoList($1, Parsetree,"DIVIDED",1);
                  SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |   DIV	{ $$ = $1;
                  Parsetree = new ParseExpr(tBinOp,"BinOp",1);
                  AppendtoList($1, Parsetree,"DIV",1);
                  SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
    |   MOD	{ $$ = $1;
                  Parsetree = new ParseExpr(tBinOp,"BinOp",1);
                  AppendtoList($1, Parsetree,"MOD",1);
                  SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |   POWER	{ $$ = $1;
                  Parsetree = new ParseExpr(tBinOp,"BinOp",1);
                  AppendtoList($1, Parsetree,"POWER",1);
                  SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
;

StringExpr: /* <selem> */ 
	STRING	{ $$ = $1;
                  Parsetree = new ParseExpr(tStringExpr,"StringExpr",1);
                  AppendtoList($1, Parsetree,"STRING",1);
                  SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
;

isOrNotElementOf: /* <selem> */ 
	_ELEMENT_OF_	 { $$ = $1;
                           Parsetree = new ParseExpr(tisOrNotElementOf,"isOrNotElementOf",1);
                           AppendtoList($1, Parsetree,"_ELEMENT_OF_",1);
                           SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |	_NOT_ELEMENT_OF_ { $$ = $1;
                           Parsetree = new ParseExpr(tisOrNotElementOf,"isOrNotElementOf",1);
                           AppendtoList($1, Parsetree,"_NOT_ELEMENT_OF_",1);
                           SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
;

recdef_csl: /* <genSort> (Sorte des (Teil-)Ausdrucks) */
	recdef
	{ $$ = $1;
          Parsetree = new ParseExpr(trecdef_csl,"recdef_csl",1);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
    |	recdef_csl COMMA recdef
	{ $$ = BinaryMaxMatch($1,$3,s_transfct,$2->pos,$2->leng,s_const);
          Parsetree = new ParseExpr(trecdef_csl,"recdef_csl",3);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,3);
          AppendtoList($2, Parsetree,"COMMA",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
;

recdef:	/* <genSort> (Sorte des (Teil-)Ausdrucks) */
	componentID _MAPSTO_ expr
	{ $$ = UnaryMaxMatch($3,s_transfct,lastElem->pos,
                                           lastElem->leng,s_const);
          Parsetree = new ParseExpr(trecdef,"recdef",3);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,3);
          AppendtoList($2, Parsetree,"_MAPSTO_",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
;

rctype_ssl: /* <genSort> (Sorte des (Teil-)Ausdrucks) */
    	rctype
	{ $$ = $1;
          Parsetree = new ParseExpr(trctype_ssl,"rctype_ssl",1);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
    |	rctype_ssl SEMICOLON rctype
	{ $$ = BinaryMaxMatch($1,$3,s_transfct,$2->pos,$2->leng,s_const);
          Parsetree = new ParseExpr(trctype_ssl,"rctype_ssl",3);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,3);
          AppendtoList($2, Parsetree,"SEMICOLON",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
;

rctype: /* <genSort> (Sorte des (Teil-)Ausdrucks) */
    	normalID_csl COLON SetExpr
	{ SymListElem *e = $1->GetFirstElem();
	  while (e)
	    {
	    e->Set_nidt(nidt_rec_comp);
	    e->SetSort(new Sort(s_const));
	    e = $1->GetNextElem();
	    }
	  delete $1;
	  $$ = UnaryMaxMatch($3,s_transfct,lastElem->pos,
                                           lastElem->leng,s_const);
          Parsetree = new ParseExpr(trctype,"rctype",3);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,3);
          AppendtoList($2, Parsetree,"COLON",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
;

boolRel: /* <selem> */ 
	_L_AND_	{ $$ = $1;
                  Parsetree = new ParseExpr(tboolRel,"boolRel",1);
                  AppendtoList($1, Parsetree,"_L_AND_",1);
                  SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
    |	_L_OR_	{ $$ = $1;
                  Parsetree = new ParseExpr(tboolRel,"boolRel",1);
                  AppendtoList($1, Parsetree,"_L_OR_",1);
                  SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
;

boolRelIB: /* <selem> */ 
    	_IMPLIES_ { $$ = $1;
                    Parsetree = new ParseExpr(tboolRelIB,"boolRelIB",1);
                    AppendtoList($1, Parsetree,"_IMPLIES_",1);
                    SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
    |	_EQUIV_	  { $$ = $1;
                    Parsetree = new ParseExpr(tboolRelIB,"boolRelIB",1);
                    AppendtoList($1, Parsetree,"_EQUIV_",1);
                    SemanticStack->SmStackPush(new SmStackElem(Parsetree));}
;
BinRel: /* <selem> */ 
	  EQ	{ $$ = $1;
                  Parsetree = new ParseExpr(tBinRel,"BinRel",1);
                  AppendtoList($1, Parsetree,"EQ",1);
                  SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
        | NE	{ $$ = $1;
                  Parsetree = new ParseExpr(tBinRel,"BinRel",1);
                  AppendtoList($1, Parsetree,"NE",1);
                  SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
        | LT	{ $$ = $1;
                  Parsetree = new ParseExpr(tBinRel,"BinRel",1);
                  AppendtoList($1, Parsetree,"LT",1);
                  SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
        | GT	{ $$ = $1;
                  Parsetree = new ParseExpr(tBinRel,"BinRel",1);
                  AppendtoList($1, Parsetree,"GT",1);
                  SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
        | LE	{ $$ = $1;
                  Parsetree = new ParseExpr(tBinRel,"BinRel",1);
                  AppendtoList($1, Parsetree,"LE",1);
                  SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
        | GE	{ $$ = $1;
                  Parsetree = new ParseExpr(tBinRel,"BinRel",1);
                  AppendtoList($1, Parsetree,"GE",1);
                  SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }

;

OperatorApplication: /* <genSort> (Sorte des (Teil-)Ausdrucks) */
	qualifierPrefixes normalID RBR_ON expr_csl RBR_OFF
							%prec PRECLOWERTHANDOT
	{ SymListElem *e;
          if ($2->is_primed())
            SimpleError("prime not allowed here",$2->pos,$2->leng);
          $2->SetSort($4);
	  $1->Append($2);
	  if (UnknownID_allowed || isHeader) {  // Ausnahme zu declare before
            UnknownIDEnv->AppendUnknown($1);	// use moeglich
            $$ = new Sort(s_joker); }
          else {				// declare before use
            e = Env->UseOfID($1);
            delete $1;
	    $$ = new Sort(e->GetSort()->GetLastElem()); }
          Parsetree = new ParseExpr(tOperatorApplication,"OperatorApplication",5);
          AppendtoList($5, Parsetree,"RBR_OFF",5);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,4);
          AppendtoList($3, Parsetree,"RBR_ON",3);
          AppendtoList($2, Parsetree,"normalID",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));    }
;

prefixLines: /* <genSort> (Sorte des (Teil-)Ausdrucks) */
	conjunctLines
	{ $$ = $1;
          Parsetree = new ParseExpr(tprefixLines,"prefixLines",1);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |	disjunctLines
	{ $$ = $1;
          Parsetree = new ParseExpr(tprefixLines,"prefixLines",1);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
;

conjunctLines: /* <genSort> (Sorte des (Teil-)Ausdrucks) */
	PREF_AND IBL expr IBR	%prec PREF_AND
	{ $$ = UnaryMaxMatch($3,s_temporal,lastElem->pos,
                                           lastElem->leng,s_bool);
          Parsetree = new ParseExpr(tconjunctLines,"conjunctLines",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,2);
          AppendtoList($1, Parsetree,"PREF_AND",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
    | 	conjunctLines PREF_AND IBL expr IBR
	{ $$ = BinaryMaxMatch(
                 UnaryMaxMatch($4,s_temporal,lastElem->pos,
                                             lastElem->leng,s_bool),
                 $1,s_temporal,lastElem->pos,lastElem->leng,s_bool);
          Parsetree = new ParseExpr(tconjunctLines,"conjunctLines",3);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,3);
          AppendtoList($2, Parsetree,"PREF_AND",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
;

disjunctLines: /* <genSort> (Sorte des (Teil-)Ausdrucks) */
	PREF_OR IBL expr IBR     %prec PREF_OR
	{ $$ = UnaryMaxMatch($3,s_temporal,lastElem->pos,
                                           lastElem->leng,s_bool);
          Parsetree = new ParseExpr(tdisjunctLines,"disjunctLines",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,2);
          AppendtoList($1, Parsetree,"PREF_OR",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
    | 	disjunctLines PREF_OR IBL expr IBR
	{ $$ = BinaryMaxMatch(
                 UnaryMaxMatch($4,s_temporal,lastElem->pos,
                                             lastElem->leng,s_bool),
                 $1,s_temporal,lastElem->pos,lastElem->leng,s_bool);
          Parsetree = new ParseExpr(tdisjunctLines,"disjunctLines",3);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,3);
          AppendtoList($2, Parsetree,"PREF_OR",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
;

isElementOfSetListQB: /* <genSort> (Sorte des (Teil-)Ausdrucks) */
		/* Seiteneffekt: lokale Bezeichner binden */
    	normalID_csl _ELEMENT_OF_ SetExpr
        { $$ = UnaryMaxMatch($3,s_transfct,lastElem->pos,
                                           lastElem->leng,s_const);
	  SymListElem *e = $1->GetFirstElem();
	  while (e)
	    {
	    e->Set_nidt(nidt_boundVar);
	    e->SetSort(new Sort (s_const));
	    e = $1->GetNextElem();
	    }
	  Env->Push();
	  Env->Append($1);
          Parsetree = new ParseExpr(tisElementOfSetListQB,"isElementOfSetListQB",3);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,3);
          AppendtoList($2, Parsetree,"_ELEMENT_OF_",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |   isElementOfSetListQB SEMICOLON normalID_csl _ELEMENT_OF_ SetExpr
	{ $$ = BinaryMaxMatch(
                 UnaryMaxMatch($5,s_transfct,lastElem->pos,
                                             lastElem->leng,s_const),
                 $1,s_transfct,lastElem->pos,lastElem->leng,s_const);
	  SymListElem *e = $3->GetFirstElem();
	  while (e)
	    {
	    e->Set_nidt(nidt_boundVar);
	    e->SetSort(new Sort (s_const));
	    e = $3->GetNextElem();
	    }
	  Env->Append($3);
          Parsetree = new ParseExpr(tisElementOfSetListQB,"isElementOfSetListQB",5);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,5);
          AppendtoList($4, Parsetree,"_ELEMENT_OF_",4);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,3);
          AppendtoList($2, Parsetree,"SEMICOLON",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
;

CaseRA_list: /* <genSort> (Sorte des (Teil-)Ausdrucks) */
    	CaseRA
	{ $$ = $1;
          Parsetree = new ParseExpr(tCaseRA_list,"CaseRA_list",1);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
    |	CaseRA_list CaseRA
	{ $$ = BinaryMaxMatch($1,$2,s_transfct,lastElem->pos,
                                               lastElem->leng,s_const);
          Parsetree = new ParseExpr(tCaseRA_list,"CaseRA_list",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
;

CaseRA:	/* <genSort> (Sorte des (Teil-)Ausdrucks) */
	expr _RIGHTARROW_ expr SEMICOLON
	{ $$ = BoolConstToConst(
                 UnaryMaxMatch($1,s_action,$2->pos,$2->leng,s_bool),
                 UnaryMaxMatch($3,s_transfct,$4->pos,$4->leng,s_const));
          Parsetree = new ParseExpr(tCaseRA,"CaseRA",4);
          AppendtoList($4, Parsetree,"SEMICOLON",4);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,3);
          AppendtoList($2, Parsetree,"_RIGHTARROW_",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
;

/*
 * function and operator definition 
 */

SymbolDef_LHS: /* <selem> (Neu definierter Bezeichner) */ 
        normalID
	{ Env->SetDefaultSort(Env->GetDefaultOpSort()); }
	TypeAndSortDecl01
	{ if ($1->is_primed())
            SimpleError("prime not allowed here",$1->pos,$1->leng);
	  $1->SetSort($3);
	  $1->Set_nidt(nidt_operator);
	  Env->Push();
          UnknownIDEnv->Push();
	  $$ = $1;
          Parsetree = new ParseExpr(tSymbolDef_LHS,"SymbolDef_LHS",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,2);
          AppendtoList($1, Parsetree,"normalID",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
    |	normalID RBR_ON
	{ Env->SetDefaultSort(Env->GetDefaultParSort());
	  Env->Push();
          UnknownIDEnv->Push(); }
	formalPar_ssl RBR_OFF
	{ Env->SetDefaultSort(Env->GetDefaultOpSort()); }
	TypeAndSortDecl01
	{ if ($1->is_primed())
            SimpleError("prime not allowed here",$1->pos,$1->leng);
	  $4->Append($7);
	  $1->SetSort($4);
	  $1->Set_nidt(nidt_operator);
	  $$ = $1;
          Parsetree = new ParseExpr(tSymbolDef_LHS,"SymbolDef_LHS",5);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,5);
          AppendtoList($5, Parsetree,"RBR_OFF",4);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,3);
          AppendtoList($2, Parsetree,"RBR_ON",2);
          AppendtoList($1, Parsetree,"normalID",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
    |  	normalID '['
	{ Env->SetDefaultSort(Env->GetDefaultParSort());
	  Env->Push();
          UnknownIDEnv->Push(); }
	formalPar_ssl ']'
	{ Env->SetDefaultSort(Env->GetDefaultOpSort()); }
	TypeAndSortDecl01
	{ if ($1->is_primed())
            SimpleError("prime not allowed here",$1->pos,$1->leng);
	  delete $4;
	  $1->SetSort($7);
	  $1->Set_nidt(nidt_fctOperator);
	  Env->AppendInEnclosingFrame($1);
	  if (! (*($7) <= s_transfct))
	    SimpleError("functions must be of sort TRANS FCN or subsort",
                        $1->pos,$1->leng);
	  $$ = $1;
          Parsetree = new ParseExpr(tSymbolDef_LHS,"SymbolDef_LHS",5);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,5);
          AppendtoList($5, Parsetree,"]",4);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,3);
          AppendtoList($2, Parsetree,"[",2);
          AppendtoList($1, Parsetree,"normalID",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |	infix '_'
	{ Env->SetDefaultSort(Env->GetDefaultParSort());
	  Env->Push();
          UnknownIDEnv->Push(); }
	formalPar '_' normalID '_' formalPar '_'
	{ Env->SetDefaultSort(Env->GetDefaultOpSort()); }
	TypeAndSortDecl01
	{ if ($6->is_primed())
            SimpleError("prime not allowed here",$6->pos,$6->leng);
	  if (! ($4->is_Singleton() && $8->is_Singleton()) )
	    SimpleError("too many formal parameter for binary infix OP",
                        $6->pos,$6->leng);
	  $4->Append($8);
	  $4->Append($11);
	  $6->SetSort($4);
	  $6->Set_nidt(nidt_operator);
	  $$ = $6;
          Parsetree = new ParseExpr(tSymbolDef_LHS,"SymbolDef_LHS",9);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,9);
          AppendtoList($9,Parsetree,"_",8);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,7);
          AppendtoList($7,Parsetree,"_",6);
          AppendtoList($6,Parsetree,"normalID",5);
          AppendtoList($5,Parsetree,"_",4);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,3);
          AppendtoList($2,Parsetree,"_",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
;

formalPar_ssl: /* genSort (Liste der Parametersorten) */ 
        formalPar
	{ $$ = $1;
          Parsetree = new ParseExpr(tformalPar_ssl,"formalPar_ssl",1);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
    |	formalPar_ssl SEMICOLON formalPar
	{ $1->Append($3);
	  $$ = $1;
          Parsetree = new ParseExpr(tformalPar_ssl,"formalPar_ssl",3);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,3);
          AppendtoList($2,Parsetree,"SEMICOLON",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
;

formalPar: /* genSort (Liste der Parametersorten) */
	/* Parameter in Umgebung binden + Sortenzuweisung */
	normalID_csl TypeAndSortDecl01
	{ Sort *r = new Sort;
	  SortListElem *se = new SortListElem($2->GetFirstElem()->GetSort());
	  SymListElem *e = $1->GetFirstElem();
	  while (e!=0)
	    {
            if (isHeader) { e->Set_nidt(nidt_formalglobPar); }
	       else       { e->Set_nidt(nidt_formalPar);     }
	    e->SetSort(se);
	    r->Append(se);
	    e = $1->GetNextElem();
	    }
	  Env->Append($1);
	  delete $2;
	  $$ = r;
          Parsetree = new ParseExpr(tformalPar,"formalPar",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
    |  	normalID
	{ Env->Push();
          UnknownIDEnv->Push(); }
	RBR_ON formalPar_ssl RBR_OFF TypeAndSortDecl01
	{ if ($1->is_primed())
            SimpleError("prime not allowed here",$1->pos,$1->leng);
	  $4->Append($6);
          if (isHeader) { $1->Set_nidt(nidt_formalglobPar); }
             else       { $1->Set_nidt(nidt_formalPar);     }
	  $1->SetSort($4);
          UnknownIDEnv->AppendAndPopUnknownIDs(isHeader,Env);
	  Env->Pop();
	  Env->Append($1);
          Sort *r = new Sort(new SortListElem(s_nested,$4->copy()));
	  $$ = r;
          Parsetree = new ParseExpr(tformalPar,"formalPar",5);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,5);
          AppendtoList($5, Parsetree,"RBR_OFF",4);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,3);
          AppendtoList($3, Parsetree,"RBR_ON",2);
          AppendtoList($1, Parsetree,"normalID",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));    }
    |	infix 
	{ Env->Push();
          UnknownIDEnv->Push(); }
	'_' formalPar '_' normalID '_' formalPar '_' TypeAndSortDecl01
	{ if ($6->is_primed())
            SimpleError("prime not allowed here",$6->pos,$6->leng);
	  if (! ($4->is_Singleton() && $8->is_Singleton()) )
	    SimpleError("too many formal parameter for binary infix OP",
                        $6->pos,$6->leng);
	  $4->Append($8);
	  $4->Append($10);
          if (isHeader) { $6->Set_nidt(nidt_formalglobPar); }
            else        { $6->Set_nidt(nidt_formalPar);     }
	  $6->SetSort($4);
          UnknownIDEnv->AppendAndPopUnknownIDs(isHeader,Env);
	  Env->Pop();
	  Env->Append($6);
          Sort *r = new Sort(new SortListElem(s_nested,$4->copy()));
	  $$ = r;
          Parsetree = new ParseExpr(tformalPar,"formalPar",9);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,9);
          AppendtoList($9, Parsetree,"_",8);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,7);
          AppendtoList($7, Parsetree,"_",6);
          AppendtoList($6, Parsetree,"normalID",5);
          AppendtoList($5, Parsetree,"_",4);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,3);
          AppendtoList($3, Parsetree,"_",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));      }
;

infix: /* <selem> */ 
	INFIX	{ $$ = $1;
                  Parsetree = new ParseExpr(tinfix,"infix",1);
                  AppendtoList($1, Parsetree,"INFIX",1);
                  SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
    |	INFIXR	{ $$ = $1;
                  Parsetree = new ParseExpr(tinfix,"infix",1);
                  AppendtoList($1, Parsetree,"INFIXR",1);
                  SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
;

/*
 * Typen/Sortenangaben
 */

TypeAndSortDecl01: /* genSort */ 
	/* empty */
	{ $$ = new Sort(Env->GetDefaultSort());
          Parsetree = new ParseExpr(tTypeAndSortDecl01,"TypeAndSortDecl01", 1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |	COLON TypeAndSortDecl
	{ $$ = $2;
          Parsetree = new ParseExpr(tTypeAndSortDecl01,"TypeAndSortDecl01",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,2);
          AppendtoList($1, Parsetree,"COLON",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));    }
;

TypeAndSortDecl: /* genSort */ 
	TypeDecl
	{ $$ = new Sort(Env->GetDefaultSort());
          if (*($$) <= s_temporal)
	    SimpleError(
             "types not allowed in connection with sort TEMPORAL or subsort",
             lastElem->pos,lastElem->leng);
          Parsetree = new ParseExpr(tTypeAndSortDecl,"TypeAndSortDecl",1);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));    }
    |	TypeDecl Sort
	{ $$ = new Sort($2);
          if (*($$) <= s_temporal)
	    SimpleError(
             "types not allowed in connection with sort TEMPORAL or subsort",
             lastElem->pos,lastElem->leng);
          Parsetree = new ParseExpr(tTypeAndSortDecl,"TypeAndSortDecl",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
    |	Sort
	{ $$ = new Sort($1);
          Parsetree = new ParseExpr(tTypeAndSortDecl,"TypeAndSortDecl",1);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
;

TypeDecl:
	{ isTypeDecl++; } /* ANY darf nur in Typausdruck vorkommen */
    	SetExpr
	{ isTypeDecl--;
          if (! (*($2) <= s_const))
	    SimpleError("type definitions must have sort CONSTANT",
                        lastElem->pos,lastElem->leng);
          Parsetree = new ParseExpr(tTypeDecl,"TypeDecl",1);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
;

Sort: /* <primSort> */ 
	CONSTANT
        { $$ = s_const;
          Parsetree = new ParseExpr(tSort,"Sort",1);
          AppendtoList($1, Parsetree,"CONSTANT",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
    |   BOOLEAN
        { $$ = s_bool;
          Parsetree = new ParseExpr(tSort,"Sort",1);
          AppendtoList($1, Parsetree,"BOOLEAN",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
    |   PREDICATE
	{ $$ = s_pred; 
	  if (Env->GetModuleType() == mt_const)
	    SimpleError(
             "sort PREDICATE not allowed in CONSTANT MODULE",
             $1->pos,$1->leng);
          Parsetree = new ParseExpr(tSort,"Sort",1);
          AppendtoList($1, Parsetree,"PREDICATE",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
    |   STATE FUNCTION
        { $$ = s_statefct;
	  if (Env->GetModuleType() == mt_const)
	    SimpleError(
             "sort STATE FUNCTION not allowed in CONSTANT MODULE",
             $1->pos,$1->leng);
          Parsetree = new ParseExpr(tSort,"Sort",2);
          AppendtoList($2, Parsetree,"FUNCTION",2);
          AppendtoList($1, Parsetree,"STATE",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
    |   ACTION
	{ $$ = s_action;
	  if (Env->GetModuleType() == mt_const)
	    SimpleError(
             "sort ACTION not allowed in CONSTANT MODULE",
             $1->pos,$1->leng);
          Parsetree = new ParseExpr(tSort,"Sort",1);
          AppendtoList($1, Parsetree,"ACTION",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
    |   TRANSITION FUNCTION
	{ $$ = s_transfct;
	  if (Env->GetModuleType() == mt_const)
	    SimpleError(
             "sort TRANSITION FUNCTION not allowed in CONSTANT MODULE",
             $1->pos,$1->leng);
          Parsetree = new ParseExpr(tSort,"Sort",2);
          AppendtoList($2, Parsetree,"FUNCTION",2);
          AppendtoList($1, Parsetree,"TRANSITION",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));    }
    |   TEMPORAL
	{ $$ = s_temporal;
	  if (Env->GetModuleType() == mt_const)
	    SimpleError(
             "sort TEMPORAL not allowed in CONSTANT MODULE",
             $1->pos,$1->leng);
          Parsetree = new ParseExpr(tSort,"Sort",1);
          AppendtoList($1, Parsetree,"TEMPORAL",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
;

/*
 *  Bezeichner
 */

ModuleName: /* <selem> */
	normalID
	{ if ($1->is_primed())
            SimpleError("prime not allowed here",$1->pos,$1->leng);
	  $$ = $1;
          Parsetree = new ParseExpr(tModuleName,"ModuleName",1);
          AppendtoList($1, Parsetree,"normalID",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   } 
;
qualifyingID: /* <selem> */
	normalID
	{ if ($1->is_primed())
            SimpleError("prime not allowed here",$1->pos,$1->leng);
	  $$ = $1;
          Parsetree = new ParseExpr(tqualifyingID,"qualifyingID",1);
          AppendtoList($1, Parsetree,"normalID",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  } 
;
componentID: /* <selem> */
	normalID
	{ if ($1->is_primed())
            SimpleError("prime not allowed here",$1->pos,$1->leng);
	  $1->Set_nidt(nidt_rec_comp);
	  $1->SetSort(new Sort (s_const));
	  $1->SetPrime(0);
	  $$ = $1;
          Parsetree = new ParseExpr(tcomponentID,"componentID",1);
          AppendtoList($1, Parsetree,"normalID",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   } 
;

ID: /* <selem> */
	qualifierPrefixes normalID	%prec PRECLOWERTHANDOT
	{ $1->Append($2);
          if ( (Env->GetModuleType() == mt_const) && $2->is_primed() )
	     SimpleError("prime not allowed in CONSTANT MODULE",
             $2->pos,$2->leng);
	  if (UnknownID_allowed || isHeader) {	// Ausnahme zu declare before
            UnknownIDEnv->AppendUnknown($1);	// use
            $$ = dummy_selem;
            dummy_selem->pos  = $1->GetFirstElem()->pos;
            dummy_selem->leng = $1->GetFirstElem()->leng; }
          else {				// declare before use
            $$ = Env->UseOfID($1);
            delete $1;
	    $$->SetPrime($2->is_primed()); }
          Parsetree = new ParseExpr(tID,"ID",2);
          AppendtoList($2, Parsetree,"normalID",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
;

qualifierPrefixes: /* <slist> ( Liste der Qualifizierer) */
	/* empty */
	{ $$ = new SymList;
          Parsetree = new ParseExpr(tqualifierPrefixes,"qualifierPrefixes",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree)); }
   |	qualifierPrefixes normalID DOT	%prec DOT
	{ if ($2->is_primed())
            SimpleError("prime not allowed here",$2->pos,$2->leng);
          $1->Append($2);
          $$ = $1;
          Parsetree = new ParseExpr(tqualifierPrefixes,"qualifierPrefixes",3);
          AppendtoList($3, Parsetree,"DOT",3);
          AppendtoList($2, Parsetree,"normalID",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
   |	qualifierPrefixes normalID RBR_ON expr_csl RBR_OFF DOT  %prec DOT
	{ if ($2->is_primed())
            SimpleError("prime not allowed here",$2->pos,$2->leng);
          $2->SetSort($4);
          $1->Append($2);
          $$ = $1;
          Parsetree = new ParseExpr(tqualifierPrefixes,"qualifierPrefixes",6);
          AppendtoList($6, Parsetree,"DOT",6);
          AppendtoList($5, Parsetree,"RBR_OFF",5);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,4);
          AppendtoList($3, Parsetree,"RBR_ON",3);
          AppendtoList($2, Parsetree,"normalID",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
;

normalID_csl: /* <slist> (Liste aller Bezeichner) */ 
	normalID
	{ if ($1->is_primed())
            SimpleError("prime not allowed here",$1->pos,$1->leng);
          $$ = new SymList($1);
          Parsetree = new ParseExpr(tnormalID_csl,"normalID_csl",1);
          AppendtoList($1, Parsetree,"normalID",1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
    |	normalID_csl COMMA normalID
	{ $1->Append($3);
	  if ($3->is_primed())
            SimpleError("prime not allowed here",$3->pos,$3->leng);
	  $$ = $1;
          Parsetree = new ParseExpr(tnormalID_csl,"normalID_csl",3);
          AppendtoList($3, Parsetree,"normalID",3);
          AppendtoList($2, Parsetree,"COMMA",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
;

ID_csl:  /* <slist> (Liste aller Bezeichner) */ 
	ID
	{ if ($1->is_primed())
            SimpleError("prime not allowed here",$1->pos,$1->leng);
          $$ = new SymList($1);
          Parsetree = new ParseExpr(tID_csl,"ID_csl",1);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));  }
    |	ID_csl COMMA ID
	{ $1->Append($3);
	  if ($3->is_primed())
            SimpleError("prime not allowed here",$3->pos,$3->leng);
	  $$ = $1;
          Parsetree = new ParseExpr(tID_csl,"ID_csl",3);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,3);
          AppendtoList($2, Parsetree,"COMMA",2);
          PopandAppend(SemanticStack->SmStackPop(),Parsetree,1);
          SemanticStack->SmStackPush(new SmStackElem(Parsetree));   }
;

%%

static void startframe(int depth)
	// beginne eine Zeile im Inhaltsverzeichnis
 {
 BeginContentsLine(depth+ModuleDepth);
 AddToContentsLine(yytext);
 writeit=1;
 }

static void endframe(SymListElem* dest)
	// beende aktuelle Zeile im Inhaltsverzeichnis
 {
 EndContentsLine(dest->pos, dest->leng);
 writeit=0;
 }

static void writeframe(int depth, char* s1, char* s2, SymListElem* dest)
	// schreibe komplette Zeile ins Inhaltsverzeichnis
 {
 WriteContentsLine(depth+ModuleDepth, s1, s2, dest->pos, dest->leng);
 }

static void act_Recover()
	// Error-Recovery Massnahmen des Parsers
 {
 ModuleDepth = Env->Recover();
 UnknownIDEnv->Recover();
 isTypeDecl = UnknownID_allowed = 0;
 if (writeit) endframe(lastElem);
 }

static SymList* act_Import(SymListElem *ModName, SymListElem *qualifier)
	// importiere Modul ModName mit eventuellem Qualifizierer.
 {
 Module_Type mt;

 ModName->Set_nidt(nidt_imported);
 if (qualifier)			// Wenn Qualifizierer angegeben
   {
   qualifier->Set_nidt(nidt_qualifier);
   qualifier->SetSort(new Sort(s_const));
   }
 SymList *s = crossref->Import(ModName,qualifier,&mt);
 if (s)
   {
   Env->Append(s,0,ModName);  // Importierte Bezeichner in Umgebung eintragen
   if (qualifier)
     {
     Env->Append(qualifier);
     }
   }
 else
   {
   SimpleError("can't find exportlist",ModName->pos,ModName->leng);
   }
 if ( (mt == mt_temporal) && (Env->GetModuleType() == mt_const) )
   {
   SimpleError("import of temporal module into constant module not allowed",
               ModName->pos,ModName->leng);
   }
 return s; // SymList wird  zurueckgegeben ->om
 }

static void act_Include(SymListElem* ModName, SymListElem* qualifier)
	// Inclusion von Modul ModName vorbereiten.
 {
 Module_Type mt;

 IncludeList = crossref->Import(ModName,qualifier,&mt);
 if (! IncludeList)
   {
   SimpleError("can't find exportlist",ModName->pos,ModName->leng);
   }
 ModName->Set_nidt(nidt_included);
 if ( (mt == mt_temporal) && (Env->GetModuleType() == mt_const) )
   {
   SimpleError("inclusion of temporal module into constant module not allowed",
               ModName->pos,ModName->leng);
   }
 }

static Sort* SortOfID(SymListElem* ID)
	// berechne Sorte eines eventuell `ge-prime-ten' Bezeichners 
 {
 Sort *s = ID->GetSort()->copy();
 if (s->GetLength() > 1)
   {
   if (ID->is_primed())
     SimpleError("prime not allowed here",ID->pos,ID->leng);
   return new Sort (new SortListElem(s_nested,s));
   }
 else
   {
   Sort_Type t = s->GetFirstElem()->GetSort();
   if (ID->is_primed())
     {
     if      (t == s_pred)     t = s_action;
     else if (t == s_statefct) t = s_transfct;
     else if ( (t != s_const) && (t != s_bool) && (t != s_joker) )
       SimpleError("prime not allowed here",ID->pos,ID->leng);
     }
   return new Sort(t);
   }
 }

static void reinitialize (FILE* ifp)
	// Parser und Scanner in Anfangszustand versetzen.
 {
 ModuleDepth = 0;
 isTypeDecl  = 0;
 isHeader    = 1;
 UnknownID_allowed = 0;
 UnknownIDEnv = new Environment;
 ExportList = new SymList;
 IncludeList = 0;
 dummy_selem = new SymListElem (0,0,0,0,new id_info); // ->om: zusaetzlicher Parameter
 dummy_selem->SetSort(new Sort(s_joker));
 Parsetree = 0;
 SemanticStack = new SmStack;


 reinitialize_lex (ifp, actfilename);
 }

// Exporte

void yyerror (char* msg)
 {
 if (lastElem)
   SimpleError(msg,lastElem->pos,lastElem->leng);
 else
   SimpleError(msg,0,0);
 }

void parse_pass2 (char* suffixedfilename, char* filename)
	// Hauptfunktion parser2
 {
 FILE* ifp = 0;
 int err,warn,dummyerr;
 SymListElem* MName;
 SymList* actuallist;
 AdjList*  tmpadjl;
 SymList *Symbollist = new SymList; 
 char *dir, *libpath = NULL;

 Node* startnode; 
 actfilename=filename;
 ifp = fopen(suffixedfilename, "r");
 if (ifp == NULL)
   {
    dir = getenv("TLALIB");
    if(dir)
     {
       libpath = new char[strlen(dir) + strlen(suffixedfilename) + 2];

       sprintf(libpath,"%s/%s",dir,suffixedfilename);

       ifp = fopen(libpath, "r");
       if(ifp == NULL)
       {
          delete libpath;
          fprintf(stderr, "Can't open specification file '%s'.\n",suffixedfilename);
          fflush(stderr);
          return;
       }
     }
    else
    {
     fprintf(stderr, "Can't open specification file '%s'.\n",suffixedfilename);
     fflush(stderr);
     return;
    }
   }



 if (!OpenContentsFile(filename, "w"))
   {
   fprintf(stderr,"cannot open contentsfile for module %s\n",filename);
   fflush(stderr);
   fclose(ifp);
   return;
   }
 if (!OpenErrorFile(filename,"w"))
   {
   fprintf(stderr,"cannot open errorfile for module %s\n",filename);
   fflush(stderr);
   fclose(ifp);
   CloseContentsFile();
   return;
   }
 if(!OpenWarningFile(filename,"w"))
   {
   fprintf(stderr,"cannot open warningfile for module %s\n",filename);
   fflush(stderr);
   fclose(ifp);
   CloseContentsFile();
   dummyerr = CloseErrorFile();
   return;
   }
   
 reinitialize (ifp);
 yyparse();
 Parsetree = SemanticStack->SmStackPop();
 //fprintf(stderr,"afterParsing*** stackelements: %d \n",SemanticStack->GetLength()); // Debuginfo
 // SemanticStack->SmStackClear();


 
 assert(Parsetree != NULL);

 
 fclose(ifp);
 CloseContentsFile();
 UnknownIDEnv->AppendRestUnknownIDs(Env);
 fprintf(stderr,"typechecking in module %s ...\n",filename);
 fflush(stderr);
// Start der Typueberpruefung
 TypeChecker(Parsetree);
 actuallist = Env->Export(ExportList);
// Zusaetzliche Behandlung von declare-after-use Typnamen,
// die im aktuellen Modul nicht genutzt werden.
 MakeActualTypeInfos(actuallist);

 tmpadjl = new AdjList;


// Transformation des Ableitungsbaums in eine Adjazenzliste mit
// anschliessender Abspeicherung in Parsetree-Datei
 startnode = tmpadjl->PTAddNode(Parsetree->GetRuleName(),'N',Parsetree->minpos,Parsetree->maxpos);
 WriteParsetree(Parsetree,startnode, tmpadjl,filename); 
 MName = Env->GetModuleName();
 if (MName)
   {
   if (strcmp(MName->Getlexem(),filename)!=0 )
     SimpleError("filename and modulname not equal",MName->pos,MName->leng);
   }

// Parsetree-Expansion:
// Ableitungsbaeme includierten bzw. importierter Module werden mit
// dem aktuellen Ableitungsbaum verknuepft.

 CorrectTree(Parsetree);  // Record-Komponentennamen werden ermittelt und gesondert dargestellt.
 TreeImport(Parsetree,Parsetree,Symbollist);
 

 crossref->Insert(filename,MName,Env->GetModuleType(),sList,
                                 actuallist,Parsetree, Symbollist);
 
 err = CloseErrorFile();
 if (err)
   {
   fprintf(stderr,"found %d error(s).\n",err);
   fflush(stderr);
   }

 warn = CloseWarningFile();
 if(warn)
  {
  fprintf(stderr,"created %d warning(s).\n",warn);
  fflush(stderr);
  }


 delete tmpadjl;
 delete dummy_selem;
 delete Env;
 delete UnknownIDEnv;
 delete ExportList;
 delete SemanticStack;
 }
