
/*----------------------------------------------------------------------*
 * Parsetree.c
 * 
 * Olaf Meier, 16.6.94
 *
 * Enhaelt die Member-Funktionen zu den Klassen ParseExpr,ParseSymb, SmStack, 
 * SmStackElem sowie einige Hilfsroutinen.
 *---------------------------------------------------------------------------------*/

#include <assert.h>
#include "parsetree.h"
#include "dbinterface.h"
#include<string.h>

/*------------------------------------------------------------------------------------
 * Klasse ParseSymb
 *----------------------------------------------------------------------------------*/



ParseSymb::~ParseSymb()
         { 
            if (notermsymb)
             
               delete notermsymb;
             
             
         }

Symbol_Type ParseSymb::GetSymbolType()
	 {  assert((st == NonTerm) || (st == Terminal)||(st == Empty));
	     return st;
	 }

SymListElem* ParseSymb::GetSymbElem()
	 {   assert((st == Terminal) && (termsymb != 0));
	     return termsymb;
	 }

ParseExpr* ParseSymb::GetParseExpr()
	 {  assert((st == NonTerm) && (notermsymb != 0));
	    return notermsymb;
	 }


/*------------------------------------------------------------------------------------
 * Klasse ParseExpr
 *----------------------------------------------------------------------------------*/


ParseExpr::ParseExpr(noTerminal_Type tp,char* name, int i )
            : slist(NULL), minpos(0), maxpos(0)
  {  assert(i-1 >= 0);
     nt_type = tp;
     ruleName = new char[strlen(name) + 1]; // Kopiere Namen der Produktionsregel
     strcpy(ruleName,name);
     maxindex = i-1;
     parsesymbols =  new ParseSymb [i];
  }
  
ParseExpr::~ParseExpr()
  {  
     delete ruleName;
     delete [maxindex + 1] parsesymbols;
  }

void ParseExpr::AddExpr(ParseExpr *pe,int i)
 { 
   

   assert((maxindex >= i-1) && (i-1 >= 0));

   
   
   if(pe->maxpos > 0)
    {
     if(pe->maxpos > maxpos)                     // Min/Maxposition des Teilbaumes im
        maxpos = pe->maxpos;                     // Spezifikationsmodul eintragen
     if((pe->minpos > 0) || (minpos > pe->minpos)) // Epsilon-Produktionsregeln werden nicht betrachtet
       minpos = pe->minpos;
    }; 
   parsesymbols[i-1].notermsymb = pe;
   parsesymbols[i-1].st = NonTerm;
 }

void ParseExpr::AddExpr(SymListElem *se,int i)
  { 
    long sum;
    assert((maxindex >= i-1) && (i-1 >= 0));
    if(se->tokenID != 999)   //  Fehlertoken (ID = 999)
                             //  werden nicht betrachtet.      
      {
       sum = se->pos + (long) se->leng -1;  // Min/Maxposition des Teilbaumes im 
                                  // Spezifikationsmodul eintragen
       
       
          if(maxpos < sum)
            maxpos = sum;
          if((minpos > se->pos)|| (minpos == 0)) 
            minpos = se->pos;
       
      }; 
    parsesymbols[i-1].termsymb  = se;
    parsesymbols[i-1].st = Terminal;
   
  }
    
ParseSymb* ParseExpr::GetExpr(int i)
 {
   assert((maxindex >= i-1)&&(i-1 >=0));
   return &parsesymbols[i-1];
 }

int ParseExpr::GetMaxIndex()
 {
    return maxindex;
  }

char* ParseExpr::GetRuleName()
 {
    return ruleName;
 }

ParseExpr* ParseExpr::GetPE(int i)
 {
   assert(GetExpr(i)->GetSymbolType() == NonTerm);
   return GetExpr(i)->GetParseExpr();
 }

SymListElem* ParseExpr::GetSE(int i)
 {
   assert(GetExpr(i)->GetSymbolType() == Terminal);
   return GetExpr(i)->GetSymbElem();
 }

void ParseExpr::AddIncludes(SymList* inclist)
 {
   slist = inclist;
 }

SymList* ParseExpr::GetIncludes()
 {
   return slist;
 }

noTerminal_Type ParseExpr::GetNonTermType()
         {
            return nt_type;
         }



/*---------------------------------------------------------------------------------
 * Definition des semantischen Stacks
 *--------------------------------------------------------------------------------*/

         
/*--------------------------------------------------------------------------------
 * Klasse SmStack
 *-------------------------------------------------------------------------------*/

int SmStack::SmStackIsEmpty()
         { return (List::GetLength() == 0);
         }

void SmStack::SmStackPush(SmStackElem *se)
         { 

             List::NewFirstElem(se);
            
          };   


            

ParseExpr* SmStack::SmStackPop()
         {
           SmStackElem* tmpse;
           ParseExpr* tmppe;
           SymListElem *tmpsyle;
           int dummy;
           tmpse = (SmStackElem*) List::GetFirstElem();
           if (tmpse)     // Stack ist nicht leer !
             {tmppe = tmpse->pexpptr;
              List::DelFirstElem(); // Loesche oberstes Stackelement !

              return tmppe; // Rueckgabe des ParseExpr-Objekts
             }
           else  // Stack ist bereits leer !
            { tmpsyle = new SymListElem(0,5,0,999); // Fehler-Token generieren!! 
              tmppe = new ParseExpr(tunknown,"missing rule on stack",1);
              tmppe->AddExpr( tmpsyle, 1);
              return tmppe;
            }
         };

void SmStack::SmStackClear() // Alle Stackelemente loeschen !
          {
            fprintf(stderr,"\n ggf Stackinhalt: \n");
            while(List::GetLength() != 0)
              {
               SmStackElem *smel;
               smel = (SmStackElem*)  List::GetFirstElem();
               if (smel->pexpptr)      // Parse-Expression ist vorhanden !
              
               delete smel->pexpptr; // loeschen ! 
               List::DelFirstElem(); // Stackelement loeschen !
              };
            
          }; 

/*--------------------------------------------------------------------
 * Non-Member RoutinenFunktionsdefinitionen
 *------------------------------------------------------------------*/

void PopandAppend(ParseExpr *pexp,ParseExpr *pexplist,int i)
  { assert(pexp != 0);
    pexplist->AddExpr(pexp,i);
   
  };  

void AppendtoList(SymListElem *sle,ParseExpr *pexplist,char* toNa, int i)
  { assert(sle != NULL);
    sle->tokenName = new char[strlen(toNa) + 1];
    strcpy(sle->tokenName,toNa); 
    pexplist->AddExpr(sle,i);
  };     



void TraverseTree(ParseExpr* tree) // DFS-Baumdurchlauf
 {   int index,max;
     max = tree->GetMaxIndex();
     SymListElem* se;
     ParseSymb* tmpps;


//     fprintf(stderr,"Prod. Regel: %s\n",tree->GetRuleName());
//      fprintf(stderr,"min : %ld max: %ld \n",tree->minpos,tree->maxpos);
     
     for(index=1;index<=max+1; index++)
        { tmpps = tree->GetExpr(index);
          if (tmpps->GetSymbolType() == Terminal) // Blatt = Terminalsymbol erreicht!
             { se = tmpps->GetSymbElem();
               if(se->tokenID == 344)
            fprintf(stderr,"tokenName: %s ->min:%ld max:%ld\n",se->Getlexem(),se->pos,  se->pos + (long) se->leng -1);
             }
          else if (tmpps->GetSymbolType() == NonTerm) // innerer Knoten ...
             TraverseTree(tmpps->GetParseExpr()); // ... fahre mit DFS fort.
        };
  };   

void WriteParsetree(ParseExpr* tree,Node* currnode, AdjList* adj,char* filename)
 { 
  Node    *n,*m;

  if (!OpenParseTreeFile(filename, "w"))
   {
    fprintf(stderr,"cannot open treefile for module %s\n",filename);
    fflush(stderr);
    return;
   }
  else
   {

    ConverttoAdjList(tree,currnode, adj);
    n = adj->GetFirstNode();
    while(n)
     {
       WritePTNode(n->GetNodeType(), n->GetName(),n->GetPTminpos() , n->GetPTmaxpos());
       m = n->GetFirstEdge();
       while (m) // Alle Ausgangskanten je Knoten 
        {
         WritePTEdge(n->GetTypeOfCurrentEdge(), m->GetName(),m->GetPTminpos(),m->GetPTmaxpos());
         m = n->GetNextEdge();

        }                  
       n = adj->GetNextNode();        
     }   
    fprintf(stderr,"...parsetree has %d nodes\n",adj->GetLength());
    fflush(stderr);
    CloseParseTreeFile();
   }
  }
void ConverttoAdjList(ParseExpr* tree,Node* currnode, AdjList* adj)
 { 
   int index, max;
   ParseSymb* ps;
   ParseExpr* pe;
   Node* tmpnode;
   SymListElem* se;
   max = tree->GetMaxIndex();
   for(index=1;index<=max+1; index++)
    { ps = tree->GetExpr(index);
      if(ps->GetSymbolType() == Terminal)
       { 
         se = ps->GetSymbElem();


          tmpnode = adj->PTAddNode(se->tokenName,'T',se->pos, se->pos + (long) se->leng -1);

          currnode->PTAddEdge(tmpnode,'T');

          
          
         

       }
      else if(ps->GetSymbolType() == NonTerm)
       { 
         pe = ps->GetParseExpr();
         if(pe->maxpos > 0) 
          { 

            tmpnode = adj->PTAddNode(pe->GetRuleName(),'N',pe->minpos, pe->maxpos );

            currnode->PTAddEdge(tmpnode,'N');
         
            ConverttoAdjList( pe,tmpnode, adj);
          };
        };
     };
 }  
 
