
/*----------------------------------------------------------------------*
 * SymList.c
 *
 * Michael Geikowski, 14.7.93
 *
 * Enthaelt die Member-Funktionen zu den Klassen SymList, SymListElem
 * und id_info.
 *----------------------------------------------------------------------*/
/* Olaf Meier, 16.6.93
 *
 * Modul um zusaetzliche Routinen fuer Typinformationen ergaenzt.
 *----------------------------------------------------------------------*/

#include <assert.h>
#include <string.h>

#include "symlist.h"
#include "dbinterface.h"







/*----------------------------------------------------------------------*
 * Klasse id_info
 *----------------------------------------------------------------------*/

id_info::id_info(char* l, normalID_Type t, char p)
    :lexem(l), primed(p), nid_type(t), the_Sort(0), TypeInfo(NULL), TypeDefinition(0) , definingModule(0),
	 definedAt(0), nextUse(0), prevUse(0)
 {
 qualifier = new SymList();
 }

id_info::~id_info()
 {
 if (lexem) delete lexem;
 if (the_Sort) delete the_Sort;
 if (qualifier) delete qualifier;
 if (TypeInfo)   // ...falls Bezeichner Typinformationen besitzt...
    delete TypeInfo; // werden sie geloescht.->om
   
   
 }

/*----------------------------------------------------------------------*
 * Klasse SymListElem
 *----------------------------------------------------------------------*/

void SymListElem::SetDefinedAt(SymListElem *e)
 {
 SymListElem *f;

 assert(e && e->id_info_ptr && id_info_ptr);

 id_info_ptr->TypeDefinition = e;  // Bezeichner 'e' besitzt  aktuelle Typinfo  
 
 if (e->id_info_ptr->definedAt != NULL)	// Wenn e Import/Include aus anderem
   {					// Modul ist,...
   e = e->id_info_ptr->definedAt;	// ...dann ist die Definition
   }					// hier zu finden
 id_info_ptr->definedAt = e;
 
 if (e->id_info_ptr->prevUse==NULL)
   {					// Erste Anwendung
   e->id_info_ptr->prevUse = this;
   e->id_info_ptr->nextUse = this;
   id_info_ptr->prevUse = e;
   id_info_ptr->nextUse = e;
   }
 else
   {					// Weitere Anwendungen
   f=e->id_info_ptr->prevUse;		// Einsortieren nach Position in Text
   while (    (strcmp(id_info_ptr->definingModule,
		   f->id_info_ptr->definingModule)==0)
	   && (f->id_info_ptr->definedAt != 0)
	   && (f->pos > pos)
	 )
     {
     f=f->id_info_ptr->prevUse;
     }
   id_info_ptr->nextUse = f->id_info_ptr->nextUse;
   id_info_ptr->prevUse = f;
   f->id_info_ptr->nextUse = this;
   id_info_ptr->nextUse->id_info_ptr->prevUse = this;
   }
 }

SymListElem* SymListElem::copy()
 {
 SymListElem *newelem;
 SymListElem *qualelem;
 SymList *qual;
 id_info *ip;
 char *name;

 assert(id_info_ptr);
 qual = new SymList;		// kopiere Qualifiziererliste
 qualelem = id_info_ptr->qualifier->GetFirstElem();
 while (qualelem)
   {
   assert(qualelem->id_info_ptr);
   name = new char[strlen(qualelem->id_info_ptr->lexem)+1];
   strcpy(name,qualelem->id_info_ptr->lexem);
   ip = new id_info(name, nidt_qualifier);
   qual->Append(new SymListElem(0,0,0,0,ip)); // -> om: zusaetzlicher Parameter
   qualelem = id_info_ptr->qualifier->GetNextElem();
   }
 name = new char[strlen(id_info_ptr->lexem)+1];	// kopiere id_info
 strcpy(name,id_info_ptr->lexem);
 ip = new id_info(name, id_info_ptr->nid_type);
 newelem = new SymListElem(pos,leng,0,tokenID,ip); // ->om: zusaetzlicher Parameter
 if (id_info_ptr->the_Sort)			// kopiere evtl. Sorte
   {
   newelem->id_info_ptr->the_Sort = id_info_ptr->the_Sort->copy();
   }
 if (id_info_ptr->TypeInfo != NULL) // kopiere evtl. Typangaben
   { 
     newelem->id_info_ptr->TypeInfo = id_info_ptr->TypeInfo->Copy(); // Typinfo kopieren ->om
   }
 newelem->id_info_ptr->qualifier = qual;
 newelem->id_info_ptr->definingModule = id_info_ptr->definingModule;
 if (id_info_ptr->definedAt==NULL)
   {
   newelem->id_info_ptr->definedAt = this;
   }
 else // wenn this ein importiertes/includiertes Symbol beschreibt.
   {
   newelem->id_info_ptr->definedAt = id_info_ptr->definedAt;
   }
 return newelem;
 }

void SymListElem::AddQualifier(SymListElem *e)
 {
 assert(e && id_info_ptr && id_info_ptr->qualifier);
 id_info_ptr->qualifier->NewFirstElem(e);
 }

void SymListElem::SetQualifier(SymList *s)
 {
 assert(s &&  id_info_ptr
	  &&  id_info_ptr->qualifier
	  && (id_info_ptr->qualifier->GetLength()==0) ) ;
 delete id_info_ptr->qualifier;
 id_info_ptr->qualifier = s;
 }

SymList* SymListElem::GetQualifier()
 {
 assert(id_info_ptr);
 return (id_info_ptr->qualifier);
 }

int SymListElem::lexemcmp(SymListElem *e)
 {
 assert(e && e->id_info_ptr);
 assert(id_info_ptr);
 return strcmp(id_info_ptr->lexem,e->id_info_ptr->lexem) == 0;
 }

void SymListElem::WriteDest(char* s1, char* s2, char* s3, char* DestModName)
	// Schreibe einen Eintrag bzgl. des Symbols `this' in eine Link-Datei.
        // Das Symbol `this' muss ein Bezeichner sein.
        // `s1', `s2' und `s3' geben den Linktext an.
        // DestModName ist der Name eines Moduls, das Ziel des `Module'-Links
        // ist oder NULL, wenn kein solches Ziel existiert.
 {
 char *DefModName, *PrevModName, *NextModName;
 long  DefPos,      PrevPos,      NextPos;
 id_info* q;

 assert(s1 && s2 && s3 && id_info_ptr);
 q = id_info_ptr;
 if (q->definedAt)	// Definition des Bezeichners
   {
   assert(q->definedAt->id_info_ptr &&
          q->definedAt->id_info_ptr->definingModule );
   DefModName = q->definedAt->id_info_ptr->definingModule;
   DefPos     = q->definedAt->pos;
   }
 else
   {
   DefModName = NULL;
   DefPos     = 0;
   }
 if (q->prevUse)	// Letztes Auftreten
   {
   assert(q->prevUse->id_info_ptr &&
          q->prevUse->id_info_ptr->definingModule );
   PrevModName = q->prevUse->id_info_ptr->definingModule;
   PrevPos     = q->prevUse->pos;
   }
 else
   {
   PrevModName = NULL;
   PrevPos     = 0;
   }
 if (q->nextUse)	// Naechstes Auftreten
   {
   assert(q->nextUse->id_info_ptr &&
          q->nextUse->id_info_ptr->definingModule );
   NextModName = q->nextUse->id_info_ptr->definingModule;
   NextPos     = q->nextUse->pos;
   }
 else
   {
   NextModName = NULL;
   NextPos     = 0;
   }
 WriteLink(pos,leng,	// Schreibe Link in Datenbasis
           s1,s2,s3,
           DefPos , DefModName,
           PrevPos, PrevModName,
           NextPos, NextModName,
           0      , DestModName);
 }
 
/*----------------------------------------------------------------------*
 * Klasse SymList
 *----------------------------------------------------------------------*/

SymListElem* SymList::GetFirstElem()
 {
 return (SymListElem*) List::GetFirstElem();
 }

SymListElem* SymList::GetNextElem()
 {
 return (SymListElem*) List::GetNextElem();
 }

SymListElem* SymList::Lookup(SymListElem *e, SymList* qual)
 {
 SymListElem *q,*r,*s;
 normalID_Type t;

 assert(e && qual);
 q = GetFirstElem();
 while (q)
   {
   if (q->lexemcmp(e)) 	  	// Vergleiche Element mit Anfrage-
     {				// element.
 
     t = q->Get_nidt();			// Wenn q Parameter ist,
     if ( ( t == nidt_formalPar)     ||	// kann q ausgegeben werden
          ( t == nidt_formalglobPar) || // (da keine qualifier)
          ( t == nidt_substPar)     )
       {
       return q;
       }
     				// Sonst : Vergleiche qualifier
 
     assert(q->GetQualifier());
     r = q->GetQualifier()->GetFirstElem();
     s = qual->GetFirstElem();
     while (r && s && r->lexemcmp(s))
       {
       r = q->GetQualifier()->GetNextElem();
       s = qual->GetNextElem();
       }
     if ((r == 0) && (s == 0))
       {
       return q;			// qualifier gleich : gefunden
       }
     }
   q = GetNextElem();
   }
 return 0;
 }





void SymList::SubstPar_with_Type(SymListElem* par, TypeComp* typ)
 { assert(par && typ);
   TypeComp* t;
   SymListElem *se = GetFirstElem();  // erstes Element der Include-Liste 
   while(se)
   {
     if(se->GetTypeInfo())  // falls Typangaben vorhanden sind...
      { t = se->GetTypeInfo();
	t->TraverseTypeInfo(par->Getlexem(), typ);  // ...durchlaufe die Typangabe und tausche
                                                    //   Typ des Bezeichners 'par' durch Typ 'typ'
                                                    //   aus.
      };
     se = GetNextElem();
   };
 }






static void CreateHelpText1(char *s, char* t, Sort* sort)
	// Hilfsfunktion zu SymList::WriteLinkFile()
 {
 Sort_Type primSort = sort->GetLastElem()->GetSort();
 strcpy(s,t);
 switch (primSort)
   {
   case s_const:    strcat(s,"constant");	break;
   case s_bool:     strcat(s,"boolean");	break;
   case s_statefct: strcat(s,"state fcn");	break;
   case s_pred:     strcat(s,"predicate");	break;
   case s_transfct: strcat(s,"transition fcn"); break;
   case s_action:   strcat(s,"action");		break;
   case s_temporal: strcat(s,"temporal formula"); break;
   default:         strcat(s,"identifier"); break;
   };
 strcat(s," '");
 }

static void CreateHelpText2(char *s, Sort* sort)
	// Hilfsfunktion zu SymList::WriteLinkFile()
 {
 strcpy(s,"' (");
 if (sort->GetLength()>1)
   {
   strcat(s,"composite sort");
   }
 else
   {
   Sort_Type primSort = sort->GetLastElem()->GetSort();
   switch (primSort)
     {
     case s_const:    strcat(s,"CONSTANT");	break;
     case s_bool:     strcat(s,"BOOLEAN");	break;
     case s_statefct: strcat(s,"STATE FCN");	break;
     case s_pred:     strcat(s,"PREDICATE");	break;
     case s_transfct: strcat(s,"TRANSITION FCN"); break;
     case s_action:   strcat(s,"ACTION");		break;
     case s_temporal: strcat(s,"TEMPORAL"); 	break;
     default:         strcat(s,"???"); 		break;
     };
   }
 strcat(s,").");
 }

void SymList::WriteLinkFile(char* filename)
 {
 SymListElem* p;
 id_info*     q;
 char s[100];

 p = (SymListElem*) List::GetLastElem();	// Maximale Position
 if (! OpenLinkFile(filename, "w", p->pos))	// Oeffnen
   {
   fprintf(stderr,"cannot create linkfile for module %s\n",filename);
   fflush(stderr);
   return;
   }
 p = (SymListElem*) List::GetFirstElem();	// Liste durchlaufen
 while (p)
  {
  q = p->id_info_ptr;
  if (q)		// Nur Bezeichner werden geschrieben
    {
    switch (q->nid_type)// abhaengig vom Bezeichnertyp
      {
      case nidt_formalPar:
        assert(q->the_Sort);
        CreateHelpText2(s,q->the_Sort);
        if (q->definedAt)
          p->WriteDest("Application of local parameter '",q->lexem,s);
        else
          p->WriteDest("Declaration of local parameter '",q->lexem,s);
        break;
      case nidt_formalglobPar:
        assert(q->the_Sort);
        CreateHelpText2(s,q->the_Sort);
        if (q->definedAt)
          p->WriteDest("Application of global parameter '",q->lexem,s);
        else
          p->WriteDest("Declaration of global parameter '",q->lexem,s);
        break;
      case nidt_boundVar:
        assert(q->the_Sort);
        CreateHelpText2(s,q->the_Sort);
        if (q->definedAt)
          p->WriteDest("Application of bound variable '",q->lexem,s);
        else
          p->WriteDest("Declaration of bound variable '",q->lexem,s);
        break;
      case nidt_rec_comp:
        WriteLink(p->pos,p->leng,
                  "Record component name '",q->lexem,"'.",
                  0,NULL, 0,NULL, 0,NULL, 0,NULL);
        break;
      case nidt_operator:
      case nidt_fctOperator:
        assert(q->the_Sort);
        if (q->definedAt)
          CreateHelpText1(s,"Application of ",q->the_Sort);
        else
          CreateHelpText1(s,"Definition of ",q->the_Sort);
        p->WriteDest(s,q->lexem,"'.");
        break;
      case nidt_expOperator:
        assert(q->definedAt && q->the_Sort);
        CreateHelpText1(s,"Export of ",q->the_Sort);
        p->WriteDest(s,q->lexem,"'.");
        break;
      case nidt_assumption:
        if (q->definedAt)
          p->WriteDest("Application of assumption '",q->lexem,"'.");
        else
          p->WriteDest("Definition of assumption '",q->lexem,"'.");
        break;
      case nidt_expTheorem:
        assert(q->definedAt);
        p->WriteDest("Export of theorem '",q->lexem,"'.");
        break;
      case nidt_theorem:
        if (q->definedAt)
          p->WriteDest("Application of theorem '",q->lexem,"'.");
        else
          p->WriteDest("Definition of theorem '",q->lexem,"'.");
        break;
      case nidt_qualifier:
        if (q->definedAt)
          p->WriteDest("Application of qualifier '",q->lexem,"'.");
        else
          p->WriteDest("Declaration of qualifier '",q->lexem,"'.");
        break;
      case nidt_imported:
        p->WriteDest("Import of module '",q->lexem,"'.",q->lexem);
        break;
      case nidt_included:
        p->WriteDest("Inclusion of module '",q->lexem,"'.",q->lexem);
        break;
      case nidt_moduleName:
        p->WriteDest("(Sub-)module name '",q->lexem,"'.");
        break;
      case nidt_substPar:
        assert(q->the_Sort);
        CreateHelpText2(s,q->the_Sort);
        p->WriteDest("Substitution of parameter '",q->lexem,s);
        break;
      case nidt_normal:
        WriteLink(p->pos,p->leng,
                  "Sorry, I don't know anything about '",q->lexem,"'.",
                  0,NULL, 0,NULL, 0,NULL, 0,NULL);
        break;
      default:
        assert(0);
      };
    }
  p = (SymListElem*) List::GetNextElem();
  }
 CloseLinkFile();
 }


   
